/*
	uemcmd.c: Emulated CCID reader commands for UEM non-CCID reader
	Copyright (C) 2011-2017   Advanced Card Systems Ltd.
	Copyright (C) 2021-2021   MicroEM JSC

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with this library; if not, write to the Free Software Foundation,
	Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <config.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <pcsclite.h>
#include <ifdhandler.h>
#include <reader.h>

#include <math.h>

#include "misc.h"
#include "uemcmd.h"
#include "openct/proto-t1.h"
#include "ccid.h"
#include "defs.h"
#include "ccid_ifdhandler.h"
#include "debug.h"

//#define SCR_BUFFER_SIZE         320
#define SCR_BUFFER_SIZE         1024

// buffer sizes
//#define SCR_MAX_PACKET_SIZE     64
#define SCR_MAX_PACKET_SIZE     256
#define SCR_ALIGNED_BUFFER_SIZE \
    ((SCR_BUFFER_SIZE + SCR_MAX_PACKET_SIZE - 1) & ~(SCR_MAX_PACKET_SIZE - 1))


#define UEM_HEADER_SIZE			1
#define UEM_STATUS_OFFSET		0

// Card type
#define CARD_TYPE_MCU_AUTO		0
#define CARD_TYPE_I2C_AUTO		1	// Not supported
#define CARD_TYPE_I2C_1K		2
#define CARD_TYPE_I2C_2K		3
#define CARD_TYPE_I2C_4K		4
#define CARD_TYPE_I2C_8K		5
#define CARD_TYPE_I2C_16K		6
#define CARD_TYPE_I2C_32K		7
#define CARD_TYPE_I2C_64K		8
#define CARD_TYPE_I2C_128K		9
#define CARD_TYPE_I2C_256K		10
#define CARD_TYPE_I2C_512K		11
#define CARD_TYPE_I2C_1024K		12
#define CARD_TYPE_AT88SC153		13
#define CARD_TYPE_AT88SC1608	14
#define CARD_TYPE_SLE4418		15
#define CARD_TYPE_SLE4428		16
#define CARD_TYPE_SLE4432		17
#define CARD_TYPE_SLE4442		18
#define CARD_TYPE_SLE4406		19
#define CARD_TYPE_SLE4436		20
#define CARD_TYPE_SLE5536		21
#define CARD_TYPE_MCU_T0		22
#define CARD_TYPE_MCU_T1		23
#define CARD_TYPE_AUTO			24	// Not supported

// UEM card type
#define UEM_CARD_TYPE_MCU_AUTO				0x00
#define UEM_CARD_TYPE_I2C_1K_16K			0x01
#define UEM_CARD_TYPE_I2C_32K_1024K			0x02
#define UEM_CARD_TYPE_AT88SC153				0x03
#define UEM_CARD_TYPE_AT88SC1608			0x04
#define UEM_CARD_TYPE_SLE4418_4428			0x05
#define UEM_CARD_TYPE_SLE4432_4442			0x06
#define UEM_CARD_TYPE_SLE4406_4436_5536		0x07
#define UEM_CARD_TYPE_MCU_T0				0x0C
#define UEM_CARD_TYPE_MCU_T1				0x0D

#define CS_ISO14443A							0x01
#define CS_ISO14443B							0x02
#define CS_ISO15693								0x04

#define CT_UNKNOWN								0x00
#define CT_MIFARE_CLASSIC						0x01
#define CT_MIFARE_ULTRALIGHT					0x02
#define CT_MIFARE_PLUS							0x03
#define CT_MIFARE_DESFIRE						0x04
#define CT_MIFARE_SMARTMX						0x05
#define CT_SAM									0x10

#define CT_ICODE_NXP							0x11
#define CT_ICODE_INFINEON						0x12
#define CT_ISO15693_TEXASINSTRUMENTS			0x16
#define CT_ISO15693_EMMICROELECTRONICMARINSA	0x17

#define TRUE	0xFF
#define FALSE	0x00

#define BOOLEAN BYTE
#define F_DATA_RATE     			0x02
#define F_ADDRESS       			0x20
#define F_OPTION        			0x40
#define ICODE_BLOCK_SIZE       		4
#define FLAG_CRD					0x02

// UEM option
#define UEM_OPTION_NONE			0x00	// No option
#define UEM_OPTION_EMV_MODE		0x10	// EMV mode
#define UEM_OPTION_MEMCARD_MODE	0x20	// Memory card mode

//#define UEM_CARD_INACTIVE_MAX_CYCLES	2
#define UEM_CARD_INACTIVE_MAX_CYCLES	1

#define CHECK_STATUS(res) \
	if (STATUS_NO_SUCH_DEVICE == res) \
		return IFD_NO_SUCH_DEVICE; \
	if (IFD_SUCCESS != res) \
		return IFD_COMMUNICATION_ERROR;

//#define MAX_T1_BLOCK_SIZE           270
#define MAX_T1_BLOCK_SIZE           1024

ULONG FSD[] = { 16, 24, 32, 40, 48, 64,  96, 128, 256,
                                512, 512, 512, 512, 512, 512, 512 };
ULONG DRS[] = { 106, 212, 424, 847 };
ULONG TA = 0x1A90;

static RESPONSECODE UEM_CmdXfrBlockTPDU_T0(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[]);

static RESPONSECODE UEM_CmdXfrBlockTPDU_T1(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[]);

static RESPONSECODE UEM_CmdXfrBlockTPDU_RAW(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[]);

static RESPONSECODE UEM_CmdSelectCardType(unsigned int reader_index, BYTE card_type);

static RESPONSECODE UEM_CmdSetOption(unsigned int reader_index, BYTE option);

RESPONSECODE UEM_CmdCRD_RF_up(unsigned int reader_index, UCHAR mode) 
{
	BYTE cmd[MAX_T1_BLOCK_SIZE];
	unsigned int cmd_length;
	status_t res;
	BYTE OutputBuffer[MAX_T1_BLOCK_SIZE];
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CmdCRD_RF_up");
#endif

    cmd[0] = 0x51;//CMD_RF;
    cmd[1] = mode; // 3-2 from icc, 1-0 - to icc
    cmd_length = 2;
    
	return_value = UEM_Transcieve(reader_index, 
		cmd, cmd_length,
		OutputBuffer, &length);
	if (return_value != IFD_SUCCESS)
		return return_value;
		
    if (OutputBuffer[0] == 0x00) {
        return IFD_SUCCESS;
    } else {
        return IFD_COMMUNICATION_ERROR;//STATUS_INVALID_DEVICE_STATE;
    }
}

RESPONSECODE UEM_CmdSetPPS_A(unsigned int reader_index) 
{
	BYTE cmd[MAX_T1_BLOCK_SIZE];
	unsigned int cmd_length;
	status_t res;
	BYTE OutputBuffer[MAX_T1_BLOCK_SIZE];
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
    //UCHAR CID = (UCHAR) ccid_descriptor->uemCID & 0x0F;
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
    DEBUG_CRITICAL("UEM_CmdSetPPS_A");
#endif
    
    UEM_CmdCRD_RF_up(reader_index, ccid_descriptor->uemDataRateCurrent);
    
    /*cmd[0] = 0x48;//CMD_CARD;
    cmd[1] = 0x03; 
    cmd[2] = 0x00;
    cmd[3] = 0xD0;// | CID;//SET_PPS | CID;
    cmd[4] = 0x11;
    cmd[5] = ccid_descriptor->uemDataRateNew;
    cmd[6] = TA & 0xFF; 
    cmd[7] = (TA >> 8) & 0xFF; 
    cmd[8] = 0x00; 
    cmd[9] = 0x00;
    cmd_length = 10;*/
    
    cmd[0] = 0x36;//PPS
    cmd[1] = 0x00;//CID
    cmd[2] = ccid_descriptor->uemDataRateNew;
    cmd_length = 3;
    
	return_value = UEM_Transcieve(reader_index, 
		cmd, cmd_length,
		OutputBuffer, &length);
		
	if (return_value != IFD_SUCCESS)
		return return_value;
    
    if (OutputBuffer[0] == 0) {// && OutputBuffer[3] == (0xD0/* | CID*/)) {
        ccid_descriptor->uemDataRateCurrent = ccid_descriptor->uemDataRateNew;
        UEM_CmdCRD_RF_up(reader_index, ccid_descriptor->uemDataRateCurrent);
		ccid_descriptor->uemReaderCapabilitiesCurrentState = SCARD_NEGOTIABLE;
        return IFD_SUCCESS;
    }
    
    // Reset controller
    
	/*return_value = UEM_Transcieve(reader_index, 
		"\004\200\001", 3,
		OutputBuffer, &length);
		
    sleep(1);
    
	return_value = UEM_Transcieve(reader_index, 
		"\020", 1,
		OutputBuffer, &length);*/
		
	DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
	
    return IFD_COMMUNICATION_ERROR;//STATUS_INVALID_DEVICE_STATE;
}

RESPONSECODE UEM_CmdPowerOn(unsigned int reader_index, unsigned int *pATRLength,
	/*@out@*/ BYTE pATR[], int voltage)
{
	BYTE cmd[256];
	unsigned int cmd_length;
	status_t res;
	BYTE OutputBuffer[256];
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	// store length of buffer[]
	length = *pATRLength;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL3("UEM_CmdPowerOn: reader_index = %d, ccid_descriptor->bCurrentSlotIndex = %d", reader_index, ccid_descriptor->bCurrentSlotIndex);
	//DEBUG_CRITICAL2("fSupportsTCL = %d", ccid_descriptor->uemfSupportsTCL);
#endif

	//if (ccid_descriptor->uemfSessionActive == TRUE)
	//	return return_value;
		
	if (ccid_descriptor->bCurrentSlotIndex == 0)
	{
		BYTE*					ATQ;
		BYTE					SAK;
		UCHAR 					ATR_stub[32];
		
		ULONG 					i;
		
		static BYTE 			FakeATR[13] = {0x3b, 0x89, 0x40, 0x14, 0x47, 0x47, 0x32, 0x34, 0x4d, 0x35, 0x32, 0x38, 0x30};
		static ULONG 			FakeATRLength = 13;
		
		// UL C		04 2E 5F FD E1 ED 25 80 A9 48 00 00 E1 10 12 00
		// Metro	04 84 E1 E9 82 B4 1F 82 AB 48 F8 FF FF FF FF FC
		// UL New	04 37 81 3A 82 73 1E 80 6F 48 00 00 00 00 00 00
		// NTAG203	04 85 30 39 4A B6 28 80 54 48 00 00 E1 10 12 00
		// NTAG213	04 BA B0 86 02 D9 38 80 63 48 00 00 E1 10 12 00
		// NTAG210	04 05 6E E7 DA 44 34 81 2B 48 00 00 E1 10 06 00
		// NTAG216	04 81 A1 AC 72 D5 38 81 1E 48 00 00 E1 10 6D 00

		BYTE	ATR_Simple_T0T1[] = {0x3B, 0x80, 0x80, 0x01, 0x01};
		
		// NFC P2P Device	3B 8F 80 01 FF 4E 46 43 20 50 32 50 20 44 65 76 69 63 65 B0
		BYTE	ATR_NFC_P2P[] = {0x3B, 0x8F, 0x80, 0x01, 0xFF, 0x4E, 0x46, 0x43, 0x20, 0x50, 0x32, 0x50, 0x20, 0x44, 0x65, 0x76, 0x69, 0x63, 0x65, 0xB0};
		// MF Mini	3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 00 26 00 00 00 00 4D
		//          3B 8F 80 01 80 4F 0C A0 00 00 03 06 .. 00 26 00 00 00 00 ..
		BYTE	ATR_MF_Mini[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x26, 0x00, 0x00, 0x00, 0x00, 0x4D};
		// MF 1k	3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 00 01 00 00 00 00 6A
		//          3B 8F 80 01 80 4F 0C A0 00 00 03 06 .. 00 01 00 00 00 00 ..
		BYTE	ATR_MF_1k[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x6A};
		// MF 4k	3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 00 02 00 00 00 00 69
		//          3B 8F 80 01 80 4F 0C A0 00 00 03 06 .. 00 02 00 00 00 00 ..
		BYTE	ATR_MF_4k[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x69};
		// MF UL	3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 00 03 00 00 00 00 68
		//          3B 8F 80 01 80 4F 0C A0 00 00 03 06 .. 00 03 00 00 00 00 ..
		BYTE	ATR_MF_UL[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x68};
		// MF DF	3B 81 80 01 80 80
		BYTE	ATR_MF_DF[] = {0x3B, 0x81, 0x80, 0x01, 0x80, 0x80};
		// MF PS3	3B 87 80 01 C1 05 2F 2F 00 35 C7 30
	//	BYTE	ATR_MF_PS3[] = {0x3B, 0x87, 0x80, 0x01, 0xC1, 0x05, 0x2F, 0x2F, 0x00, 0x35, 0xC7, 0x30};
		// MF PX2	3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 FF 11 00 00 00 00 85
		BYTE	ATR_MF_PX2[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x03, 0xFF, 0x11, 0x00, 0x00, 0x00, 0x00, 0x85};
		// MF PX3	3B 87 80 01 C1 05 2F 2F 01 BC D6 A9
		BYTE	ATR_MF_PX3[] = {0x3B, 0x87, 0x80, 0x01, 0xC1, 0x05, 0x2F, 0x2F, 0x01, 0xBC, 0xD6, 0xA9};
		// MF Prox	3B 8E 80 01 80 42 9F 40 55 50 45 52 46 11 73 04 21 00 01
	//	BYTE	ATR_MF_Prox[] = {0x3B, 0x8E, 0x80, 0x01, 0x80, 0x42, 0x9F, 0x40, 0x55, 0x50, 0x45, 0x52, 0x46, 0x11, 0x73, 0x04, 0x21, 0x00, 0x01};
		// MF Smart	3B 88 80 01 00 73 C8 40 13 00 90 00 71
		BYTE	ATR_MF_Smart[] = {0x3B, 0x88, 0x80, 0x01, 0x00, 0x73, 0xC8, 0x40, 0x13, 0x00, 0x90, 0x00, 0x71};
		// ICode NXP   3B 8F 80 01 80 4F 0C A0 00 00 03 06 0B 00 14 00 00 00 00 77
		BYTE	ATR_ICode_NXP[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x0B, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x77};
		// ICode Inf   3B 8F 80 01 80 4F 0C A0 00 00 03 06 0B 00 0E 00 00 00 00 6D
		BYTE	ATR_ICode_Infineon[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x0B, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x77};
		// ICode TI    3B 8F 80 01 80 4F 0C A0 00 00 03 06 0B 00 12 00 00 00 00 71
		BYTE	ATR_ISO15693_TexasInstrument[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x0B, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x77};
		// ICode EM    3B 8F 80 01 80 4F 0C A0 00 00 03 06 0B 00 00 00 00 00 00 63
		BYTE	ATR_ISO15693_EMMicroelectronicMarinSA[] = {0x3B, 0x8F, 0x80, 0x01, 0x80, 0x4F, 0x0C, 0xA0, 0x00, 0x00, 0x03, 0x06, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x63};
		// https://smartcard-atr.appspot.com/parse?ATR=3B8F8001804F0CA0000003060300260000000045
		/* 
		Code 0x01: Motorola
		Code 0x02: ST Microelectronics
		Code 0x03: Hitachi
		Code 0x04: NXP Semiconductors
		Code 0x05: Infineon Technologies
		Code 0x06: Cylinc
		Code 0x07: Texas Instruments Tag-it™
		Code 0x08: Fujitsu Limited
		Code 0x09: Matsushita Electric Industrial
		Code 0x0A: NEC
		Code 0x0B: Oki Electric
		Code 0x0C: Toshiba
		Code 0x0D: Mitsubishi Electric
		Code 0x0E: Samsung Electronics
		Code 0x0F: Hyundai Electronics
		Code 0x10: LG Semiconductors
		Code 0x16: EM Microelectronic-Marin
		Code 0x1F: Melexis
		Code 0x2B: Maxim
		Code 0x33: AMIC
		*/
		
		if (ccid_descriptor->uemfModeTCL == TRUE) {
			*pATRLength = ccid_descriptor->uemATRLength;
			if (ccid_descriptor->uemATRLength > 0) {
				memcpy(pATR, ccid_descriptor->uemATR, ccid_descriptor->uemATRLength);
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
				DEBUG_CRITICAL2("uemfModeTCL == TRUE, return IFD_SUCCESS, ATRLength = %d", ccid_descriptor->uemATRLength);
#endif
				return IFD_SUCCESS;
			} else {
				ccid_descriptor->uemfPowerOff = TRUE;
				//ccid_descriptor->uemfModeTCL == FALSE;
				return_value = UEM_CmdGetSlotStatus(reader_index, OutputBuffer);
			}
		}
		
		if (ccid_descriptor->uemCardStatus != CCID_ICC_PRESENT_ACTIVE) {
			DEBUG_CRITICAL("IFD_ERROR_POWER_ACTION");
			return IFD_ERROR_POWER_ACTION;
		}
		
		ccid_descriptor->uemCardType = CT_UNKNOWN;
		
		switch (ccid_descriptor->uemCardStandard)
		{
		case CS_ISO14443A:
		
			ATQ = ccid_descriptor->uemCardATQ;
			SAK = ccid_descriptor->uemCardSAK;

			if( SAK == 0x28 && ATQ[0] == 0x04 && ATQ[1] == 0x00 )
			{
				//str += "     Smart MX";
				ccid_descriptor->uemCardType = CT_MIFARE_SMARTMX;
				*pATRLength = sizeof(ATR_MF_Smart);
				memmove(
					pATR,
					ATR_MF_Smart,
					*pATRLength
				);
			}
			
			if( SAK == 0x00 && ATQ[0] == 0x44 && ATQ[1] == 0x00 )  // if it is ULTRALight Card
			{
				if (
				(ccid_descriptor->uemCardUID[0] == 0x05) && 
				((ccid_descriptor->uemCardUID[1] & 0xF0) == 0x30))
				{
					//str += "     my-d move";
					ccid_descriptor->uemCardType = CT_MIFARE_ULTRALIGHT;
					*pATRLength = sizeof(ATR_MF_UL);
					memmove(
						pATR,
						ATR_MF_UL,
						*pATRLength
					);
				}
				else
				{
					//str += "     Mifare Ultralight (C)";
					ccid_descriptor->uemCardType = CT_MIFARE_ULTRALIGHT;
					*pATRLength = sizeof(ATR_MF_UL);
					memmove(
						pATR,
						ATR_MF_UL,
						*pATRLength
					);
				}
			}

			if( SAK == 0x09 && ATQ[0] == 0x04 && ATQ[1] == 0x00 )
			{
				//str += "     Mifare Mini";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_Mini);
				memmove(
					pATR,
					ATR_MF_Mini,
					*pATRLength
				);
			}
			
			if( SAK == 0x08 && ATQ[0] == 0x04 && ATQ[1] == 0x00 )
			{
				//str += "     Mifare Classic 1k";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_1k);
				memmove(
					pATR,
					ATR_MF_1k,
					*pATRLength
				);
			}
			
			if( SAK == 0x18 && ATQ[0] == 0x02 && ATQ[1] == 0x00 )
			{
				//str += "     Mifare Classic 4k";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_4k);
				memmove(
					pATR,
					ATR_MF_4k,
					*pATRLength
				);
			}
			
			if( SAK == 0x08 && ATQ[0] == 0x44 && ATQ[1] == 0x00 )
			{
				//str += "     NXP Mifare Classic 1k / Plus 2k 7 Byte UID";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_1k);
				memmove(
					pATR,
					ATR_MF_1k,
					*pATRLength
				);
			}
			
			if( SAK == 0x88 && ATQ[0] == 0x44 && ATQ[1] == 0x00 )
			{
				//str += "     Infineon Mifare Classic 1k / Plus 2k 7 Byte UID";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_1k);
				memmove(
					pATR,
					ATR_MF_1k,
					*pATRLength
				);
			}
			
			if( SAK == 0x18 && ATQ[0] == 0x42 && ATQ[1] == 0x00 )
			{
				//str += "     Mifare Classic 4k / Plus 4k 7 Byte UID";
				ccid_descriptor->uemCardType = CT_MIFARE_CLASSIC;
				*pATRLength = sizeof(ATR_MF_4k);
				memmove(
					pATR,
					ATR_MF_4k,
					*pATRLength
				);
			}
			
			if( ATQ[0] == 0x04 && ATQ[1] == 0x00 )
			{
				if ( SAK == 0x10 )
				{
					//str += "     Mifare Plus SL2 2k 4 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX2);
					memmove(
						pATR,
						ATR_MF_PX2,
						*pATRLength
					);
				}
				if ( SAK == 0x20 )
				{
					//str += "     Mifare Plus SL3/SL0 2k 4 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX3);
					memmove(
						pATR,
						ATR_MF_PX3,
						*pATRLength
					);
				}
			}

			if( ATQ[0] == 0x02 && ATQ[1] == 0x00 )
			{
				if ( SAK == 0x11 )
				{
					//str += "     Mifare Plus SL2 4k 4 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX2);
					memmove(
						pATR,
						ATR_MF_PX2,
						*pATRLength
					);
				}
				if ( SAK == 0x20 )
				{
					//str += "     Mifare Plus SL3/SL0 4k 4 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX3);
					memmove(
						pATR,
						ATR_MF_PX3,
						*pATRLength
					);
				}
			}

			if( ATQ[0] == 0x44 && ATQ[1] == 0x00 )
			{
				if ( SAK == 0x10 )
				{
					//str += "     Mifare Plus SL2 2k 7 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX2);
					memmove(
						pATR,
						ATR_MF_PX2,
						*pATRLength
					);
				}
				if ( SAK == 0x20 )
				{
					//str += "     Mifare Plus SL3/SL0 2k 7 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX3);
					memmove(
						pATR,
						ATR_MF_PX3,
						*pATRLength
					);
				}
			}

			if( ATQ[0] == 0x42 && ATQ[1] == 0x00 )
			{
				if ( SAK == 0x11 )
				{
					//str += "     Mifare Plus SL2 4k 7 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX2);
					memmove(
						pATR,
						ATR_MF_PX2,
						*pATRLength
					);
				}
				if ( SAK == 0x20 )
				{
					//str += "     Mifare Plus SL3/SL0 4k 7 Byte UID";
					ccid_descriptor->uemCardType = CT_MIFARE_PLUS;
					*pATRLength = sizeof(ATR_MF_PX3);
					memmove(
						pATR,
						ATR_MF_PX3,
						*pATRLength
					);
				}
			}

			if( SAK == 0x20 && ATQ[0]== 0x44 && ATQ[1] == 0x03 ) // if it is DESFire Card
			{
				//str += "     Mifare DESFire (EV1)\r\nNFC Forum Type 4 Tag";
				ccid_descriptor->uemCardType = CT_MIFARE_DESFIRE;
				*pATRLength = sizeof(ATR_MF_DF);
				memmove(
					pATR,
					ATR_MF_DF,
					*pATRLength
				);
			}
			
			if( SAK == 0x40 && ATQ[0]== 0x01 && ATQ[1] == 0x01 ) // if it is Prestigio NFC tablet
			{
				//str += "     ISO/IEC 18092 (NFC P2P)";
				//ACR122 - Nokia N9: 3B 8F 80 01 80 4F 0C A0 00 00 03 06 03 FF 40 00 00 00 00 D4
				//PR533 - Rinkai Suica Octopus, MTR network from Hong Kong, 2014: 3B 8F 80 01 80 4F 0C A0 00 00 03 06 11 00 3B 00 00 00 00 42
				//ccid_descriptor->uemCardType = CT_ISO14443A;
				ccid_descriptor->uemCardType = CT_UNKNOWN;
				*pATRLength = sizeof(ATR_NFC_P2P);
				memmove(
					pATR,
					ATR_NFC_P2P,
					*pATRLength
				);
			}
			
			if( ( SAK & 0x44 ) == 0x40 )
			{
				//str += "     ISO/IEC 18092 (NFC P2P)";
				//ccid_descriptor->uemCardType = CT_ISO14443A;
				ccid_descriptor->uemCardType = CT_UNKNOWN;
				*pATRLength = sizeof(ATR_NFC_P2P);
				memmove(
					pATR,
					ATR_NFC_P2P,
					*pATRLength
				);
			}
			
			if( ( SAK & 0x44 ) == 0x00 )
			{
				//str += ""; //not compliant with ISO/IEC 18092
			}
			
			if( ( SAK & 0x24 ) == 0x00 )
			{
				//str += ""; //not compliant with ISO/IEC 14443-4
			}
			
			if( ( SAK & 0x24 ) == 0x20 )
			{
				//str += "     T=CL";
				ccid_descriptor->uemfSupportsTCL = TRUE;
			}
			
			if (
				//ccid_descriptor->uemCardType == CT_MIFARE_PLUS || 
				ccid_descriptor->uemCardType == CT_MIFARE_CLASSIC || 
				//ccid_descriptor->uemCardType == CT_MIFARE_DESFIRE ||
				ccid_descriptor->uemCardType == CT_MIFARE_ULTRALIGHT
			)
			{
				ccid_descriptor->uemfSupportsTCL = FALSE;
			}
				
			if (ccid_descriptor->uemfSupportsTCL == TRUE)
			{
				BYTE ATS[256*256] = { 0 };
				int nATS = 0;
				
#ifdef UEM_USE_EXTENDED_ACTIVATION
				cmd[0] = 0x2B;
				cmd[1] = 0x00;// | m_CID & 0x0F;
				cmd[2] = 0x00;
				cmd_length = 3;

				return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						OutputBuffer, &length);
						
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("return_value = %ul", return_value);
					return return_value;
				}
						
				if (length < 3) {
					DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
					return IFD_COMMUNICATION_ERROR;
				}
#else
				/*cmd[0] = 0x48;//CMD_CARD;
				cmd[1] = 0x02; 
				cmd[2] = 0x00;
				cmd[3] = 0xE0;//GET_PPS;
				cmd[4] = 0x50;//0x5F;//0x50 | (UCHAR)(DeviceExtension->CID & 0x0F);
				cmd[5] = TA & 0xFF;//0x90;// 
				cmd[6] = (TA >> 8) & 0xFF;//0x1A;// 
				cmd[7] = 0x00; 
				cmd[8] = 0x00;
				cmd_length = 9;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					OutputBuffer, &length);*/
					
				cmd[0] = 0x35; //RATS
				cmd[1] = 0x00; //CID
				cmd_length = 2;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					OutputBuffer, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("return_value = %ul", return_value);
					return return_value;
				}
					
				if (length < 3) {
					DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
					return IFD_COMMUNICATION_ERROR;
				}
#endif
				if (OutputBuffer[0] == 0x00) { // PPS gotten
					memcpy(&nATS, &OutputBuffer[1], 2);
					memcpy(ATS, &OutputBuffer[3], nATS);

					memcpy(&ccid_descriptor->uemATSLength, &nATS, 2);
					memcpy(ccid_descriptor->uemATS, ATS, ccid_descriptor->uemATSLength);

					ULONG idx=2;

					ccid_descriptor->uemIFSC = ATS[1] & 0x0F;
					ccid_descriptor->uemIFSD = UEM_READER_IFSD;
					if (ATS[1] & 0x10) { // have TA1
						ccid_descriptor->uemDataRatesSupported = ATS[idx];
						ccid_descriptor->uemDataRateCurrent = 0x00;
						ccid_descriptor->uemDataRateNew = 0x00;
						memmove(&ccid_descriptor->uemReaderCapabilitiesDataRateDefault, DRS, 4);
						for (i = 0; i < 3; i++) { // parse speeds
							if (ATS[idx] & (0x01 << i)) {
								memmove(&ccid_descriptor->uemReaderCapabilitiesDataRateMax, &DRS[i+1], 4);
							}
						}
						idx++;
					}
					if (ATS[1] & 0x20) { // have TB1
						ccid_descriptor->uemCardCapabilitiesN = 
							(256 * 16) / 128 << ((ATS[idx] & 0x0F) - 1); // SFGT in 128/fc
						ccid_descriptor->uemTimeOut = 33 << (((ATS[idx] >> 4) & 0x0F)); // FWT-> BWT
						ccid_descriptor->uemCardCapabilitiesT1BWI = (BYTE)ccid_descriptor->uemTimeOut;
						idx++;
					}
					if (ATS[1] & 0x40) { // have TC1
						if (ATS[idx] & 0x02) // CID support
							;
						idx++;
					}
					
					if (ccid_descriptor->uemCardType != CT_MIFARE_DESFIRE)
					{
						/*
						// Store historical bytes
						ATR_stub[0] = 0x3B;
						ATR_stub[1] = (UCHAR) (OutputBuffer[3] + 3 - idx);
						memmove(ATR_stub+2, OutputBuffer+idx, ATR_stub[1]);
						i = ATR_stub[1] + 2;
						memmove(ccid_descriptor->uemCardCapabilitiesATRBuffer, ATR_stub, i);
						ccid_descriptor->uemCardCapabilitiesATRLength = (BYTE)i;
						*/
						
						WORD wIndex;
						BYTE bLength = 0;
						BYTE bOffset = 0;
						BYTE bAtsCopyLen = 0;

						/* Len + Format are always present */
						//bOffset = 1 + idx;
						bOffset = idx;

						if (ATS[0] - bOffset >= 0) {
							bAtsCopyLen = (ATS[0] - bOffset);
						}
						else {
							bAtsCopyLen = 0;
						}
						bLength = (5 + bAtsCopyLen);
						ATR_stub[0] = 0x3B;
						ATR_stub[1] = (BYTE)(0x80 | bAtsCopyLen);
						ATR_stub[2] = 0x80;
						ATR_stub[3] = 0x01;
						/*Copy ATS bytes to global ATR array*/
						memcpy(&ATR_stub[4], &ATS[bOffset], bAtsCopyLen);

						ATR_stub[(0x04 + bAtsCopyLen)] = 0x00;
						/*copy ATR validation code*/
						for (wIndex = 1; wIndex < (0x04 + bAtsCopyLen); wIndex++)
						{
							ATR_stub[(0x04 + bAtsCopyLen)] =
								(BYTE)(ATR_stub[(0x04 + bAtsCopyLen)] ^ ATR_stub[wIndex]);
						}
						
#ifndef UEM_USE_EXTENDED_ACTIVATION
						return_value = UEM_CmdSetPPS_A(reader_index);
#endif						

						*pATRLength = bLength;//2 + ATR_stub[1];
						memmove(
							pATR,
							ATR_stub,
							*pATRLength
						);
						
						ccid_descriptor->uemATRLength = *pATRLength;
						memcpy(ccid_descriptor->uemATR, pATR, ccid_descriptor->uemATRLength);
					}
					
					ccid_descriptor->uemfModeTCL = TRUE;
					ccid_descriptor->uembnBit = 0;
					ccid_descriptor->uemReaderCapabilitiesCurrentState = SCARD_NEGOTIABLE;
				} else { // GetPPS failed
					/*cmd[0] = 0x1D; // Halt card
					cmd_length = 1;
					return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						OutputBuffer, &length);
					if (return_value != IFD_SUCCESS)
						return return_value;*/
				}
				/*DeviceExtension->doStopChecker = FALSE;*/
			}
			break;

		case CS_ISO14443B:
			if (ccid_descriptor->uemfSupportsTCL == TRUE)
			{
				cmd[0] = 0xFF;
				
#ifdef UEM_USE_EXTENDED_ACTIVATION
				cmd[0] = 0x7C;
				cmd[1] = 0x00;
				cmd[2] = 0x00;
				memcpy(&cmd[3], &ccid_descriptor->uemATQBLength, 2);
				memcpy(&cmd[5], ccid_descriptor->uemATQB, ccid_descriptor->uemATQBLength);
				cmd_length = 5 + ccid_descriptor->uemATQBLength;
					
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					OutputBuffer, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("return_value = %ul", return_value);
					return return_value;
				}

				if (OutputBuffer[0] == 0x00) {
					if (ccid_descriptor->uemATRLength > 0) {
						*pATRLength = ccid_descriptor->uemATRLength;
						memcpy(pATR, ccid_descriptor->uemATR, ccid_descriptor->uemATRLength);
					}
				}
#else
				cmd[0] = 0x54;
				cmd[1] = 0x08;
				memcpy(&cmd[2], ccid_descriptor->uemCardUID, 4);
				cmd[6] = 0x00;
				cmd[7] = UEM_READER_IFSD;//0x05; // 64 bytes buffer
				cmd[8] = 0x01;
				cmd[9] = 0x00;
				cmd_length = 10;

				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					OutputBuffer, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("return_value = %ul", return_value);
					return return_value;
				}
				
				if (OutputBuffer[0] == 0x00 && length > 1 && ccid_descriptor->uemATRLength > 0) {
					BYTE MBLI = OutputBuffer[1] >> 4;
					
					*pATRLength = ccid_descriptor->uemATRLength;
					memcpy(pATR, ccid_descriptor->uemATR, *pATRLength);
					
					pATR[11] = 0x00 | (MBLI << 4);
					pATR[12] = 0x00;

					for (WORD wIndex = 1; wIndex < 12; wIndex++) {
						pATR[12] ^= pATR[wIndex];
					}

					memcpy(ccid_descriptor->uemATR, pATR, ccid_descriptor->uemATRLength);
				}
				
#endif
				if (OutputBuffer[0] == 0x00) {
					ccid_descriptor->uemfModeTCL = TRUE;
					ccid_descriptor->uembnBit = 0;
					ccid_descriptor->uemReaderCapabilitiesCurrentState = SCARD_NEGOTIABLE;
				}
				
				/*DeviceExtension->SmartcardExtension.ReaderCapabilities.CurrentState = SCARD_NEGOTIABLE;
				
				for (i = 0; i < ICC_RETRIES; i++) {
					return_value = CRD_SetPPS_B_Get_ATR(DeviceExtension, pATR, pATRLength);
					if (return_value == IFD_SUCCESS) {
						DeviceExtension->doStopChecker = FALSE;
						break;
					} else
						doDelay(100);
				}
				
				if (i == ICC_RETRIES)
					DeviceExtension->SmartcardExtension.ReaderCapabilities.CurrentState = SCARD_ABSENT;*/
			}
			else
			{	
				*pATRLength = sizeof(ATR_Simple_T0T1);
				memmove(pATR, ATR_Simple_T0T1, *pATRLength);

				ccid_descriptor->uemATRLength = *pATRLength;
				memcpy(ccid_descriptor->uemATR, pATR, ccid_descriptor->uemATRLength);
				
				/*DeviceExtension->SmartcardExtension.ReaderCapabilities.CurrentState = SCARD_PRESENT;*/
			}
			
			break;
			
		case CS_ISO15693:
			//ccid_descriptor->uemCardType = CT_ISO15693;
			//ccid_descriptor->uemCardType = CT_UNKNOWN;
			
			return_value = IFD_SUCCESS;
			
			switch (ccid_descriptor->uemCardUID[6])
			{
			case 0x04:
				ccid_descriptor->uemCardType = CT_ICODE_NXP;
				*pATRLength = sizeof(ATR_ICode_NXP);
				memmove(
					pATR,
					ATR_ICode_NXP,
					*pATRLength
				);
				break;
			case 0x05:
				ccid_descriptor->uemCardType = CT_ICODE_INFINEON;
				*pATRLength = sizeof(ATR_ICode_Infineon);
				memmove(
					pATR,
					ATR_ICode_Infineon,
					*pATRLength
				);
				break;
			case 0x07:
				ccid_descriptor->uemCardType = CT_ISO15693_TEXASINSTRUMENTS;
				*pATRLength = sizeof(ATR_ISO15693_TexasInstrument);
				memmove(
					pATR,
					ATR_ISO15693_TexasInstrument,
					*pATRLength
				);
				break;
			case 0x16:
				ccid_descriptor->uemCardType = CT_ISO15693_EMMICROELECTRONICMARINSA;
				*pATRLength = sizeof(ATR_ISO15693_EMMicroelectronicMarinSA);
				memmove(
					pATR,
					ATR_ISO15693_EMMicroelectronicMarinSA,
					*pATRLength
				);
				break;
			default:
				*pATRLength = 0;
			}
			
			ccid_descriptor->uemATRLength = *pATRLength;
			memcpy(ccid_descriptor->uemATR, pATR, ccid_descriptor->uemATRLength);
			
			break;
			
		default:
			*pATRLength = 0;
		}
	}
	else
	{
		BYTE SAMATR[] = {0x3B, 0x80, 0x80, 0x01, 0x01};//{0x3B, 0xDF, 0x18, 0xFF, 0x81, 0xF1, 0xFE, 0x43, 0x00, 0x3F, 0x03, 0x83, 0x4D, 0x49, 0x46, 0x41, 0x52, 0x45, 0x20, 0x50, 0x6C, 0x75, 0x73, 0x20, 0x53, 0x41, 0x4D, 0x3B };

		// Power off SAM

		ccid_descriptor->uemCardType = CT_SAM;
		
		UCHAR SAMATRLength = sizeof(SAMATR);

		// Set Holder
		cmd[0] = 0xB2;
		cmd[1] = 0x00 + (BYTE)ccid_descriptor->bCurrentSlotIndex; 
		cmd_length = 2;
		
		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			OutputBuffer, &length);
			
		if (return_value != IFD_SUCCESS)
		{
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			DEBUG_CRITICAL2("return_value = %ul", return_value);
			return return_value;
		}
			
		if (OutputBuffer[0] != 0x00)
		{
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
			DEBUG_CRITICAL("SAM Set Holder CCID_ICC_ABSENT");
#endif
			return IFD_SUCCESS;
		}
			
		//if (OutputBuffer[0] == 0x00) {}
		
		// Reset
		cmd[0] = 0xBA;
		cmd[1] = 0x21;
		cmd[2] = 0x43;
		cmd[3] = 0x65;
		cmd[4] = 0x87;
		cmd_length = 5;

		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			OutputBuffer, &length);
			
		if (return_value != IFD_SUCCESS)
		{
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			DEBUG_CRITICAL2("return_value = %ul", return_value);
			return return_value;
		}
		
		if (OutputBuffer[0] != 0x00)
		{
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
			DEBUG_CRITICAL("SAM Set Holder CCID_ICC_ABSENT");
#endif
			return IFD_SUCCESS;
		}
		
		// Set Holder Get ATR	
		cmd[0] = 0xB2;
		cmd[1] = 0x10 + (BYTE)ccid_descriptor->bCurrentSlotIndex;
		cmd_length = 2;

		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			OutputBuffer, &length);
		//if (return_value != IFD_SUCCESS)
		//	return return_value;
		
		/*if (return_value != IFD_SUCCESS)
		{
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			DEBUG_CRITICAL2("return_value = %ul", return_value);
			return return_value;
		}*/
		
		if ((length > 0)&&(OutputBuffer[0] == 0x00))
		{
			*pATRLength = OutputBuffer[1] + OutputBuffer[2] * 256;
			memmove(pATR, &OutputBuffer[3], *pATRLength);
		}
		else if ((length > 0)&&(OutputBuffer[0] == 0xC4)) // Ver <= 0x4F
		{
			*pATRLength = SAMATRLength;
			memmove(pATR, SAMATR, *pATRLength);
		}
		else
		{
			if (length == 0)
			{
				ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
				DEBUG_CRITICAL("SCARD_F_COMM_ERROR");
				return SCARD_F_COMM_ERROR;
			}
			if (OutputBuffer[0] != 0x00)
			{
				ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
				DEBUG_CRITICAL("SAM Set Holder CCID_ICC_ABSENT");
#endif
				return IFD_SUCCESS;
			}
		}
		//DEBUG_CRITICAL("SAM OK!");
	}

	/*res = WritePort(reader_index, cmdLen, cmd);
	CHECK_STATUS(res)

	// reset available buffer size
	// needed if we go back after a switch to ISO mode
	*nlength = length;

	res = ReadPort(reader_index, nlength, buffer);
	CHECK_STATUS(res)

	if (buffer[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(buffer[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}*/

	/* extract the ATR */
	//*pATRLength -= UEM_HEADER_SIZE;
	//memmove(pATR, pATR + UEM_HEADER_SIZE, *pATRLength);

#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	DEBUG_CRITICAL2("return_value = %ul", return_value);
#endif
	return return_value;
}

/*RESPONSECODE UEM_CmdPowerOn1(unsigned int reader_index, unsigned int *nlength,
	//out 
	unsigned char buffer[], 
	int voltage)
{
	unsigned char cmd[6];
	unsigned int cmdLen;
	status_t res;
	unsigned int length;
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	unsigned char option = UEM_OPTION_NONE;

	// store length of buffer[]
	length = *nlength;

	DEBUG_CRITICAL("UEM_CmdPowerOn");

	if (ccid_descriptor->bCurrentSlotIndex == 0)
	{
		// Set option
		if ((ccid_descriptor->cardType == UEM_CARD_TYPE_MCU_AUTO) ||
			(ccid_descriptor->cardType == UEM_CARD_TYPE_MCU_T0) ||
			(ccid_descriptor->cardType == UEM_CARD_TYPE_MCU_T1))
		{
			option = UEM_OPTION_NONE;
		}
		else
		{
			option = UEM_OPTION_MEMCARD_MODE;
		}
		//(void)UEM_CmdSetOption(reader_index, option);

		// Select card type
		//(void)UEM_CmdSelectCardType(reader_index, ccid_descriptor->cardType);
	}

	if (ccid_descriptor->bCurrentSlotIndex == 0)
	{
		// RESET_WITH_SPECIFIC_VOLTAGE
		cmd[0] = 0x01;
		cmd[1] = 0x80;
		cmd[2] = 0x00;
		cmd[3] = 0x01;
		cmd[4] = ccid_descriptor->cardVoltage;
		cmdLen = 5;
	}
	else
	{
		// Power off SAM
		(void)UEM_CmdPowerOff(reader_index);
		usleep(10 * 1000);

		// RESET_WITH_5_VOLTS_DEFAULT_SAM
		cmd[0] = 0x01;
		cmd[1] = 0x90;
		cmd[2] = 0x00;
		cmd[3] = 0x00;
		cmdLen = 4;
	}

	res = WritePort(reader_index, cmdLen, cmd);
	CHECK_STATUS(res)

	// reset available buffer size
	// needed if we go back after a switch to ISO mode
	*nlength = length;

	res = ReadPort(reader_index, nlength, buffer);
	CHECK_STATUS(res)

	if (buffer[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(buffer[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}

	// If memory card had been selected
	if ((ccid_descriptor->bCurrentSlotIndex == 0) &&
		(option == UEM_OPTION_MEMCARD_MODE))
	{
		unsigned char buffer_tmp[16];
		unsigned int bufferLen = sizeof(buffer_tmp);

		// SELECT_CARD_TYPE
		cmd[0] = 0xFF;
		cmd[1] = 0xA4;
		cmd[2] = 0x00;
		cmd[3] = 0x00;
		cmd[4] = 0x01;
		cmd[5] = ccid_descriptor->cardType;
		cmdLen = 6;

		(void)UEM_CmdXfrBlockTPDU_T0(reader_index, cmdLen, cmd, &bufferLen, buffer_tmp);
	}

	// extract the ATR
	*nlength -= UEM_HEADER_SIZE;
	memmove(buffer, buffer + UEM_HEADER_SIZE, *nlength);

	return return_value;
}*/

RESPONSECODE UEM_CmdPowerOff(unsigned int reader_index)
{
	BYTE InputBuffer[256];
	unsigned int InputBufferLength;
	status_t res;
	ULONG ulBytesReadOrWritten;
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE OutputBuffer[256];
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	BYTE fCardDetected = FALSE;
	ULONG i;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CmdPowerOff");
#endif
	
	if (ccid_descriptor->uemfSessionActive == TRUE)
		return return_value;
		
	switch (ccid_descriptor->uemCardStandard)
	{
		case CS_ISO14443A:
			if (ccid_descriptor->uemfModeTCL == TRUE)
			{
#ifndef UEM_USE_EXTENDED_ACTIVATION
				// DESELECT
				InputBuffer[0] = 0x48;
				InputBuffer[1] = 0x01; 
				InputBuffer[2] = 0x00;
				//InputBuffer[3] = 0xBA;// | ccid_descriptor->uembnBit; 
				//ccid_descriptor->uembnBit ^= 0x01;
				InputBuffer[3] = 0xC2;//0xCA;
				//InputBuffer[4] = ccid_descriptor->uemCID & 0x0F;
				//InputBuffer[5] = 0xC2;
				InputBuffer[4] = TA & 0xFF;//0x90;// 
				InputBuffer[5] = (TA >> 8) & 0xFF;//0x1A;// 
				InputBuffer[6] = 0; 
				InputBuffer[7] = 0;
				InputBufferLength = 8;
				
				return_value = UEM_Transcieve(reader_index, 
							InputBuffer, InputBufferLength,
							OutputBuffer, &ulBytesReadOrWritten);

				ccid_descriptor->uemfModeTCL = FALSE;
				ccid_descriptor->uembnBit = 0;
				
				if (
					(return_value != IFD_SUCCESS) || 
					(ulBytesReadOrWritten == 0) || 
					(OutputBuffer[0] != 0x00)
					)
				{
					DEBUG_CRITICAL("DESELECT failed!");
					//fCardDetected = FALSE;
					//WdfWaitLockRelease( deviceExtension->UEM_TranscieveHandleLock );
					break;
				}
#endif
			}
		default:
		;
	}
	
	ccid_descriptor->uemfPowerOff = TRUE;
	
	return UEM_CmdGetSlotStatus(reader_index, OutputBuffer);
	
	/*
	// POWER_OFF or POWER_OFF_SAM
	cmd[0] = 0x20;
	//cmd[1] = ccid_descriptor->bCurrentSlotIndex == 0 ? 0x81 : 0x91;
	cmd[1] = 0x00;
	cmd[2] = 0x00;

	return_value = UEM_Transcieve(reader_index, 
		cmd, sizeof(cmd),
		buffer_tmp, &length);
	if (return_value != IFD_SUCCESS)
		return return_value;
			
	if (buffer_tmp[UEM_STATUS_OFFSET] != 0x00)
	{
		uem_error(buffer_tmp[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return_value = IFD_COMMUNICATION_ERROR;
	}
	*/
	
	switch (ccid_descriptor->uemCardStandard)
	{
		case CS_ISO14443A:
			if (ccid_descriptor->uemfSupportsTCL == TRUE)
			{
				/*InputBuffer[0] = CMD_CARD;
				InputBuffer[1] = LOBYTE(2); 
				InputBuffer[2] = HIBYTE(2);
				InputBuffer[3] = 0xBA | (deviceExtension->bnBit ^ 1); 
				InputBuffer[4] = (UCHAR)deviceExtension->CID & 0x0F;
				InputBuffer[5] = LOBYTE(TA); 
				InputBuffer[6] = HIBYTE(TA); 
				InputBuffer[7] = 0; 
				InputBuffer[8] = 0;
				InputBufferLength = 9;
				
				status = UEM_Transcieve(
					reader_index, 
					InputBuffer, 
					InputBufferLength, 
					OutputBuffer, 
					&ulBytesReadOrWritten
				);

				if (
					(!NT_SUCCESS(status)) || 
					(ulBytesReadOrWritten == 0) || 
					(OutputBuffer[0] != 0x00)
					)
				{
					fCardDetected = FALSE;
					WdfWaitLockRelease( deviceExtension->UEM_TranscieveHandleLock );
					break;
				}*/
			}
			else
			{
				/*if (deviceExtension->fAuthenticated == TRUE)
				{	// repeat last auth command
			
					for (i = 0; i < 3; i++)
					{
						status = UEM_Transcieve(
							reader_index,
							deviceExtension->LastAuthCommand,
							deviceExtension->LastAuthCommandLength,
							OutputBuffer,
							&ulBytesReadOrWritten
						);
						if (
							(NT_SUCCESS(status)) && 
							(ulBytesReadOrWritten > 0) && 
							(OutputBuffer[0] == 0x00)
							)
						{
							break;
						}
					}
					
					if (
						(!NT_SUCCESS(status)) || 
						(ulBytesReadOrWritten == 0) || 
						(OutputBuffer[0] != 0x00)
						)
					{
						deviceExtension->fAuthenticated = FALSE;
						fCardDetected = FALSE;
					}
				}
				else
				{*/
					//DEBUG_CRITICAL("Halt");
					InputBuffer[0] = 0x1D;// Halt
					InputBufferLength = 1;

					for (i = 0; i < 4; i++)
					{
						return_value = UEM_Transcieve(reader_index, 
							InputBuffer, InputBufferLength,
							OutputBuffer, &ulBytesReadOrWritten);
						if (return_value != IFD_SUCCESS)
							return return_value;
						
						if (
							(return_value == IFD_SUCCESS) && 
							(ulBytesReadOrWritten > 0) && 
							(OutputBuffer[0] == 0x00)
							)
						{
							break;
						}
					}
					
					if (
						(return_value != IFD_SUCCESS) || 
						(ulBytesReadOrWritten == 0) || 
						(OutputBuffer[0] != 0x00)
						)
					{
						fCardDetected = FALSE;
						break;
					}
					
					usleep(100 * 1000);
					
					//DEBUG_CRITICAL("Wakeup");
					InputBuffer[0] = 0x44;// Wakeup
					InputBuffer[1] = ccid_descriptor->uemCardUIDLength;
					memmove(
						&InputBuffer[2],
						ccid_descriptor->uemCardUID,
						ccid_descriptor->uemCardUIDLength
					);
					InputBufferLength = 2 + ccid_descriptor->uemCardUIDLength;

					return_value = UEM_Transcieve(reader_index, 
						InputBuffer, InputBufferLength,
						OutputBuffer, &ulBytesReadOrWritten);
					if (return_value != IFD_SUCCESS)
						return return_value;
					
					if (
						(return_value != IFD_SUCCESS) || 
						(ulBytesReadOrWritten == 0) || 
						(OutputBuffer[0] != 0x00)
						)
					{
						fCardDetected = FALSE;
					}
				/*}*/
			}
			
			break;
		case CS_ISO14443B:
			if (ccid_descriptor->uemfSupportsTCL == TRUE)
			{
				/*InputBuffer[0] = CMD_CARD;
				InputBuffer[1] = LOBYTE(7); 
				InputBuffer[2] = HIBYTE(7);
				InputBuffer[3] = 0x0A | (deviceExtension->bnBit); 
				InputBuffer[4] = (UCHAR)deviceExtension->CID & 0x0F;
				InputBuffer[5] = 0x00; 
				InputBuffer[6] = 0xB0; 
				InputBuffer[7] = 0x81; 
				InputBuffer[8] = 0x00; 
				InputBuffer[9] = 0x01; 
				InputBuffer[10] = LOBYTE(TA); 
				InputBuffer[11] = HIBYTE(TA); 
				InputBuffer[12] = 0x00; 
				InputBuffer[13] = 0x00;
				InputBufferLength = 14;

				UEM_Transcieve(
					reader_index, 
					InputBuffer, 
					InputBufferLength, 
					OutputBuffer, 
					&ulBytesReadOrWritten
				);
				
				if (
					(!NT_SUCCESS(status)) || 
					(ulBytesReadOrWritten == 0) || 
					(OutputBuffer[0] != 0x00)
					)
				{
					fCardDetected = FALSE;
					WdfWaitLockRelease( deviceExtension->UEM_TranscieveHandleLock );
					break;
				}

				deviceExtension->bnBit ^= 1;*/
			}
			else
			{
				InputBuffer[0] = 0x55;// Halt
				memmove(
					&InputBuffer[1],
					ccid_descriptor->uemCardUID,
					ccid_descriptor->uemCardUIDLength
				);
				InputBufferLength = 5;

				return_value = UEM_Transcieve(reader_index, 
					InputBuffer, InputBufferLength,
					OutputBuffer, &ulBytesReadOrWritten);
				if (return_value != IFD_SUCCESS)
					return return_value;
				
				if (
					(return_value != IFD_SUCCESS) || 
					(ulBytesReadOrWritten == 0) || 
					(OutputBuffer[0] != 0x00)
					)
				{
					fCardDetected = FALSE;
					break;
				}

				usleep(100 * 1000);
				
				InputBuffer[0] = 0x57;// Wakeup
				InputBuffer[1] = 0x00;//AFI
				InputBuffer[2] = 0x00;//SNI
				InputBufferLength = 3;

				return_value = UEM_Transcieve(reader_index, 
					InputBuffer, InputBufferLength,
					OutputBuffer, &ulBytesReadOrWritten);
				if (return_value != IFD_SUCCESS)
					return return_value;
				
				if (
					(return_value != IFD_SUCCESS) || 
					(ulBytesReadOrWritten == 0) || 
					(OutputBuffer[0] != 0x00)
					)
				{
					fCardDetected = FALSE;
				}
			}
			
			break;
		case CS_ISO15693:
			InputBuffer[0] = 0x31;// Inventory 15693
			InputBuffer[1] = 0x26;
			InputBuffer[2] = 0x01;
			InputBuffer[3] = 0x00;
			InputBuffer[4] = 0x00;
			InputBufferLength = 5;

			return_value = UEM_Transcieve(reader_index, 
				InputBuffer, InputBufferLength,
				OutputBuffer, &ulBytesReadOrWritten);
			if (return_value != IFD_SUCCESS)
				return return_value;
			
			fCardDetected = FALSE;
			
			if (
				(return_value == IFD_SUCCESS) && 
				(ulBytesReadOrWritten > 0) && 
				(OutputBuffer[0] == 0x00)
				)
			{
				if (OutputBuffer[1] > 0) // RespLen > 0
				{
					if (OutputBuffer[3] == 0) // RetCode
					{
						if (OutputBuffer[4] == 10) // UID Size
						{
							//OutputBuffer[5] // Flag
							//OutputBuffer[6] // DSFID
							if (memcmp(
								&OutputBuffer[7], 
								ccid_descriptor->uemCardUID, 
								ccid_descriptor->uemCardUIDLength) == 0//ccid_descriptor->uemCardUIDLength
								)
							{
								fCardDetected = TRUE;
							}
						}
					}
				}
			}
			
			break;
		default:
		;
	}


	return return_value;
}

void UEM_SuspendCards(unsigned int reader_index, BYTE ucFlagsSuspend)
{
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	ucFlagsSuspend = ucFlagsSuspend & ccid_descriptor->uemFlagsDetected;
	
	if (ccid_descriptor->uemFlagsSuspend == ucFlagsSuspend)
		return;

	ccid_descriptor->uemFlagsSuspend = ucFlagsSuspend;
	
	if (ucFlagsSuspend == 0x00)
		return;

	while (ccid_descriptor->uemFlagsSuspended != ccid_descriptor->uemFlagsSuspend)
	{
		//SysDelay(10);
		usleep(10*1000);
	}
}

void UEM_CalculateFCS( unsigned short * pFCS, BYTE b )
{
    BYTE i;
    unsigned short w;

    w = ( b ^ (*pFCS) ) & 0xFF;
    i = 8;
    do{
        if( w & 1 ) 
        {
            w >>= 1;
            w ^= 0x8408;
        } 
        else 
        {
            w >>= 1;
        }
    } while( --i );
    *pFCS = w ^ ( (*pFCS) >> 8 );
}

void UEM_ByteStuff( BYTE TX_buf[], unsigned int * pTX_cnt, BYTE b )
{
    if( b < 0xFD )
        TX_buf[(*pTX_cnt)++] = b;
    else                            // ByteStuffing
    {      
        TX_buf[(*pTX_cnt)++] = 0xFF;
        TX_buf[(*pTX_cnt)++] = 0xFF - b;
    }
}

void UEM_PostByte( BYTE TX_buf[], unsigned int * pTX_cnt, unsigned short * pTX_FCS, BYTE b )
{
    UEM_CalculateFCS( pTX_FCS, b );
    UEM_ByteStuff( TX_buf, pTX_cnt, b );
}

RESPONSECODE UEM_DirectIOCard(unsigned int reader_index,
	/*@out@*/ BYTE out_buffer[], unsigned int out_length,
	ULONG timeout,
	/*@in@*/ BYTE in_buffer[],
	ULONG * in_length
	)
{
	const int BUF_SIZE = 1024;// 300 * 10;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength;
	DWORD RecvLength;
	
	SendBuffer[0] = 0x48;
	SendBuffer[1] = out_length & 0x000000FF;
	SendBuffer[2] = (out_length & 0x0000FF00) >> 8;
	memcpy(&SendBuffer[3], out_buffer, out_length);
	memcpy(&SendBuffer[3 + out_length], &timeout, 4);
	SendLength = 3 + out_length + 4;
	
	RecvLength = sizeof(RecvBuffer);
	RESPONSECODE Status = UEM_Transcieve(reader_index, 
		SendBuffer, SendLength,
		RecvBuffer, 
		&RecvLength);
	
	if (Status != SCARD_S_SUCCESS) {
		*in_length = 0;
		DEBUG_CRITICAL2("error=%lu", Status);
		return Status;
	}
	
	if (RecvLength < 1) {
		*in_length = 0;
		DEBUG_CRITICAL("SCARD_F_COMM_ERROR");
		return SCARD_F_COMM_ERROR;
	}
	
	if (RecvBuffer[0] != 0x00) {
		*in_length = 0;
		DEBUG_CRITICAL2("SCARD_F_INTERNAL_ERROR 0x%02X", RecvBuffer[0]);
		return SCARD_F_INTERNAL_ERROR;
	}
	
	if (RecvLength < 3) {
		*in_length = 0;
		DEBUG_CRITICAL("SCARD_F_COMM_ERROR");
		return SCARD_F_COMM_ERROR;
	}
	
	DWORD recLen = 0;
	memcpy(&recLen, &RecvBuffer[1], 2);
	
	if (recLen == 0) {
		*in_length = 0;
		DEBUG_CRITICAL("SCARD_F_COMM_ERROR");
		return SCARD_F_COMM_ERROR;
	}
	
	if (recLen > *in_length) {
		*in_length = 0;
		DEBUG_CRITICAL("SCARD_E_INSUFFICIENT_BUFFER");
		return SCARD_E_INSUFFICIENT_BUFFER;
	}
	
	*in_length = recLen;
	
	memcpy(in_buffer, &RecvBuffer[3], *in_length);
	
/*#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	{
		int resp_log_pos = 0;
		char resp_log[SCR_BUFFER_SIZE * 3 + 1] = {0};
		for( unsigned int jj = 0; jj < *in_length; jj++ ) {
			//DEBUG_CRITICAL2("Response: %02X", in_buffer[jj]);
			resp_log_pos += sprintf(resp_log + resp_log_pos, "%02X ", in_buffer[jj]);
		}
		DEBUG_CRITICAL2("UEM_DirectIOCard Response: %s", resp_log);
	}	
#endif*/

	return Status;
}

BYTE Cmd_ID;

RESPONSECODE UEM_Transcieve(unsigned int reader_index,
	/*@out@*/ BYTE out_buffer[], unsigned int out_length,
	/*@in@*/ BYTE in_buffer[],
	ULONG * in_length
	)
{
	status_t res;
	//BYTE TX_buf[256];
	BYTE TX_buf[SCR_BUFFER_SIZE];
	unsigned int TX_cnt = 0;
    unsigned short TX_FCS = 0xFFFF;

	//BYTE RX_buf[256];
	BYTE RX_buf[SCR_ALIGNED_BUFFER_SIZE];
	unsigned int RX_cnt = 0;
    unsigned short RX_FCS = 0xFFFF;
	
	//unsigned int in_buffer_length = sizeof(in_buffer);
	unsigned int in_buffer_length = SCR_BUFFER_SIZE;
	
	ULONG i, j, total, rest, result;
	
    TX_buf[TX_cnt++] = 0xFD;  // Frame begin
    UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, 0x00 );
    UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, Cmd_ID++ );
    
    for( i = 0; i < out_length; i++ ) {
		UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, out_buffer[i] );
		//DEBUG_CRITICAL2("Command: %02X", out_buffer[i] );
	}
	
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	{
		int cmd_log_pos = 0;
		char cmd_log[SCR_BUFFER_SIZE * 3 + 1] = {0};
		for( unsigned int jj = 0; jj < out_length; jj++ ) {
			cmd_log_pos += sprintf(cmd_log + cmd_log_pos, "%02X ", out_buffer[jj]);
		}
		DEBUG_CRITICAL2("Command: %s", cmd_log);
	}	
#endif
	
    TX_FCS ^= 0xFFFF;
    UEM_ByteStuff( TX_buf, &TX_cnt, TX_FCS & 0xFF );
    UEM_ByteStuff( TX_buf, &TX_cnt, ( TX_FCS >> 8 ) & 0xFF );
    TX_buf[TX_cnt++] = 0xFE;  // Frame end
	                                
	//for( unsigned int j = 0; j < TX_cnt; j++ )	                                
	//	DEBUG_CRITICAL3("WritePort %d: %02X", reader_index, TX_buf[j]);
	                                
	res = WritePort(reader_index, TX_cnt, TX_buf);
	//DEBUG_CRITICAL2("WritePort RES: %d", res);
	if (res == 0x000000FA)
		res = STATUS_SUCCESS;
	if (res != STATUS_SUCCESS)
		return IFD_COMMUNICATION_ERROR;
	//CHECK_STATUS(res)
	
	RX_cnt = sizeof(RX_buf);
	res = ReadPort(reader_index, &RX_cnt, RX_buf);
	//DEBUG_CRITICAL2("ReadPort RES: %d", res);
	if (res == 0x000000FA)
		res = STATUS_SUCCESS;
	if (res != STATUS_SUCCESS)
		return IFD_COMMUNICATION_ERROR;
	//CHECK_STATUS(res)
	
	//for( unsigned int jjj = 0; jjj < RX_cnt; jjj++ )	                                
	//	DEBUG_CRITICAL2("ReadPort: %02X", RX_buf[jjj]);
	
	//DEBUG_CRITICAL2("Data received: %d bytes", RX_cnt);

	if (RX_cnt < 8)
	{
		DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_buf[0] != 0xFD)
	{
		DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_buf[RX_cnt-1] != 0xFE)
	{
		DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_cnt - 7 > in_buffer_length)
	{
		DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	
	
	
	
	
	
	
// Start, [ Addr, FID, Cmd, Status, <Response>, 2*FSC ], Stop
    for (i = 1, j = 0; i < RX_cnt - 1; i++) {
        if (RX_buf[i] == 0xFF)
            RX_buf[j++] = 0xFF - RX_buf[++i];
        else
            RX_buf[j++] = RX_buf[i];
    }

    // Addr, FID, Cmd, Status, <Response>, 2*FSC
    if (j < 6)
    {
		DEBUG_CRITICAL("STATUS_DEVICE_PROTOCOL_ERROR j < 6");
        return IFD_COMMUNICATION_ERROR;//STATUS_DEVICE_PROTOCOL_ERROR;
	}

    // check FID
    //if (buff[1] != context->FID) {
    //    DevPrint(DBG_WARN, "Got FID %d, expected %d\n", buff[1], context->FID);
    //    return STATUS_DEVICE_PROTOCOL_ERROR;
    //}

    // calculate and check FSC
    for (i = 0; i < j - 2; i++)
		UEM_CalculateFCS(&RX_FCS, RX_buf[i]);
    RX_FCS ^= 0xFFFF;

    if (RX_FCS != (RX_buf[i] | (RX_buf[i+1] << 8))) {
        //DevPrint(DBG_WARN, "Transfer FSC mismatch\n");
        DEBUG_CRITICAL("STATUS_CRC_ERROR Transfer FSC mismatch");
        return IFD_COMMUNICATION_ERROR;//STATUS_CRC_ERROR;
    }

    // Addr, FID, Cmd, [ Status, <Response>, ] 2*FSC
    result = j - 5;
    if (result > in_buffer_length) {
        //result = in_buffer_length;
        DEBUG_CRITICAL("STATUS_BUFFER_OVERFLOW result > in_buffer_length");
        return IFD_COMMUNICATION_ERROR;//STATUS_BUFFER_OVERFLOW;
    }
	
	//RtlCopyMemory(bin, buff + 3, result);
    //*rlin = result;
	
	
	
	
	
	
	
	//DEBUG_CRITICAL2("result = %d", result);
	
	
	*in_length = 0;
	
    //for( unsigned int k = 4; k < RX_cnt - 3; k++ ) {
	for( unsigned int k = 3; k < 3 + result; k++ ) {
		in_buffer[(*in_length)] = RX_buf[k];
		//DEBUG_CRITICAL3("Response: *in_length=%d, %02X", *in_length, in_buffer[(*in_length)]);
		(*in_length)++;
	}
	
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	{
		int resp_log_pos = 0;
		char resp_log[SCR_BUFFER_SIZE * 3 + 1] = {0};
		for( unsigned int jj = 0; jj < *in_length; jj++ ) {
			//DEBUG_CRITICAL2("Response: %02X", in_buffer[jj]);
			resp_log_pos += sprintf(resp_log + resp_log_pos, "%02X ", in_buffer[jj]);
		}
		DEBUG_CRITICAL2("Response: %s", resp_log);
	}	
#endif
	
	return IFD_SUCCESS;
}

RESPONSECODE UEM_TranscieveOld(unsigned int reader_index,
	/*@out@*/ BYTE out_buffer[], unsigned int out_length,
	/*@in@*/ BYTE in_buffer[], unsigned int in_length,
	unsigned int * in_actual_length
	)
{
	status_t res;
	BYTE TX_buf[256];
	unsigned int TX_cnt = 0;
    unsigned short TX_FCS = 0xFFFF;

	BYTE RX_buf[256];
	unsigned int RX_cnt = 0;
	
    TX_buf[TX_cnt++] = 0xFD;  // Frame begin
    UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, 0x00 );
    UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, Cmd_ID++ );
    for( unsigned int i = 0; i < out_length; i++ ) {
		UEM_PostByte( TX_buf, &TX_cnt, &TX_FCS, out_buffer[i] );
		DEBUG_CRITICAL2("Command: %02X", out_buffer[i] );
	}
    TX_FCS ^= 0xFFFF;
    UEM_ByteStuff( TX_buf, &TX_cnt, TX_FCS & 0xFF );
    UEM_ByteStuff( TX_buf, &TX_cnt, ( TX_FCS >> 8 ) & 0xFF );
    TX_buf[TX_cnt++] = 0xFE;  // Frame end
	                                
	for( unsigned int j = 0; j < TX_cnt; j++ )	                                
		DEBUG_CRITICAL3("WritePort %d: %02X", reader_index, TX_buf[j]);
	                                
	res = WritePort(reader_index, TX_cnt, TX_buf);
	DEBUG_CRITICAL2("WritePort RES: %08X", res);
	if (res == 0x000000FA)
		res = STATUS_SUCCESS;
	if (res != STATUS_SUCCESS)
		return IFD_COMMUNICATION_ERROR;
	//CHECK_STATUS(res)
	
	RX_cnt = sizeof(RX_buf);
	res = ReadPort(reader_index, &RX_cnt, RX_buf);
	DEBUG_CRITICAL2("ReadPort RES: %d", res);
	if (res == 0x000000FA)
		res = STATUS_SUCCESS;
	if (res != STATUS_SUCCESS)
		return IFD_COMMUNICATION_ERROR;
	//CHECK_STATUS(res)
	
	for( unsigned int jjj = 0; jjj < RX_cnt; jjj++ )	                                
		DEBUG_CRITICAL2("ReadPort: %02X", RX_buf[jjj]);
	
	DEBUG_CRITICAL2("Data received: %d bytes", RX_cnt);

	if (RX_cnt < 8)
	{
		DEBUG_CRITICAL("UEM_Transcieve: return IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_buf[0] != 0xFD)
	{
		DEBUG_CRITICAL("UEM_Transcieve: return IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_buf[RX_cnt-1] != 0xFE)
	{
		DEBUG_CRITICAL("UEM_Transcieve: return IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	if (RX_cnt - 7 > in_length)
	{
		DEBUG_CRITICAL("UEM_Transcieve: return IFD_COMMUNICATION_ERROR");
		return IFD_COMMUNICATION_ERROR;
	}
	
	*in_actual_length = 0;
	
    for( unsigned int k = 4; k < RX_cnt - 3; k++ ) {
		in_buffer[(*in_actual_length)++] = RX_buf[k];
		//DEBUG_CRITICAL2("Response: %02X", in_buffer[k-4]);
	}
	for( unsigned int jj = 0; jj < *in_actual_length; jj++ )	                                
		DEBUG_CRITICAL2("Response: %02X", in_buffer[jj]);
		
	return IFD_SUCCESS;
}

RESPONSECODE UEM_SignalCardFound(unsigned int reader_index)
{
	BYTE cmd[SCR_BUFFER_SIZE] = {0};
	unsigned int cmd_length;
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE buffer_tmp[SCR_BUFFER_SIZE] = {0};
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	if (ccid_descriptor->uemfPowerOff == TRUE) 
		return return_value;
		
	cmd[0] = 0x07;
	cmd[1] = 0x02; // Blink Color
	cmd[2] = 0x00;//0x01; // Blink Count
	cmd[3] = 0x02; //0x00; // Post Color
	cmd_length = 4;
	return_value = UEM_Transcieve(reader_index, 
		cmd, cmd_length,
		buffer_tmp, &length);
		
	cmd[0] = 0x05;
	cmd[1] = 0x01; // Beep count
	cmd_length = 2;
	return_value = UEM_Transcieve(reader_index, 
		cmd, cmd_length,
		buffer_tmp, &length);
		
	return return_value;
}

RESPONSECODE UEM_SignalCardLost(unsigned int reader_index)
{
	BYTE cmd[SCR_BUFFER_SIZE] = {0};
	unsigned int cmd_length;
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE buffer_tmp[SCR_BUFFER_SIZE] = {0};
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	if (ccid_descriptor->uemfPowerOff == TRUE) 
		return return_value;
		
	cmd[0] = 0x07;
	cmd[1] = 0x01;//0x00; // Blink Color
	cmd[2] = 0x00;//0x01; // Blink Count
	cmd[3] = 0x01;//0x00; // Post Color
	cmd_length = 4;
	return_value = UEM_Transcieve(reader_index, 
		cmd, cmd_length,
		buffer_tmp, &length);
		
	return return_value;
}

RESPONSECODE UEM_CmdGetSlotStatus(unsigned int reader_index,
	/*@out@*/ BYTE buffer[])
{
	BYTE cmd[SCR_BUFFER_SIZE] = {0};
	unsigned int cmd_length;
	status_t res;
	unsigned int lengthOld;
	ULONG length;
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE buffer_tmp[SCR_BUFFER_SIZE] = {0};
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	DEBUG_CRITICAL3("UEM_CmdGetSlotStatus reader_index = %d, ccid_descriptor->bCurrentSlotIndex = %d", reader_index, ccid_descriptor->bCurrentSlotIndex);
#endif
	
	if (ccid_descriptor->uemfSessionActive == TRUE)
	{
		buffer[7] = CCID_ICC_PRESENT_ACTIVE;
		return return_value;
	}
	
	if (ccid_descriptor->bCurrentSlotIndex == 0)
	{
		//DEBUG_CRITICAL("UEM_CmdGetSlotStatus");
		// FD 00 02 05 01 57 26 FE
		//cmd[0] = 0xFD;
		//cmd[1] = 0x00;
		//cmd[2] = 0x02;
		//cmd[3] = 0x05;
		//cmd[4] = 0x01;
		//cmd[5] = 0x57;
		//cmd[6] = 0x26;
		//cmd[7] = 0xFE;

		//cmd[0] = 0x05;
		//cmd[1] = 0x01;

		//cmd[0] = 0x07;
		//cmd[1] = 0x02;
		//cmd[2] = 0x01;
		//cmd[3] = 0x00;
		//cmd_length = 4;
		
		if (ccid_descriptor->uemCardStatus != CCID_ICC_PRESENT_ACTIVE || ccid_descriptor->uemfPowerOff == TRUE)
		{
			ccid_descriptor->uemfModeTCL = FALSE;
			ccid_descriptor->uembnBit = 0;
			
			/*//MFRC ON
			cmd[0] = 0x10;
			cmd_length = 1;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
			//lengthOld = sizeof(buffer_tmp);
			//return_value = UEM_TranscieveOld(reader_index, 
			//	cmd, cmd_length,
			//	buffer_tmp, lengthOld, &lengthOld);
			if (return_value != IFD_SUCCESS)
			{
				//DEBUG_CRITICAL2("UEM_CmdGetSlotStatus: error=%lu", return_value);
				return return_value;
			}*/
			
			/*//RF OFF ON
			cmd[0] = 0x20;
			cmd[1] = 0x0A;
			cmd[2] = 0x00;
			cmd_length = 3;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
			if (return_value != IFD_SUCCESS)
				return return_value;*/
			
			//RF MODE 14443A
			ccid_descriptor->uemCardStandard = CS_ISO14443A;
			
			ccid_descriptor->uemATQBLength = 0;
			ccid_descriptor->uemATRLength = 0;
			ccid_descriptor->uemATSLength = 0;
					
#ifdef UEM_USE_EXTENDED_ACTIVATION
			cmd[0] = 0x75;
			cmd[1] = 0x00;
			cmd[2] = 0xAA;
			cmd[3] = 0x80;// | m_CID & 0x0F;
			cmd_length = 4;

			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);

			if (return_value != IFD_SUCCESS)
			{
				DEBUG_CRITICAL2("error=%lu", return_value);
				ccid_descriptor->uemfPowerOff = FALSE;
				return return_value;
			}
#else
			cmd[0] = 0x51;
			cmd[1] = 0x00;
			cmd_length = 2;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
				
			if (return_value != IFD_SUCCESS)
			{
				DEBUG_CRITICAL2("error=%lu", return_value);
				ccid_descriptor->uemfPowerOff = FALSE;
				return return_value;
			}
			
			//RF OFF ON
			cmd[0] = 0x20;
			cmd[1] = 0x0A;
			cmd[2] = 0x00;
			cmd_length = 3;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
				
			ccid_descriptor->uemfModeTCL = FALSE;
			ccid_descriptor->uembnBit = 0;
			
			if (return_value != IFD_SUCCESS)
			{
				DEBUG_CRITICAL2("error=%lu", return_value);
				ccid_descriptor->uemfPowerOff = FALSE;
				return return_value;
			}
			
			sleep(1);
			
			//INVENTORY 14443A
			cmd[0] = 0x43;
			cmd_length = 1;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
				
			if (return_value != IFD_SUCCESS)
			{
				DEBUG_CRITICAL2("error=%lu", return_value);
				ccid_descriptor->uemfPowerOff = FALSE;
				return return_value;
			}
			
			//res = WritePort(reader_index, sizeof(cmd), cmd);
			//CHECK_STATUS(res)

			//DEBUG_CRITICAL2("WritePort RES: %d", res);
			
			//length = sizeof(buffer_tmp);
			//res = ReadPort(reader_index, &length, buffer_tmp);
			//CHECK_STATUS(res)

			//DEBUG_CRITICAL2("ReadPort RES: %d", res);
			//DEBUG_CRITICAL2("Data received: %d bytes", length);
			
			//if (length < sizeof(buffer_tmp))
			//{
			//	DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
			//	return IFD_COMMUNICATION_ERROR;
			//}
#endif
			if ((buffer_tmp[UEM_STATUS_OFFSET] != 0x00) && (buffer_tmp[UEM_STATUS_OFFSET] != 0xFF))
			{
				uem_error(buffer_tmp[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
				DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
				ccid_descriptor->uemfPowerOff = FALSE;
				return IFD_COMMUNICATION_ERROR;
			}

			if ((buffer_tmp[UEM_STATUS_OFFSET] == 0x00) && (length < 9))
			{
				uem_error(buffer_tmp[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
				DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
				ccid_descriptor->uemfPowerOff = FALSE;
				return IFD_COMMUNICATION_ERROR;
			}

			/* simulate a CCID bStatus */
			switch (buffer_tmp[UEM_STATUS_OFFSET])
			{
				case 0x00:
					ccid_descriptor->uemCardATQ[0] = buffer_tmp[1];
					ccid_descriptor->uemCardATQ[1] = buffer_tmp[2];
					ccid_descriptor->uemCardSAK = buffer_tmp[3];
					ccid_descriptor->uemCardUIDLength = buffer_tmp[4];
					memmove(
						ccid_descriptor->uemCardUID, 
						&buffer_tmp[5], 
						ccid_descriptor->uemCardUIDLength);
						
					ccid_descriptor->uemCardInactiveCounter = UEM_CARD_INACTIVE_MAX_CYCLES;
					
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;
					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					
					return_value = UEM_SignalCardFound(reader_index);
					
					break;

				//case 1:
				//	buffer[7] = CCID_ICC_PRESENT_INACTIVE;
				//	break;

				case 0xFF:
					buffer[7] = CCID_ICC_ABSENT;
					ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
					break;

				default:
					return_value = IFD_COMMUNICATION_ERROR;
					break;
			}
			
			if (ccid_descriptor->uemCardStatus != CCID_ICC_PRESENT_ACTIVE)
			{
				DWORD dwPUPI = 0;
				BOOL fSupportsTCL = FALSE;
				BYTE appData[4] = { 0 };
				BYTE protInfo[4] = { 0 };

				//RF MODE 14443B
				ccid_descriptor->uemCardStandard = CS_ISO14443B;
				
				buffer_tmp[0] == 0xFF;
				
				ccid_descriptor->uemATQBLength = 0;
				ccid_descriptor->uemATRLength = 0;
				ccid_descriptor->uemATSLength = 0;
					
#ifdef UEM_USE_EXTENDED_ACTIVATION
				cmd[0] = 0x75;
				cmd[1] = 0x10;
				cmd[2] = 0xAA;
				cmd[3] = 0x80;// | m_CID & 0x0F;
				cmd[4] = 0x00;
				cmd[5] = 0x00;
				cmd_length = 6;

				return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						buffer_tmp, &length);

				if (return_value != IFD_SUCCESS)
				{
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
			
				if (buffer_tmp[0] == 0x00) {
					if (length < 8) {
						ccid_descriptor->uemfPowerOff = FALSE;
						DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
						return IFD_COMMUNICATION_ERROR;
					}
					
					BYTE MBLI = buffer_tmp[1];
					BYTE ATQLength = buffer_tmp[2];
					BYTE ATQ[256] = { 0 };
					memcpy(ATQ, &buffer_tmp[3], ATQLength);
					if (ATQLength > 8) {
						const int ATRLength = 13;
						BYTE ATR[ATRLength];// = { 0x3B, 0x88, 0x80, 0x01 };
						ATR[0] = 0x3B; ATR[1] = 0x88; ATR[2] = 0x80; ATR[3] = 0x01;
						memcpy(&dwPUPI, &ATQ[1], 4);
						memcpy(appData, &ATQ[5], 4);
						if (ATQLength - 9 <= 4) {
							memcpy(protInfo, &ATQ[9], ATQLength - 9);
						}

						fSupportsTCL = ((protInfo[1] & 0x01) == 0x00) ? FALSE : TRUE;

						//DEBUG_CRITICAL2("fSupportsTCL = %d", fSupportsTCL);

						memcpy(&ATR[4], &ATQ[5], 7);

						/* MBLI */
						ATR[11] = 0x00 | (MBLI << 4);
						ATR[12] = 0x00;

						for (WORD wIndex = 1; wIndex < 12; wIndex++) {
							ATR[12] ^= ATR[wIndex];
						}

						ccid_descriptor->uemATRLength = ATRLength;
						memcpy(ccid_descriptor->uemATR, ATR, ccid_descriptor->uemATRLength);

						ccid_descriptor->uemATQBLength = ATQLength;
						memcpy(ccid_descriptor->uemATQB, ATQ, ccid_descriptor->uemATQBLength);
					}

				}
#else
				cmd[0] = 0x51;
				cmd[1] = 0x10;
				cmd_length = 2;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				if (return_value != IFD_SUCCESS)
				{
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
				
				//RF OFF ON
				cmd[0] = 0x20;
				cmd[1] = 0x0A;
				cmd[2] = 0x00;
				cmd_length = 3;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				ccid_descriptor->uemfModeTCL = FALSE;
				ccid_descriptor->uembnBit = 0;
				
				if (return_value != IFD_SUCCESS)
				{
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
				
				sleep(1);
				
				//INVENTORY 14443B
				cmd[0] = 0x56;
				cmd[1] = 0x00;
				cmd[2] = 0x00;
				cmd_length = 3;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
				
				if (buffer_tmp[0] == 0x00) {
					if (length < 12) {
						ccid_descriptor->uemfPowerOff = FALSE;
						DEBUG_CRITICAL("IFD_COMMUNICATION_ERROR");
						return IFD_COMMUNICATION_ERROR;
					}
					
					memcpy(&dwPUPI, &buffer_tmp[1], 4);
					memcpy(appData, &buffer_tmp[5], 4);
					memcpy(protInfo, &buffer_tmp[9], 3);
					
					const int ATRLength = 13;
					BYTE ATR[ATRLength];// = { 0x3B, 0x88, 0x80, 0x01 };
					ATR[0] = 0x3B; ATR[1] = 0x88; ATR[2] = 0x80; ATR[3] = 0x01;
					
					BYTE MBLI = 0x00; // !!!!!!!!!!!! Cannot get MBLI from Activate Idle B !!!!!!!!!!
					BYTE ATQLength = length;
					BYTE ATQ[256] = { 0 };
					
					memcpy(&ATQ[1], &buffer_tmp[1], length - 1);
					
					fSupportsTCL = ((protInfo[1] & 0x01) == 0x00) ? FALSE : TRUE;
					
					memcpy(&ATR[4], &ATQ[5], 7);
					
					/* MBLI */
					ATR[11] = 0x00 | (MBLI << 4);
					ATR[12] = 0x00;

					for (WORD wIndex = 1; wIndex < 12; wIndex++) {
						ATR[12] ^= ATR[wIndex];
					}

					ccid_descriptor->uemATRLength = ATRLength;
					memcpy(ccid_descriptor->uemATR, ATR, ccid_descriptor->uemATRLength);

					ccid_descriptor->uemATQBLength = ATQLength;
					memcpy(ccid_descriptor->uemATQB, ATQ, ccid_descriptor->uemATQBLength);
				}
#endif
				if (buffer_tmp[0] == 0x00) {
					
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;
					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					
					return_value = UEM_SignalCardFound(reader_index);
					
					ccid_descriptor->uemCardStandard = CS_ISO14443B;
					ccid_descriptor->uemCardUIDLength = 4; // UID Length
					memcpy(ccid_descriptor->uemCardUID, &dwPUPI, 4); //1,2,3,4
					memcpy(ccid_descriptor->uemAppData, appData, 4); //5,6,7,8
					memcpy(ccid_descriptor->uemProtInfo, protInfo, 4); //9,10,11
					ccid_descriptor->uemfSupportsTCL = fSupportsTCL;

					ccid_descriptor->uembnBit = 0;
					ccid_descriptor->uemCID = 0xF00001;
					ccid_descriptor->uemReaderCapabilitiesChannel = ccid_descriptor->uemCID & 0xFFFF;
					//ccid_descriptor->uemTimeOut = TA;

					//ccid_descriptor->uemT1IFSC = (protInfo[1] >> 4) & 0x0F; //(OutputBuffer[10] >> 4) & 0x0F;
					//ccid_descriptor->uemT1IFSD = UEM_READER_IFSD;
					ccid_descriptor->uemIFSC = (protInfo[1] >> 4) & 0x0F; //(OutputBuffer[10] >> 4) & 0x0F;
					ccid_descriptor->uemIFSD = UEM_READER_IFSD;

					ccid_descriptor->uemDataRatesSupported = protInfo[0]; //OutputBuffer[9];
					ccid_descriptor->uemDataRateCurrent = 0x10;
					ccid_descriptor->uemDataRateNew = 0x10;
					memcpy(&ccid_descriptor->uemReaderCapabilitiesDataRateDefault, DRS, 4);
					for (int i = 0; i < 3; i++) {
						if (protInfo[0] & (0x01 << i)) { //if (OutputBuffer[9] & (0x01 << i)) {
							memcpy(&ccid_descriptor->uemReaderCapabilitiesDataRateMax, &DRS[i + 1], 4);
						}
					}
					ccid_descriptor->uemCardCapabilitiesN = 0;
					ccid_descriptor->uemTimeOut = 33 << ((protInfo[2] >> 4) & 0x0F); // ((OutputBuffer[11] >> 4) & 0x0F); // FWT-> BWT
					ccid_descriptor->uemCardCapabilitiesT1BWI = (BYTE)ccid_descriptor->uemTimeOut;
				}
				else {
					buffer[7] = CCID_ICC_ABSENT;
					ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
				}
			}
			
			if (ccid_descriptor->uemCardStatus != CCID_ICC_PRESENT_ACTIVE)
			{
				//RF MODE 15693
				ccid_descriptor->uemCardStandard = CS_ISO15693;
				
				cmd[0] = 0x51;
				cmd[1] = 0x40;
				cmd_length = 2;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("error=%lu", return_value);
					//buffer[7] = CCID_ICC_ABSENT;
					//ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
			
				//RF OFF ON
				cmd[0] = 0x20;
				cmd[1] = 0x0A;//m_device->m_performanceSettings.QueryStandards.ucQueryISO15693_ResetTime
				cmd[2] = 0x00;
				cmd_length = 3;
				
				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				ccid_descriptor->uemfModeTCL = FALSE;
				ccid_descriptor->uembnBit = 0;
				
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
				
				usleep(1 * 1000);//m_device->m_performanceSettings.QueryStandards.ucQueryISO15693_FeedTime * 1000)

				//BYTE pbRecvBuffer[256] = { 0 };
				//DWORD dwRecvLength = sizeof(pbRecvBuffer);
				//status = m_device->m_pUemReader->iso15693Inventory(pbRecvBuffer, &dwRecvLength);

				cmd[0] = 0x31;// Inventory 15693
				cmd[1] = 0x26;
				cmd[2] = 0x01;
				cmd[3] = 0x00;
				cmd[4] = 0x00;
				cmd_length = 5;

				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				ccid_descriptor->uemATQBLength = 0;
				ccid_descriptor->uemATRLength = 0;
				ccid_descriptor->uemATSLength = 0;
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("error=%lu", return_value);
					ccid_descriptor->uemfPowerOff = FALSE;
					return return_value;
				}
				
				{
					BOOL fCardDetected = FALSE;
					
					if (
						(length > 0) && 
						(buffer_tmp[0] == 0x00)
					) {
						if (buffer_tmp[1] > 0) { // RespLen > 0
							if (buffer_tmp[3] == 0) { // RetCode
								if (buffer_tmp[4] == 10) { // UID Size
									//OutputBuffer[5] // Flag
									//OutputBuffer[6] // DSFID
									fCardDetected = TRUE;
									ccid_descriptor->uemCardStandard = CS_ISO15693;
									ccid_descriptor->uemCardUIDLength = 8; // UID Length
									memcpy(ccid_descriptor->uemCardUID, &buffer_tmp[7], ccid_descriptor->uemCardUIDLength);

									buffer[7] = CCID_ICC_PRESENT_ACTIVE;
									ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
									
									return_value = UEM_SignalCardFound(reader_index);
									//DEBUG_CRITICAL2("UEM_SignalCardFound: error=%lu", return_value);
								}
							}
						}
					}
					
					if (fCardDetected == FALSE) {
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE					
						DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif						
						return_value = UEM_SignalCardLost(reader_index);
								
						ccid_descriptor->uemfPowerOff = FALSE;
						
						return IFD_SUCCESS;
					}
				}
			}
			
			ccid_descriptor->uemfPowerOff = FALSE;
		}
		else {
			switch (ccid_descriptor->uemCardStandard)
			{
			case CS_ISO14443A:
				if (ccid_descriptor->uemfModeTCL != TRUE) {
					// HALT
					cmd[0] = 0x1D;
					cmd_length = 1;
					
					return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						buffer_tmp, &length);
						
					if (return_value != IFD_SUCCESS) {
						DEBUG_CRITICAL2("error=%lu", return_value);
						return return_value;
					}
						
					if (buffer_tmp[UEM_STATUS_OFFSET] != 0x00) {
						//DEBUG_CRITICAL2("HALT fail: uemCardInactiveCounter=%d", ccid_descriptor->uemCardInactiveCounter);
						ccid_descriptor->uemCardInactiveCounter--;
						if (ccid_descriptor->uemCardInactiveCounter <= 0) {
							buffer[7] = CCID_ICC_ABSENT;
							ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE						
							DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
							return_value = UEM_SignalCardLost(reader_index);
							
							return return_value;
						}
						else {
							//RF OFF ON
							cmd[0] = 0x20;
							cmd[1] = 0x0A;
							cmd[2] = 0x00;
							cmd_length = 3;
							return_value = UEM_Transcieve(reader_index, 
								cmd, cmd_length,
								buffer_tmp, &length);
								
							ccid_descriptor->uemfModeTCL = FALSE;
							ccid_descriptor->uembnBit = 0;
								
							//INVENTORY 14443A
							cmd[0] = 0x43;
							cmd_length = 1;
							return_value = UEM_Transcieve(reader_index, 
								cmd, cmd_length,
								buffer_tmp, &length);
							
							if (ccid_descriptor->uemCardUIDLength != buffer_tmp[4] || 
								memcmp(ccid_descriptor->uemCardUID, &buffer_tmp[5], buffer_tmp[4]) != 0
							) {
								buffer[7] = CCID_ICC_ABSENT;
								ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE						
								DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif							
								return_value = UEM_SignalCardLost(reader_index);
								
								return return_value;
							}
						
							buffer[7] = CCID_ICC_PRESENT_ACTIVE;
							ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
							
							return IFD_SUCCESS;
						}			
					}
						
					usleep(100 * 1000);
					
					//DEBUG_CRITICAL2("Wakeup: uemCardUIDLength=%lu", ccid_descriptor->uemCardUIDLength);
					// Wakeup
					cmd[0] = 0x44;
					cmd[1] = ccid_descriptor->uemCardUIDLength;
					memmove(
						&cmd[2],
						ccid_descriptor->uemCardUID,
						ccid_descriptor->uemCardUIDLength
					);
					cmd_length = 2 + ccid_descriptor->uemCardUIDLength;

					return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						buffer_tmp, &length);
						
					if (return_value != IFD_SUCCESS) {
						DEBUG_CRITICAL2("error=%lu", return_value);
						return return_value;
					}
						
					if (buffer_tmp[UEM_STATUS_OFFSET] != 0x00) {
						//DEBUG_CRITICAL2("Wakeup fail: uemCardInactiveCounter=%d", ccid_descriptor->uemCardInactiveCounter);
						ccid_descriptor->uemCardInactiveCounter--;
						if (ccid_descriptor->uemCardInactiveCounter <= 0) {
							buffer[7] = CCID_ICC_ABSENT;
							ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE						
							DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif							
							return_value = UEM_SignalCardLost(reader_index);
							
							return return_value;
						}
						else {
							//RF OFF ON
							cmd[0] = 0x20;
							cmd[1] = 0x0A;
							cmd[2] = 0x00;
							cmd_length = 3;
							return_value = UEM_Transcieve(reader_index, 
								cmd, cmd_length,
								buffer_tmp, &length);
								
							ccid_descriptor->uemfModeTCL = FALSE;
							ccid_descriptor->uembnBit = 0;
							
							//INVENTORY 14443A
							cmd[0] = 0x43;
							cmd_length = 1;
							
							return_value = UEM_Transcieve(reader_index, 
								cmd, cmd_length,
								buffer_tmp, &length);
								
							/*ccid_descriptor->uemCardUIDLength = buffer_tmp[4];
							memmove(
								ccid_descriptor->uemCardUID, 
								&buffer_tmp[5], 
								ccid_descriptor->uemCardUIDLength);*/
							
							if (ccid_descriptor->uemCardUIDLength != buffer_tmp[4] || 
								memcmp(ccid_descriptor->uemCardUID, &buffer_tmp[5], buffer_tmp[4]) != 0
							) {
								buffer[7] = CCID_ICC_ABSENT;
								ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE					
								DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif							
								return_value = UEM_SignalCardLost(reader_index);
								
								return return_value;
							}
						
							buffer[7] = CCID_ICC_PRESENT_ACTIVE;
							ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
							
							return return_value;
						}
					}
					
					ccid_descriptor->uemCardInactiveCounter = UEM_CARD_INACTIVE_MAX_CYCLES;
					
					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;	
				}				
				else // Mode TCL
				{
#ifdef UEM_USE_EXTENDED_ACTIVATION
					cmd[0] = 0x76;
					cmd_length = 1;
#else
					cmd[0] = 0x48;//CMD_CARD;
					cmd[1] = 0x01;//LOBYTE(2); 
					cmd[2] = 0x00;//HIBYTE(2);
					cmd[3] = 0xB2 | ccid_descriptor->uembnBit;//0xBA;// | ccid_descriptor->uembnBit; 
					//ccid_descriptor->uembnBit ^= 0x01;
					//cmd[4] = ccid_descriptor->uemCID & 0x0F;
					cmd[4] = TA & 0xFF;//0x90;// 
					cmd[5] = (TA >> 8) & 0xFF;//0x1A;// 
					cmd[6] = 0; 
					cmd[7] = 0;
					cmd_length = 8;
#endif						
					return_value = UEM_Transcieve(reader_index, 
								cmd, cmd_length,
								buffer_tmp, &length);

					if (
						(return_value != IFD_SUCCESS) || 
						(length == 0) || 
						(buffer_tmp[0] != 0x00)
					) {
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
						
						return_value = UEM_SignalCardLost(reader_index);
						
						return return_value;
					}
					
					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;	
				}
				
				break;
			
			case CS_ISO14443B:
				if (ccid_descriptor->uemfModeTCL == TRUE)
				{
#ifdef UEM_USE_EXTENDED_ACTIVATION
					cmd[0] = 0x76;
					cmd_length = 1;
#else
					cmd[0] = 0x48;//CMD_CARD;
					cmd[1] = 0x06;//0x07;//LOBYTE(7); 
					cmd[2] = 0x00;//HIBYTE(7);
					cmd[3] = 0xB2 | (ccid_descriptor->uembnBit); 
					//ccid_descriptor->uembnBit ^= 0x01;
					//cmd[4] = ccid_descriptor->uemCID & 0x0F;
					cmd[4] = 0x00; 
					cmd[5] = 0xB0; 
					cmd[6] = 0x81; 
					cmd[7] = 0x00; 
					cmd[8] = 0x01; 
					cmd[9] = TA & 0xFF;//0x90;// 
					cmd[10] = (TA >> 8) & 0xFF;//0x1A;// 
					cmd[11] = 0x00; 
					cmd[12] = 0x00;
					cmd_length = 13;
#endif
					UEM_Transcieve(
						reader_index, 
						cmd, 
						cmd_length, 
						buffer_tmp, 
						&length
					);
					
					if (
						(return_value != IFD_SUCCESS) || 
						(length == 0) || 
						(buffer_tmp[0] != 0x00)
					) {
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE				
						DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
						return_value = UEM_SignalCardLost(reader_index);
								
						return return_value;
					}
					
					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;	
				}
				else
				{
					cmd[0] = 0x55;// Halt
					memmove(
						&cmd[1],
						ccid_descriptor->uemCardUID,
						ccid_descriptor->uemCardUIDLength
					);
					cmd_length = 5;

					return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						buffer_tmp, &length);
						
					if (return_value != IFD_SUCCESS) {
						DEBUG_CRITICAL2("error=%lu", return_value);
						return return_value;
					}
					
					if (
						(length == 0) || 
						(buffer_tmp[0] != 0x00)
					) {
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE					
						DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
						return_value = UEM_SignalCardLost(reader_index);
								
						return IFD_SUCCESS;
					}

					usleep(100 * 1000);
					
					cmd[0] = 0x57;// Wakeup
					cmd[1] = 0x00;//AFI
					cmd[2] = 0x00;//SNI
					cmd_length = 3;

					return_value = UEM_Transcieve(reader_index, 
						cmd, cmd_length,
						buffer_tmp, &length);
						
					if (return_value != IFD_SUCCESS) {
						DEBUG_CRITICAL2("error=%lu", return_value);
						return return_value;
					}
					
					if (
						(length == 0) || 
						(buffer_tmp[0] != 0x00)
					) {
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE						
						DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
						return_value = UEM_SignalCardLost(reader_index);
								
						return IFD_SUCCESS;
					}

					ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
					buffer[7] = CCID_ICC_PRESENT_ACTIVE;	
				}
				
				break;
			case CS_ISO15693:
				cmd[0] = 0x31;// Inventory 15693
				cmd[1] = 0x26;
				cmd[2] = 0x01;
				cmd[3] = 0x00;
				cmd[4] = 0x00;
				cmd_length = 5;

				return_value = UEM_Transcieve(reader_index, 
					cmd, cmd_length,
					buffer_tmp, &length);
					
				if (return_value != IFD_SUCCESS) {
					DEBUG_CRITICAL2("error=%lu", return_value);
					return return_value;
				}
				
				{
					BOOL fCardDetected = FALSE;
					
					if (
						(length > 0) && 
						(buffer_tmp[0] == 0x00)
					) {
						if (buffer_tmp[1] > 0) { // RespLen > 0
							if (buffer_tmp[3] == 0) { // RetCode
								if (buffer_tmp[4] == 10) { // UID Size
									//OutputBuffer[5] // Flag
									//OutputBuffer[6] // DSFID
									if (memcmp(
										&buffer_tmp[7], 
										ccid_descriptor->uemCardUID, 
										ccid_descriptor->uemCardUIDLength) == 0//ccid_descriptor->uemCardUIDLength
									) {
										fCardDetected = TRUE;
										//DEBUG_CRITICAL("fCardDetected = TRUE");
									}
								}
							}
						}
					}
					if (fCardDetected == FALSE) {
						//DEBUG_CRITICAL("CCID_ICC_ABSENT");
						buffer[7] = CCID_ICC_ABSENT;
						ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE							
						DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
						return_value = UEM_SignalCardLost(reader_index);
								
						return return_value;
					} else {
						ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
						buffer[7] = CCID_ICC_PRESENT_ACTIVE;	
						//return IFD_SUCCESS;		
					}
				}
				break;
			default:
				;
				//*pATRLength = 0;
			}
		}
	}
	else {
		// SAM card
		// Set Holder
		cmd[0] = 0xB2;
		cmd[1] = 0x00 + (BYTE)ccid_descriptor->bCurrentSlotIndex; 
		cmd_length = 2;
		
		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			buffer_tmp, &length);
		if ((return_value != IFD_SUCCESS) || (buffer_tmp[0] != 0x00)) {
			ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			buffer[7] = CCID_ICC_ABSENT;
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE							
			DEBUG_CRITICAL("CCID_ICC_ABSENT");
#endif	
			//DEBUG_CRITICAL("SAM CCID_ICC_ABSENT");
		}
		else {
			ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
			buffer[7] = CCID_ICC_PRESENT_ACTIVE;
			//DEBUG_CRITICAL("SAM CCID_ICC_PRESENT_ACTIVE");
		}
		return IFD_SUCCESS;
	}
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE							
	DEBUG_CRITICAL3("return_value=%lu buffer[7]=0x%02X", return_value, buffer[7]);
#endif	
	return return_value;
}

RESPONSECODE UEM_CmdXfrBlock(unsigned int reader_index, unsigned int tx_length,
	BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[], int protocol)
{
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	int old_read_timeout;

#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	DEBUG_CRITICAL3("UEM_CmdXfrBlock %08X, %d", ccid_descriptor->dwFeatures, protocol);
#endif

	// Use infinite timeout in T=0, short APDU and extended APDU
	old_read_timeout = ccid_descriptor->readTimeout;

	/* APDU or TPDU? */
	switch (ccid_descriptor->dwFeatures & CCID_CLASS_EXCHANGE_MASK)
	{
		case CCID_CLASS_TPDU:
			if (protocol == T_0)
			{
				ccid_descriptor->readTimeout = 0;	// Infinite
				return_value = UEM_CmdXfrBlockTPDU_T0(reader_index,
					tx_length, tx_buffer, rx_length, rx_buffer);
			}
			else if (protocol == T_1)
			{
				return_value = UEM_CmdXfrBlockTPDU_T1(reader_index, tx_length,
					tx_buffer, rx_length, rx_buffer);
			}
			else if (protocol == 4)
			{
				return_value = UEM_CmdXfrBlockTPDU_RAW(reader_index, tx_length,
					tx_buffer, rx_length, rx_buffer);
			}
			else
				return_value = IFD_PROTOCOL_NOT_SUPPORTED;
			break;

		default:
			return_value = IFD_COMMUNICATION_ERROR;
			break;
	}

	ccid_descriptor->readTimeout = old_read_timeout;
	return return_value;
}

unsigned int FSDItoFSD(unsigned int fsdi)
{
	if (fsdi == 0)
		return 16;
	else if (fsdi == 1)
		return 24;
	else if (fsdi == 2)
		return 32;
	else if (fsdi == 3)
		return 40;
	else if (fsdi == 4)
		return 48;
	else if (fsdi == 5)
		return 64;
	else if (fsdi == 6)
		return 96;
	else if (fsdi == 7)
		return 128;
	else if (fsdi == 8)
		return 256;
	else
		return -1;
}

RESPONSECODE transmit_ISO_14443_4(unsigned int reader_index,
	LPCBYTE pbSendBuffer,
	DWORD cbSendLength,
	LPBYTE pbRecvBuffer,
	LPDWORD pcbRecvLength)
{
	//                                  0    1    2    3    4    5    6    7
	static DWORD cCLSCRF_DecodeBufSize[] = { 16, 24, 32, 40, 48, 64, 96, 128,
		//                                  8    9   10   11   12   13   14   15
		256, 256, 256, 256, 256, 256, 256, 256 };

	const unsigned char cCLSCRF_PCD_FSDI = 5;        // PCD FIFO Size = 64

	const DWORD cCLSCRF_PCD_BufSize = cCLSCRF_DecodeBufSize[cCLSCRF_PCD_FSDI];

	const int BUF_SIZE = 1024;// 300 * 10;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	//    DWORD SendLength;
	DWORD RecvLength;
	LONG Status = SCARD_S_SUCCESS;

	unsigned char PCB;
	int TCardLength = 0;
	int ParamLength = 0;
	int RChainLength = 0;
	int TChainPointer;
	int TChainRemainder;
	int BufferWorkSize = cCLSCRF_PCD_BufSize;
	DWORD CardBufSize = cCLSCRF_PCD_BufSize;

	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	switch (ccid_descriptor->uemCardStandard)
	{
	case CS_ISO14443A:
		if (ccid_descriptor->uemATSLength > 0 && ccid_descriptor->uemATS[0] > 0)
			CardBufSize = cCLSCRF_DecodeBufSize[ccid_descriptor->uemATS[1] & 0x0f];
		else
			CardBufSize = 32;   // Default
		break;

	case CS_ISO14443B:
		CardBufSize = cCLSCRF_DecodeBufSize[(ccid_descriptor->uemProtInfo[1] >> 4) & 0x0f];
		break;
	}

	if (CardBufSize < cCLSCRF_PCD_BufSize)
		BufferWorkSize = CardBufSize;

	//LPSCARD_T0_COMMAND pCmd = (LPSCARD_T0_COMMAND)pbSendBuffer;

	if (cbSendLength < 5) {
		*pcbRecvLength = 0;
		return SCARD_E_INVALID_PARAMETER;
	}

	TChainPointer = 0;
	TChainRemainder = cbSendLength;
	TCardLength = 1 + TChainRemainder;  // PCB + Params
	if (TCardLength + 2 > BufferWorkSize)    // 2 bytes CRC trailer
		TCardLength = BufferWorkSize - 2;
	SendBuffer[0] = 0x02 | ccid_descriptor->uembnBit;
	ParamLength = TCardLength - 1;
	if (ParamLength > 0) { // Size of Params
		memcpy(&SendBuffer[1], &pbSendBuffer[TChainPointer], ParamLength);
		TChainRemainder -= ParamLength;
		TChainPointer += ParamLength;
	}

	if (TChainRemainder > 0)
		SendBuffer[0] |= 0x10;

	*pcbRecvLength = 0;
	BOOL bComplete = TRUE;
	do {
		RecvLength = BUF_SIZE;

		Status = UEM_DirectIOCard(reader_index, 
			SendBuffer, TCardLength,
			ccid_descriptor->uemTimeOut,
			RecvBuffer, &RecvLength);
			
		if (Status) {
			bComplete = TRUE;
		} else {
			ccid_descriptor->uembnBit ^= 0x01;
			PCB = RecvBuffer[0] & 0xfe;
			if ((PCB & 0xef) == 0x02) // 0x02 or 0x12
			{
				RChainLength = 0;
				memcpy(&RChainLength, &RecvLength, 2);
				RChainLength -= 1;
				if (RChainLength > 0)
				{
					memcpy(&pbRecvBuffer[*pcbRecvLength], &RecvBuffer[1], RChainLength);
					*pcbRecvLength += RChainLength;
				}
			}
			if (PCB == 0x12 || PCB == 0xa2)
			{
				TCardLength = 1 + TChainRemainder;  // PCB + Params
				if (TCardLength + 2 > BufferWorkSize)    // 2 bytes CRC trailer
					TCardLength = BufferWorkSize - 2;

				SendBuffer[0] = 0x02 | ccid_descriptor->uembnBit;
				//SendBuffer[1] = m_CID;
				ParamLength = TCardLength - 1;
				if (ParamLength > 0)  // Size of Params
				{
					memcpy(&SendBuffer[1], &pbSendBuffer[TChainPointer], ParamLength);
					TChainRemainder -= ParamLength;
					TChainPointer += ParamLength;
				}
				if (TChainRemainder > 0)
					SendBuffer[0] |= 0x10;
				else if (PCB == 0x12)
					SendBuffer[0] |= 0xa0;
				bComplete = FALSE;
			}
			else
				bComplete = TRUE;
		}
#ifdef UEM_PAUSE_BETWEEN_CHAINING_COMMANDS
        usleep(UEM_PAUSE_BETWEEN_CHAINING_COMMANDS * 1000);
#endif
	} while (bComplete != TRUE);

/*#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	{
		int resp_log_pos = 0;
		char resp_log[SCR_BUFFER_SIZE * 3 + 1] = {0};
		for( unsigned int jj = 0; jj < *pcbRecvLength; jj++ ) {
			resp_log_pos += sprintf(resp_log + resp_log_pos, "%02X ", pbRecvBuffer[jj]);
		}
		DEBUG_CRITICAL2("transmit_ISO_14443_4 Response: %s", resp_log);
	}	
#endif*/

	return Status;
}

RESPONSECODE
UEM_CRD_toCard(
			unsigned int reader_index,
			BYTE*      bout,
			unsigned int       lout,
			BYTE*      bin,
			unsigned int       lin,
			unsigned int*      rlin
			) 
{
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
    UCHAR   cmd[MAX_T1_BLOCK_SIZE], rd[MAX_T1_BLOCK_SIZE];
    USHORT  inflen = 0; 
    //UCHAR CID = (UCHAR)ccid_descriptor->uemCID & 0x0F;
    unsigned int ifsc = 0, partOut = 0, remind;
    ULONG partIn;
    PUCHAR pIn = bin, pOut = bout;//&bout[0];//&bout[2];
    
#ifdef UEM_ENABLE_DEBUG_OUTPUT
    DEBUG_CRITICAL("UEM_CRD_toCard");
#endif
    
	//ifsc = FSDItoFSD(ccid_descriptor->uemT1IFSC);
	//partOut = FSDItoFSD(ccid_descriptor->uemT1IFSD);
	ifsc = FSDItoFSD(ccid_descriptor->uemIFSC);
	partOut = FSDItoFSD(ccid_descriptor->uemIFSD);
	if (ifsc == 0) // Jure
		//ifsc = FSDItoFSD(ccid_descriptor->uemReaderCapabilitiesMaxIFSD); // Jure
		ifsc = FSDItoFSD(0); // Jure
	if (partOut == 0) // Jure
		//partOut = FSDItoFSD(ccid_descriptor->uemReaderCapabilitiesMaxIFSD); // Jure
		partOut = FSDItoFSD(0); // Jure
    if (partOut < ifsc)
        ifsc = partOut;
	if (ifsc >= 15) // Jure
		ifsc -= 15; // for service info

    if ((ifsc == 0) || (ifsc == (unsigned int)-1)) { // not known yet // == >>>>> <= Jure
        ifsc = 30;
    }
    remind = lout;//-2;   
    if (remind > partOut) {
        partOut = ifsc;
        remind -= ifsc;
    } else {
        partOut = remind;
        remind = 0;
    }

    cmd[0] = 0x48;//CMD_CARD;
    cmd[1] = (BYTE)((1 + partOut)&0x000000FF); 
	cmd[2] = (BYTE)(((1 + partOut)&0x0000FF00) >> 8);
    cmd[3] = 0x02;//bout[0]; // PCB
    if (remind)
        cmd[3] |= 0x10; // chaining

    if ((cmd[3] & 0x80) == 0) { // I or R
        cmd[3] |= ccid_descriptor->uembnBit;
    }
    //cmd[4] = CID; //bout[1] | CID;
    memcpy(&cmd[4], pOut, partOut); // copy INF
    memcpy(&cmd[4 + partOut], &ccid_descriptor->uemTimeOut, 4);
#ifdef UEM_ENABLE_DEBUG_OUTPUT
    DEBUG_CRITICAL5("TimeOut=%lu uemIFSC=%lu, ifsc=%lu, partOut=%lu", ccid_descriptor->uemTimeOut, ccid_descriptor->uemIFSC, ifsc, partOut);
#endif
    *rlin = 0;
    for (;;) {
		return_value = UEM_Transcieve(reader_index, 
			cmd, partOut+8,
			rd, 
			//MAX_T1_BLOCK_SIZE,
			&partIn);

		if (return_value != IFD_SUCCESS) {
			DEBUG_CRITICAL2("error=%lu", return_value);
			return return_value;
		}

		//usleep(10 * 1000);
        
        if (rd[0] == 0x00) { // PCD ok
            inflen = rd[1] + (rd[2] << 8);
            if ((rd[3] & 0xEE) == 0x02) { // I-block
                memcpy(pIn, rd + 4, inflen-1);
                pIn += inflen-1;
                *rlin += inflen-1;
                if (ccid_descriptor->uembnBit == (rd[3] & 0x01)) // block number is equal (rule B)
                    ccid_descriptor->uembnBit ^= 0x01;
            }
            if ((rd[3] & 0xEE) == 0xA2) { // R-block
                if (ccid_descriptor->uembnBit == (rd[3] & 0x01)) // block number is equal (rule B)
                    ccid_descriptor->uembnBit ^= 0x01;
            }
            if ((rd[3] & 0xCE) == 0xC2) { // S-block
				if ((rd[3] & 0x3E) == 0x32) { // WTX
					cmd[0] = 0x48;//UEM_CMD_CARD;
					cmd[1] = 0x02;//0x03; 
					cmd[2] = 0x00;
					cmd[3] = 0xF2;//UEM_PCB_SWTX;
					//cmd[4] = CID;
					cmd[4] = rd[4] & 0x3F;
					memcpy(&cmd[5], &ccid_descriptor->uemTimeOut, 4);
					partOut = 1;
					continue;
				}
				if ((rd[3] & 0x32) == 0x02) { // DESELECT
					DEBUG_CRITICAL("SCARD_E_NOT_TRANSACTED");
					return SCARD_E_NOT_TRANSACTED;
				}
				//if (m_bnBit == (rd[3] & 0x01)) // block number is equal (rule B)
				//	m_bnBit ^= 0x01;
			}
            if (rd[3] & 0x10) { // ICC chaining -- send R(ACK)
                cmd[0] = 0x48;//CMD_CARD;
                cmd[1] = 0x01;//0x02;//LOBYTE(2); 
				cmd[2] = 0x00;//HIBYTE(2);
                cmd[3] = 0xA2 | ccid_descriptor->uembnBit;//PCB_RACK | ccid_descriptor->uembnBit;
                //cmd[4] = CID;
                memcpy(&cmd[4], &ccid_descriptor->uemTimeOut, 4);
                partOut = 0;
            } else {
                if (remind) {
                    pOut += partOut;
                    if (remind > ifsc) {
                        partOut = ifsc;
                        remind -= ifsc;
                    } else {
                        partOut = remind;
                        remind = 0;
                    }
                    cmd[0] = 0x48;//CMD_CARD;
					cmd[1] = (BYTE)((1 + partOut)&0x000000FF); 
					cmd[2] = (BYTE)(((1 + partOut)&0x0000FF00) >> 8);
					cmd[3] = 0x02;//bout[0]; // PCB
                    if (remind)
                        cmd[3] |= 0x10; // chaining
                    if ((cmd[3] & 0x80) == 0) { // I or R
                        cmd[3] |= ccid_descriptor->uembnBit;
                    }
                    //cmd[4] = CID; //bout[1] | CID;
                    memcpy(&cmd[4], pOut, partOut);
                    memcpy(&cmd[4 + partOut], &ccid_descriptor->uemTimeOut, 4);
                } else {
                    break;
                }
            }
        } else {// PCD failed
			DEBUG_CRITICAL2("IFD_RESPONSE_TIMEOUT err = 0x%02X", rd[0]);
            return IFD_RESPONSE_TIMEOUT;//STATUS_IO_TIMEOUT;
        }
#ifdef UEM_PAUSE_BETWEEN_CHAINING_COMMANDS
        usleep(UEM_PAUSE_BETWEEN_CHAINING_COMMANDS * 1000);
#endif
    }
    
    return return_value;
    //return IFD_SUCCESS;
}

/*RESPONSECODE
UEM_CRD_toCard(
			unsigned int reader_index,
			BYTE*      bout,
			unsigned int       lout,
			BYTE*      bin,
			unsigned int       lin,
			unsigned int*      rlin
			) 
{
	RESPONSECODE return_value = IFD_SUCCESS;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
    UCHAR   cmd[MAX_T1_BLOCK_SIZE], rd[MAX_T1_BLOCK_SIZE];
    USHORT  inflen = 0; 
    UCHAR CID = (UCHAR)ccid_descriptor->uemCID & 0x0F;
    unsigned int ifsc = 0, partOut = 0, remind;
    ULONG partIn;
    PUCHAR pIn = bin, pOut = bout;//&bout[0];//&bout[2];
    
#ifdef UEM_ENABLE_DEBUG_OUTPUT
    DEBUG_CRITICAL("UEM_CRD_toCard");
#endif
    
	ifsc = FSDItoFSD(ccid_descriptor->uemT1IFSC);
	partOut = FSDItoFSD(ccid_descriptor->uemT1IFSD);
	if (ifsc == 0) // Jure
		ifsc = FSDItoFSD(ccid_descriptor->uemReaderCapabilitiesMaxIFSD); // Jure
	if (partOut == 0) // Jure
		partOut = FSDItoFSD(ccid_descriptor->uemReaderCapabilitiesMaxIFSD); // Jure
    if (partOut < ifsc)
        ifsc = partOut;
	if (ifsc >= 15) // Jure
		ifsc -= 15; // for service info

    if ((ifsc == 0) || (ifsc == (unsigned int)-1)) { // not known yet // == >>>>> <= Jure
        ifsc = 30;
    }
    remind = lout;//-2;   
    if (remind > partOut) {
        partOut = ifsc;
        remind -= ifsc;
    } else {
        partOut = remind;
        remind = 0;
    }

    cmd[0] = 0x48;//CMD_CARD;
    cmd[1] = (BYTE)((partOut+2)&0x000000FF); 
	cmd[2] = (BYTE)(((partOut+2)&0x0000FF00) >> 8);
    cmd[3] = 0x0A;//bout[0]; // PCB
    if (remind)
        cmd[3] |= 0x10; // chaining

    if ((cmd[3] & 0x80) == 0) { // I or R
        cmd[3] |= ccid_descriptor->uembnBit;
    }
    cmd[4] = CID; //bout[1] | CID;
    memmove(&cmd[5], pOut, partOut); // copy INF
    memmove(&cmd[partOut + 5], &ccid_descriptor->uemTimeOut, 4);
    *rlin = 0;
    for (;;) {
		return_value = UEM_Transcieve(reader_index, 
			cmd, partOut+9,
			rd, 
			//MAX_T1_BLOCK_SIZE,
			&partIn);
		if (return_value != IFD_SUCCESS)
			return return_value;
        
        if (rd[0] == 0x00) { // PCD ok
            inflen = rd[1] + (rd[2] << 8);
            if ((rd[3] & 0xEE) == 0x0A) { // I-block
                memmove(pIn, rd + 5, inflen-2);
                pIn += inflen-2;
                *rlin += inflen-2;
                if (ccid_descriptor->uembnBit == (rd[3] & 0x01)) // block number is equal (rule B)
                    ccid_descriptor->uembnBit ^= 0x01;
            }
            if ((rd[3] & 0xEE) == 0xAA) { // R-block
                if (ccid_descriptor->uembnBit == (rd[3] & 0x01)) // block number is equal (rule B)
                    ccid_descriptor->uembnBit ^= 0x01;
            }
            if ((rd[3] & 0xCE) == 0xCA) { // S-block
				if ((rd[3] & 0x3E) == 0x3A) { // WTX
					cmd[0] = 0x48;//UEM_CMD_CARD;
					cmd[1] = 0x03; 
					cmd[2] = 0x00;
					cmd[3] = 0xFA;//UEM_PCB_SWTX;
					cmd[4] = CID;
					cmd[5] = rd[5] & 0x3F;
					memcpy(&cmd[6], &ccid_descriptor->uemTimeOut, 4);
					partOut = 1;
					continue;
				}
				if ((rd[3] & 0x3E) == 0x0A) { // DESELECT
					return SCARD_E_NOT_TRANSACTED;
				}
				//if (m_bnBit == (rd[3] & 0x01)) // block number is equal (rule B)
				//	m_bnBit ^= 0x01;
			}
            if (rd[3] & 0x10) { // ICC chaining -- send R(ACK)
                cmd[0] = 0x48;//CMD_CARD;
                cmd[1] = 0x02;//LOBYTE(2); 
				cmd[2] = 0x00;//HIBYTE(2);
                cmd[3] = 0xAA | ccid_descriptor->uembnBit;//PCB_RACK | ccid_descriptor->uembnBit;
                cmd[4] = CID;
                memmove(&cmd[5], &ccid_descriptor->uemTimeOut, 4);
                partOut = 0;
            } else {
                if (remind) {
                    pOut += partOut;
                    if (remind > ifsc) {
                        partOut = ifsc;
                        remind -= ifsc;
                    } else {
                        partOut = remind;
                        remind = 0;
                    }
                    cmd[0] = 0x48;//CMD_CARD;
					cmd[1] = (BYTE)((partOut+2)&0x000000FF); 
					cmd[2] = (BYTE)(((partOut+2)&0x0000FF00) >> 8);
					cmd[3] = 0x0A;//bout[0]; // PCB
                    if (remind)
                        cmd[3] |= 0x10; // chaining
                    if ((cmd[3] & 0x80) == 0) { // I or R
                        cmd[3] |= ccid_descriptor->uembnBit;
                    }
                    cmd[4] = CID; //bout[1] | CID;
                    memmove(&cmd[5], pOut, partOut);
                    memmove(&cmd[partOut + 5], &ccid_descriptor->uemTimeOut, 4);
                } else {
                    break;
                }
            }
        } else {// PCD failed
            return IFD_RESPONSE_TIMEOUT;//STATUS_IO_TIMEOUT;
        }
    }
    
    return return_value;
    //return IFD_SUCCESS;
}*/

RESPONSECODE UEM_TransmitT0(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI)
{
	BYTE cmd[UEM_HEADER_SIZE + tx_length];	/* Header + APDU buffer */
	status_t ret;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_TransmitT0 NOT IMPLEMENTED");
#endif
	
	/*
	// EXCHANGE_TPDU_T0 or EXCHANGE_SAM_TPDU_T0
	cmd[0] = 0x01;
	cmd[1] = ccid_descriptor->bCurrentSlotIndex == 0 ? 0xA0 : 0xB0;
	cmd[2] = (tx_length >> 8) & 0xFF;
	cmd[3] = tx_length & 0xFF;
	memcpy(cmd + UEM_HEADER_SIZE, tx_buffer, tx_length);

	ret = WritePort(reader_index, UEM_HEADER_SIZE + tx_length, cmd);
	CHECK_STATUS(ret)
	*/
	
	return IFD_SUCCESS;
}

RESPONSECODE UEM_TransmitT1(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI)
{
	BYTE cmd[UEM_HEADER_SIZE + tx_length];	/* Header + APDU buffer */
	status_t ret;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	RESPONSECODE status = IFD_SUCCESS;
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_TransmitT1");

	for( unsigned int i = 0; i < tx_length; i++ )
		DEBUG_CRITICAL2("Transmit: %02X", tx_buffer[i] );
#endif
	
	return IFD_NOT_SUPPORTED ;
	
#ifdef UEM_USE_EXTENDED_ACTIVATION
	BYTE					InputBuffer[SCR_BUFFER_SIZE];
	ULONG					InputBufferLength = sizeof(InputBuffer);
	BYTE					OutputBuffer[SCR_BUFFER_SIZE];
	ULONG					OutputBufferLength = sizeof(OutputBuffer);
	DWORD					ulBytesReadOrWritten = 0;
	
	//tx_buffer += 3;
	//tx_length -= 4;
	
	InputBuffer[0] = 0x47;
	InputBuffer[1] = 0x00;
	InputBuffer[2] = 0x00;
	memcpy(&InputBuffer[3], &tx_length, 2);
	memcpy(&InputBuffer[5], tx_buffer, tx_length);
	InputBufferLength = 5 + tx_length;

	ulBytesReadOrWritten = OutputBufferLength;
	status = UEM_Transcieve(reader_index, 
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten);

	/*if (status == SCARD_F_INTERNAL_ERROR) {
		//m_fModeTCL = FALSE;
		//m_state = SCARD_POWERED;

		usleep(200 * 1000);
		status = m_device->m_pUemReader->iso14443AHalt();
		usleep(200 * 1000);
		status = m_device->m_pUemReader->iso14443AWakeUp(m_cardUID, m_ucCardUIDLength);

		//if (status != 0) {
		//	Sleep(200);
		//	status = m_device->m_pUemReader->iso14443AHalt();
		//	Sleep(200);
		//	status = m_device->m_pUemReader->iso14443AWakeUp(m_cardUID, m_ucCardUIDLength);
		//}

		if (status != 0) {
			m_state = SCARD_ABSENT;
			return status;
		}

		//Sleep(200);

		if (QueryATR(m_ATR, &m_ATRLength, false)) {
			rx_buffer[0] = 0x63;
			rx_buffer[1] = 0x00;
			*rx_length = 2;
			status = SCARD_S_SUCCESS;
		}
		else {
			return status;
		}

		//Sleep(200);
	}*/

	if (status == SCARD_S_SUCCESS) {
		if (ulBytesReadOrWritten < 3) {
			return SCARD_F_COMM_ERROR;
		}
		else {
			memcpy(&ccid_descriptor->uemRxLength, &OutputBuffer[1], 2);
			//if (rx_length < ccid_descriptor->uemRxLength) {
			//	DEBUG_CRITICAL("SCARD_E_INSUFFICIENT_BUFFER");
			//	return SCARD_E_INSUFFICIENT_BUFFER;
			//}
			if (ulBytesReadOrWritten - 3 < ccid_descriptor->uemRxLength) {
				DEBUG_CRITICAL("SCARD_E_COMM_DATA_LOST");
				return SCARD_E_COMM_DATA_LOST;
			}
			memcpy(ccid_descriptor->uemRxBuffer, &OutputBuffer[3], ccid_descriptor->uemRxLength);
		}
	}
#else
	//status = UEM_CRD_toCard(reader_index, tx_buffer, tx_length, rx_buffer, *rx_length, rx_length);
	status = UEM_CRD_toCard(reader_index, tx_buffer + 3, tx_length - 4, ccid_descriptor->uemRxBuffer, rx_length, &ccid_descriptor->uemRxLength);
	//status = UEM_CRD_toCard(reader_index, tx_buffer + 3, tx_length - 4, ccid_descriptor->uemRxBuffer, rx_length, &ccid_descriptor->uemRxLength);
#endif
	
	/*
	// EXCHANGE_TPDU_T1 or EXCHANGE_SAM_TPDU_T1
	cmd[0] = 0x01;
	cmd[1] = ccid_descriptor->bCurrentSlotIndex == 0 ? 0xA1 : 0xB1;
	cmd[2] = (tx_length >> 8) & 0xFF;
	cmd[3] = tx_length & 0xFF;
	memcpy(cmd + UEM_HEADER_SIZE, tx_buffer, tx_length);

	ret = WritePort(reader_index, UEM_HEADER_SIZE + tx_length, cmd);
	CHECK_STATUS(ret)
	*/
	
	return status;
}

RESPONSECODE UEM_TransmitPPS(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI)
{
	BYTE cmd[UEM_HEADER_SIZE + tx_length];	/* Header + APDU buffer */
	status_t ret;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_TransmitPPS NOT IMPLEMENTED");
#endif
	
	for( unsigned int i = 0; i < tx_length; i++ ) 
#ifdef UEM_ENABLE_DEBUG_OUTPUT
		DEBUG_CRITICAL2("Transmit: %02X", tx_buffer[i] );
#endif
	
	/*
	// SET_CARD_PPS or SET_SAM_CARD_PPS
	cmd[0] = 0x01;
	cmd[1] = ccid_descriptor->bCurrentSlotIndex == 0 ? 0x0A : 0x0C;
	cmd[2] = (tx_length >> 8) & 0xFF;
	cmd[3] = tx_length & 0xFF;
	memcpy(cmd + UEM_HEADER_SIZE, tx_buffer, tx_length);

	ret = WritePort(reader_index, UEM_HEADER_SIZE + tx_length, cmd);
	CHECK_STATUS(ret)
	*/
	
	/*if (rx_length < tx_length)
	{
		DEBUG_CRITICAL2("overrun by %d bytes", tx_length - rx_length);
		ccid_descriptor->uemRxLength = 0;
		return IFD_ERROR_INSUFFICIENT_BUFFER;
	}*/
	
	ccid_descriptor->uemRxLength = tx_length;
	memcpy(ccid_descriptor->uemRxBuffer, tx_buffer, ccid_descriptor->uemRxLength);

	//BYTE PPS_RESPONSE[3] = {0xFF, 0x00, 0xFF};
	//ccid_descriptor->uemRxLength = sizeof(PPS_RESPONSE);
	//memcpy(ccid_descriptor->uemRxBuffer, PPS_RESPONSE, ccid_descriptor->uemRxLength);
	
	return IFD_SUCCESS;
}

RESPONSECODE UEM_Receive(unsigned int reader_index,
	/*@out@*/ unsigned int *rx_length,
	/*@out@*/ BYTE rx_buffer[], BYTE *chain_parameter)
{
	BYTE cmd[UEM_HEADER_SIZE + CMD_BUF_SIZE];	/* Header + APDU buffer */
	RESPONSECODE return_value = IFD_SUCCESS;
	status_t ret;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_Receive");
#endif

	for( unsigned int i = 0; i < ccid_descriptor->uemRxLength; i++ ) 
#ifdef UEM_ENABLE_DEBUG_OUTPUT
		DEBUG_CRITICAL2("Receive: %02X", ccid_descriptor->uemRxBuffer[i] );
#endif
	
	/*if (ccid_descriptor->uemRxBuffer[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(ccid_descriptor->uemRxBuffer[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}*/

	//ccid_descriptor->uemRxLength -= 3;//UEM_HEADER_SIZE;
	if (ccid_descriptor->uemRxLength <= *rx_length)
		*rx_length = ccid_descriptor->uemRxLength;
	else
	{
		DEBUG_CRITICAL2("overrun by %d bytes", ccid_descriptor->uemRxLength - (*rx_length));
		ccid_descriptor->uemRxLength = *rx_length;
		return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
	}

	memcpy(rx_buffer, ccid_descriptor->uemRxBuffer /*+ 3*/ /*UEM_HEADER_SIZE*/, ccid_descriptor->uemRxLength);

	return return_value;
}

RESPONSECODE UEM_Receive1(unsigned int reader_index,
	/*@out@*/ unsigned int *rx_length,
	/*@out@*/ BYTE rx_buffer[], BYTE *chain_parameter)
{
	BYTE cmd[UEM_HEADER_SIZE + CMD_BUF_SIZE];	/* Header + APDU buffer */
	unsigned int length;
	RESPONSECODE return_value = IFD_SUCCESS;
	status_t ret;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_Receive");
#endif

	length = sizeof(cmd);
	ret = ReadPort(reader_index, &length, cmd);
	CHECK_STATUS(ret)

	if (cmd[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(cmd[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}

	length -= UEM_HEADER_SIZE;
	if (length <= *rx_length)
		*rx_length = length;
	else
	{
		DEBUG_CRITICAL2("overrun by %d bytes", length - (*rx_length));
		length = *rx_length;
		return_value = IFD_ERROR_INSUFFICIENT_BUFFER;
	}

	memcpy(rx_buffer, cmd + UEM_HEADER_SIZE, length);

	return return_value;
}

RESPONSECODE UEM_SetParameters(unsigned int reader_index, char protocol,
	unsigned int length, BYTE buffer[])
{
	BYTE cmd[SCR_BUFFER_SIZE];
	int i;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	status_t res;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_SetParameters NOT IMPLEMENTED");
#endif
	
	DEBUG_COMM2("length: %d bytes", length);

	/*
	// SET_READER_PPS or SET_SAM_READER_PPS
	cmd[0] = 0x01;
	cmd[1] = ccid_descriptor->bCurrentSlotIndex == 0 ? 0x0B : 0x0D;
	cmd[2] = 0x00;
	cmd[3] = 0x04;
	cmd[4] = 0xFF;							// PPSS
	cmd[5] = protocol == 0 ? 0x10 : 0x11;	// PPS0: 0x10 (T=0), 0x11 (T=1)
	cmd[6] = buffer[0];						// PPS1: TA(1)
	cmd[7] = 0;								// PCK: XOR checksum
	for (i = 4; i < 7; i++)
		cmd[7] ^= cmd[i];

	res = WritePort(reader_index, sizeof(cmd), cmd);
	CHECK_STATUS(res)

	length = sizeof(cmd);
	res = ReadPort(reader_index, &length, cmd);
	CHECK_STATUS(res)

	if (cmd[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(cmd[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}
	*/
	
	return IFD_SUCCESS;
	//return IFD_ERROR_SET_FAILURE;
}

RESPONSECODE UEM_SetCardVoltage(unsigned int reader_index,
	BYTE TxBuffer[], unsigned int TxLength,
	BYTE RxBuffer[], unsigned int *RxLength)
{
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE cardVoltage;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_SetCardVoltage");
#endif

	if ((TxBuffer != NULL) && (TxLength >= 1))
	{
		cardVoltage = TxBuffer[0];
		if (cardVoltage <= 3)
			ccid_descriptor->cardVoltage = cardVoltage;
		else
		{
			DEBUG_CRITICAL2("Card voltage %d is not supported", cardVoltage);
			ccid_descriptor->cardVoltage = 0;
			return_value = IFD_COMMUNICATION_ERROR;
		}
	}

	if (RxLength != NULL)
		*RxLength = 0;

	DEBUG_INFO2("cardVoltage: %d", ccid_descriptor->cardVoltage);
	return return_value;
}

RESPONSECODE UEM_SetCardType(unsigned int reader_index,
	BYTE TxBuffer[], unsigned int TxLength,
	BYTE RxBuffer[], unsigned int *RxLength)
{
	RESPONSECODE return_value = IFD_SUCCESS;
	DWORD cardModel;
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_SetCardType");
#endif

	ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;

	if ((TxBuffer != NULL) && (TxLength >= sizeof(cardModel)))
	{
		cardModel = *((PDWORD) TxBuffer);
		switch (cardModel)
		{
			case CARD_TYPE_MCU_AUTO:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_MCU_AUTO;
				break;

			case CARD_TYPE_I2C_1K:
			case CARD_TYPE_I2C_2K:
			case CARD_TYPE_I2C_4K:
			case CARD_TYPE_I2C_8K:
			case CARD_TYPE_I2C_16K:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_I2C_1K_16K;
				break;

			case CARD_TYPE_I2C_32K:
			case CARD_TYPE_I2C_64K:
			case CARD_TYPE_I2C_128K:
			case CARD_TYPE_I2C_256K:
			case CARD_TYPE_I2C_512K:
			case CARD_TYPE_I2C_1024K:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_I2C_32K_1024K;
				break;

			case CARD_TYPE_AT88SC153:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_AT88SC153;
				break;

			case CARD_TYPE_AT88SC1608:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_AT88SC1608;
				break;

			case CARD_TYPE_SLE4418:
			case CARD_TYPE_SLE4428:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_SLE4418_4428;
				break;

			case CARD_TYPE_SLE4432:
			case CARD_TYPE_SLE4442:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_SLE4432_4442;
				break;

			case CARD_TYPE_SLE4406:
			case CARD_TYPE_SLE4436:
			case CARD_TYPE_SLE5536:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_SLE4406_4436_5536;
				break;

			case CARD_TYPE_MCU_T0:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_MCU_T0;
				break;

			case CARD_TYPE_MCU_T1:
				ccid_descriptor->cardModel = UEM_CARD_TYPE_MCU_T1;
				break;

			default:
				DEBUG_CRITICAL2("Card model " DWORD_D " is not supported", cardModel);
				ccid_descriptor->cardModel = UEM_CARD_TYPE_MCU_AUTO;
				return_value = IFD_COMMUNICATION_ERROR;
				break;
		}
	}

	if (RxLength != NULL)
		*RxLength = 0;

	DEBUG_INFO2("cardModel: %d", ccid_descriptor->cardModel);
	return return_value;
}

RESPONSECODE UEM_GetFirmwareVersion(unsigned int reader_index,
	BYTE firmwareVersion[])
{
	BYTE cmd[1];
	status_t res;
	ULONG length = 7;
	RESPONSECODE return_value = IFD_SUCCESS;
	BYTE buffer_tmp[7] = {0};

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_GetFirmwareVersion");
#endif
	
	cmd[0] = 0x64;

	return_value = UEM_Transcieve(reader_index, 
		cmd, sizeof(cmd),
		buffer_tmp, &length);
	if (return_value != IFD_SUCCESS)
		return return_value;
			
	//if (length < sizeof(buffer_tmp))
	//{
	//	DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
	//	return IFD_COMMUNICATION_ERROR;
	//}

	if (buffer_tmp[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(buffer_tmp[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return IFD_COMMUNICATION_ERROR;
	}

	memcpy(firmwareVersion, buffer_tmp + UEM_HEADER_SIZE, 6);
	
	return return_value;
}

/*static RESPONSECODE UEM_CmdXfrBlockTPDU_T0(unsigned int reader_index,
	unsigned int tx_length, unsigned char tx_buffer[], unsigned int *rx_length,
	unsigned char rx_buffer[])
{
	RESPONSECODE return_value = IFD_SUCCESS;

	DEBUG_CRITICAL("UEM_CmdXfrBlockTPDU_T0");

	DEBUG_COMM2("T=0: %d bytes", tx_length);

	return_value = UEM_TransmitT0(reader_index, tx_length, tx_buffer, 0, 0);
	if (return_value != IFD_SUCCESS)
		return return_value;

	return UEM_Receive(reader_index, rx_length, rx_buffer, NULL);
}*/

static RESPONSECODE UEM_CmdXfrBlockTPDU_T0(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CmdXfrBlockTPDU_T0 REDIRECT TO T1");
#endif
	return UEM_CmdXfrBlockTPDU_T1(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
}

void UEM_AddressToSectorAndType(
	ULONG ulAddress,
	BYTE* pucUemKeyType,
	BYTE* pucSectorNumber,
	BYTE* pucBlockNumber
	)
{
	/*KFLOATING_SAVE  saveData;
	NTSTATUS        status;
	
	status = KeSaveFloatingPointState(&saveData);
    if (!NT_SUCCESS(status))
    {
		*pucUemKeyType = 0xFF;
		*pucSectorNumber = 0xFF;
		*pucBlockNumber = 0xFF;
        return;
    }*/
	if (ulAddress <= 0x007F) {
		*pucUemKeyType = 0x00;
		if (pucSectorNumber != NULL)
			*pucSectorNumber = (BYTE)floor((double)ulAddress / 4);
		if (pucBlockNumber != NULL)
			*pucBlockNumber = ulAddress % 4;
	} else if (ulAddress <= 0x00FF) {
		*pucUemKeyType = 0x00;
		if (pucSectorNumber != NULL)
			*pucSectorNumber = 32 + (BYTE)floor(((double)ulAddress - 128) / 16);
		if (pucBlockNumber != NULL)
			*pucBlockNumber = (ulAddress - 128) % 16;
	} else if (ulAddress == 0xB000) {
		*pucUemKeyType = 0xB0;
	} else if (ulAddress == 0xB001) {
		*pucUemKeyType = 0xB1;
	} else if (ulAddress == 0xB002) {
		*pucUemKeyType = 0xB2;
	} else if (ulAddress == 0xB003) {
		*pucUemKeyType = 0xB3;
	} else if (ulAddress <= 0x403F) {
		if ((ulAddress % 2) == 0)
			*pucUemKeyType = 0x4A;
		else
			*pucUemKeyType = 0x4B;
	} else if (ulAddress <= 0x404F) {
		if ((ulAddress % 2) == 0)
			*pucUemKeyType = 0x4A;
		else
			*pucUemKeyType = 0x4B;
	} else if (ulAddress == 0x8000) {
		*pucUemKeyType = 0x80;
	} else if (ulAddress == 0x9000) {
		*pucUemKeyType = 0x90;
	} else if (ulAddress == 0x9001) {
		*pucUemKeyType = 0x91;
	} else if (ulAddress == 0x9002) {
		*pucUemKeyType = 0x92;
	} else if (ulAddress == 0x9003) {
		*pucUemKeyType = 0x93;
	} else if (ulAddress == 0x9004) {
		*pucUemKeyType = 0x94;
	} else if (ulAddress == 0xA000) {
		*pucUemKeyType = 0xA0;
	} else if (ulAddress == 0xA001) {
		*pucUemKeyType = 0xA1;
	} else if (ulAddress == 0xA080) {
		*pucUemKeyType = 0xA2;
	} else if (ulAddress == 0xA081) {
		*pucUemKeyType = 0xA3;
	}
	//KeRestoreFloatingPointState(&saveData);
}

RESPONSECODE UEM_TranslateError(BYTE ucErrorCode)
{
	RESPONSECODE Status;

	switch (ucErrorCode)
	{
	case 0x01:
		Status = SCARD_E_UNSUPPORTED_FEATURE;
		break;
	case 0x02:
		Status = SCARD_E_COMM_DATA_LOST;
		break;
	case 0x03:
		Status = SCARD_E_INVALID_PARAMETER;
		break;
	case 0x0F:
		Status = SCARD_E_UNEXPECTED;
		break;
	case 0x010:
		Status = SCARD_E_INVALID_PARAMETER;
		break;
	case 0x011:
		Status = SCARD_E_NO_ACCESS;
		break;
	case 0x012:
		Status = SCARD_E_NO_ACCESS;
		break;
	case 0x013:
		Status = SCARD_E_NOT_TRANSACTED;
		break;
	case 0x014:
		Status = SCARD_E_NOT_TRANSACTED;
		break;
	default:
		Status = SCARD_E_UNEXPECTED;
	}
	if ((ucErrorCode > 0xA0) && (ucErrorCode < 0xDF))
		Status = SCARD_E_UNEXPECTED;

	return Status;
}

RESPONSECODE
UEM_CRD_IncrementDecrementValue(
	unsigned int reader_index,
	unsigned int *rx_length,
	BYTE rx_buffer[],
	BYTE ucCmdType,
	DWORD dwSrcAddress,
	DWORD dwDstAddress,
	DWORD dwValue)
{	// FF D6 00 0A 04 00 00 00 02
    RESPONSECODE   			status = IFD_SUCCESS;
	
    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	BYTE 					ucSrcSectorNumber;
	BYTE 					ucDstSectorNumber;
	BYTE 					ucSrcBlockNumber;
	BYTE 					ucDstBlockNumber;
	BYTE					ucUemValueType;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	switch (ccid_descriptor->uemCardType)
	{
	case CT_MIFARE_CLASSIC: // FF D6 00 28 10 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 02
	
		if ((dwSrcAddress == 3)||(dwSrcAddress == 15))
		{
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x6A;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return SCARD_S_SUCCESS;
		}
		
		InputBuffer[0] = 0x1B;
		InputBuffer[1] = (ucCmdType == 0xA0) ? 0xC1 : 0xC0; // Increment
		InputBuffer[2] = (BYTE)dwSrcAddress;
		memcpy(
			&InputBuffer[3],
			&dwValue,
			4
		);
		InputBuffer[7] = (BYTE)dwDstAddress;
		InputBufferLength = 8;
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		if (status != IFD_SUCCESS) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x62;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return SCARD_S_SUCCESS;
		}
		
		if (ulBytesReadOrWritten == 0)
			return SCARD_F_COMM_ERROR;
		if (OutputBuffer[0] != 0x00)
			return SCARD_F_INTERNAL_ERROR;

		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;

		break;	
	case CT_MIFARE_PLUS: // FF D6 00 28 10 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01
		UEM_AddressToSectorAndType(dwSrcAddress, &ucUemValueType, 
								&ucSrcSectorNumber, &ucSrcBlockNumber);
	
		UEM_AddressToSectorAndType(dwDstAddress, &ucUemValueType, 
								&ucDstSectorNumber, &ucDstBlockNumber);
	
		if ((ucSrcBlockNumber == 3)||(ucSrcBlockNumber == 15)||(ucDstBlockNumber == 3)||(ucDstBlockNumber == 15))
		{
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x6A;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return SCARD_S_SUCCESS;
		}
		
		InputBuffer[0] = 0xA1;
		InputBuffer[1] = (ucCmdType == 0xA0) ? 0xB0 : 0xB2; // Increment
		InputBuffer[2] = 0x01; // 0x06
		InputBuffer[3] = ucSrcSectorNumber;
		InputBuffer[4] = ucSrcBlockNumber;
		memcpy(
			&InputBuffer[5],
			&dwValue,
			4
		);
		InputBufferLength = 9;
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		if (status != IFD_SUCCESS) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x62;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return SCARD_S_SUCCESS;
		}
		
		if (ulBytesReadOrWritten == 0)
			return SCARD_F_COMM_ERROR;
		if (OutputBuffer[0] != 0x00)
			return SCARD_F_INTERNAL_ERROR;

		InputBuffer[0] = 0xA1;
		InputBuffer[1] = 0xB4; // Transfer
		InputBuffer[2] = 0x01; // 0x06
		InputBuffer[3] = ucDstSectorNumber;
		InputBuffer[4] = ucDstBlockNumber;
		InputBufferLength = 5;
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		if (status != IFD_SUCCESS) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x62;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return SCARD_S_SUCCESS;
		}
		
		if (ulBytesReadOrWritten == 0)
			return SCARD_F_COMM_ERROR;
		if (OutputBuffer[0] != 0x00)
			return SCARD_F_INTERNAL_ERROR;

		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;

		break;
	default:
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x6A;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}

	return status;
}

RESPONSECODE
UEM_CRD_ISO15693_ReadSingleBlock( unsigned int reader_index,
	BYTE   bFlags,
	BYTE   bBlockNumber,
	BYTE*  pbFlags,
	BYTE*  pbBlockSecurityStatus,
	BYTE*  pbData,
	BYTE*  pbErrorCode)
{
	RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	
	const DWORD Timeout = 1200;         // 11,3 ms
	BOOLEAN fOption = bFlags & 0x40;
	*pbErrorCode = 0;

	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	InputBuffer[0] = 0x48;
	InputBuffer[1] = 0x0B;
	InputBuffer[2] = 0x00;
	InputBuffer[3] = bFlags;
	InputBuffer[4] = 0x20; // Read Single Block
	memcpy(
		&InputBuffer[5],
		ccid_descriptor->uemCardUID,
		8
	);
	InputBuffer[13] = bBlockNumber;
	memcpy(
		&InputBuffer[14],
		&Timeout,
		4
	);
	InputBufferLength = 18;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);

	*pbFlags = 0x01;

	if (status != IFD_SUCCESS) {
		return status;
	}
	
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}

	if (ulBytesReadOrWritten < 5)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	
	*pbFlags = OutputBuffer[3];
	*pbBlockSecurityStatus = 0;
	if ((*pbFlags) & 0x01)
	{
		*pbErrorCode = OutputBuffer[4];
		status = UEM_TranslateError(*pbErrorCode);

		memset(pbData, 0x00, ICODE_BLOCK_SIZE);
	}
	else
	{
		DWORD Expected = (fOption) ? 9 : 8;
		DWORD Index = 4;
		if (ulBytesReadOrWritten < Expected)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		if (fOption)
			*pbBlockSecurityStatus = OutputBuffer[Index++];
		memcpy(
			pbData,
			&OutputBuffer[Index],
			ICODE_BLOCK_SIZE
		);
	}

	return status;
}

RESPONSECODE
UEM_CRD_ISO15693_WriteSingleBlock( unsigned int reader_index,
	BYTE   bFlags,
	BYTE   bBlockNumber,
	BYTE*  pbData,
	BYTE*  pbFlags,
	BYTE*  pbErrorCode)
{
	RESPONSECODE   			status = IFD_SUCCESS;
	
    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	const DWORD Timeout = 2500;         // 23,6 ms
	*pbErrorCode = 0;

	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);

	InputBuffer[0] = 0x48;
	InputBuffer[1] = 0x0B;
	InputBuffer[2] = 0x00;
	InputBuffer[3] = bFlags;
	InputBuffer[4] = 0x21; // Write Single Block
	memcpy(
		&InputBuffer[5],
		ccid_descriptor->uemCardUID,
		8
	);
	InputBuffer[13] = bBlockNumber;
	memcpy(
		&InputBuffer[14],
		pbData,
		ICODE_BLOCK_SIZE
	);
	memcpy(
		&InputBuffer[18],
		&Timeout,
		4
	);
	InputBufferLength = 22;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);

	*pbFlags = 0x01;

	if (status != IFD_SUCCESS) {
		return status;
	}
	
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}

	if (ulBytesReadOrWritten < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = OutputBuffer[3];
	if ((*pbFlags) & 0x01)
	{
		if (ulBytesReadOrWritten < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		else
		{
			*pbErrorCode = OutputBuffer[4];
			status = UEM_TranslateError(*pbErrorCode);
		}
	}

	return status;
}

RESPONSECODE
UEM_CRD_CmdARATS(unsigned int reader_index)
{	
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	
	InputBuffer[0] = 0x35;
	InputBuffer[1] = 0x00;
	InputBufferLength = 2;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);
	if (status != IFD_SUCCESS) {
	}
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}

	return status;
}

RESPONSECODE
UEM_CRD_CmdAPPS(unsigned int reader_index)
{	
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	
	InputBuffer[0] = 0x36;
	InputBuffer[1] = 0x00;
	InputBuffer[2] = 0x00; // 106 - 106
	InputBufferLength = 3;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);
	if (status != IFD_SUCCESS) {
	}
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}	

	return status;
}

RESPONSECODE
UEM_CRD_CmdRfOn(unsigned int reader_index)
{	
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CRD_CmdRfOn");
#endif
	
	//RF OFF ON
	InputBuffer[0] = 0x20;
	InputBuffer[1] = 0x0A;
	InputBuffer[2] = 0x00;
	InputBufferLength = 3;
	
	status = UEM_Transcieve(reader_index, 
		InputBuffer, InputBufferLength,
		OutputBuffer, &ulBytesReadOrWritten);
		
	/*// MFRC OFF
	InputBuffer[0] = 0x10;
	InputBufferLength = 1;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);*/
	
	if (status != IFD_SUCCESS) {
	}
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}

	return status;
}

RESPONSECODE
UEM_CRD_CmdRfOff(unsigned int reader_index)
{	
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CRD_CmdRfOff");
#endif
	
	//RF OFF ON
	InputBuffer[0] = 0x20;
	InputBuffer[1] = 0x00;
	InputBuffer[2] = 0x00;
	InputBufferLength = 3;
	
	status = UEM_Transcieve(reader_index, 
		InputBuffer, InputBufferLength,
		OutputBuffer, &ulBytesReadOrWritten);
		
	/*//MFRC OFF
	InputBuffer[0] = 0x04;
	InputBuffer[1] = 0x80;
	InputBuffer[2] = 0x01;
	InputBufferLength = 3;
	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);*/
	if (status != IFD_SUCCESS) {
	}
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}
	if (OutputBuffer[0] != 0x00) {
		return SCARD_F_INTERNAL_ERROR;
	}	

	return status;
}

RESPONSECODE
UEM_CRD_CmdLoadKeys(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{	
	// FF 82 20 0A 06 FF FF FF FF FF FF
	// FF 82 20 00 10 A5 A4 A3 A2 A1 A0 18 77 8E 69 F0 F0 F0 F0 F0 F0
    RESPONSECODE   			status = IFD_SUCCESS;
	
    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	BYTE					ucKeyStructure;
	BYTE					ucKeyNumber;
	BYTE					ucKeyLength;
	BYTE					lpKey[32] = {0};
	BOOLEAN					fReaderKey;
	BOOLEAN					fSecuredTransmission;
	BOOLEAN					fWriteToNonVolatileMemory;
	BYTE					ucReaderKeyNumber;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CRD_CmdLoadKeys");
#endif
	
	ucKeyStructure = tx_buffer[2];
	ucKeyNumber = tx_buffer[3];
	ucKeyLength = tx_buffer[4];
	fReaderKey = (ucKeyStructure & 0x80) ? TRUE : FALSE;
	fSecuredTransmission = (ucKeyStructure & 0x40) ? TRUE : FALSE;
	fWriteToNonVolatileMemory = (ucKeyStructure & 0x20) ? TRUE : FALSE;
	
	if ((!fReaderKey)
		&&(ccid_descriptor->uemCardType != CT_MIFARE_CLASSIC)
		&&(ccid_descriptor->uemCardType != CT_MIFARE_PLUS)
		&&(ccid_descriptor->uemCardType != CT_MIFARE_ULTRALIGHT)
		)
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x82;//card key not supported
		
		return IFD_SUCCESS;
	}
	
	if (fSecuredTransmission == TRUE)
	{
		ucReaderKeyNumber = ucKeyStructure & 0x0F;
	}
	
	if (fWriteToNonVolatileMemory == FALSE)
	{	// 69 86
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x86;
		return IFD_SUCCESS;
	}
	
	if (ucKeyLength > 16)
	{
		// 69 89
		*rx_length = 2;
		rx_buffer[0] = 0x69;
		rx_buffer[1] = 0x89;
		return IFD_SUCCESS;
	}		
	
	memcpy(
		lpKey,
		&tx_buffer[5],
		ucKeyLength
	);
	
	if (TRUE == fSecuredTransmission)
	{ // Auth reader
		InputBuffer[0] = 0x6F;
		InputBuffer[1] = 0xAA;
		InputBuffer[2] = ucKeyStructure;
		InputBuffer[3] = 0x00;
		InputBufferLength = 4;
		
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		
		if (status != IFD_SUCCESS)
			return status;
		
		if (ulBytesReadOrWritten == 0)
			return SCARD_F_COMM_ERROR;
		if (OutputBuffer[0] != 0x00)
			return SCARD_F_INTERNAL_ERROR;
	}
	
	if (fReaderKey == TRUE)
	{ // Reader key
		InputBuffer[0] = 0x79; // Write flash
		InputBuffer[1] = ucKeyNumber;
		InputBuffer[2] = 0x00;
		memcpy(
			&InputBuffer[3],
			lpKey,
			16
		);
		InputBufferLength = 19;
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		
		if (status != IFD_SUCCESS)
			goto quitCrypto;
		
		if (ulBytesReadOrWritten == 0) {
			status = SCARD_F_COMM_ERROR;
			goto quitCrypto;
		}
		if (OutputBuffer[0] != 0x00) {
			status = SCARD_F_INTERNAL_ERROR;
			goto quitCrypto;
		}
	}
	else
	{ // Card key
		InputBuffer[0] = 0x79; // Write flash
		InputBuffer[1] = ucKeyNumber;
		InputBuffer[2] = 0x00;
		memset(&InputBuffer[3], 0x00, 16);
		memcpy(
			&InputBuffer[3],
			lpKey,
			ucKeyLength
		);
		InputBufferLength = 19;
		
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		
		if (status != IFD_SUCCESS)
			goto quitCrypto;
		
		if (ulBytesReadOrWritten == 0) {
			status = SCARD_F_COMM_ERROR;
			goto quitCrypto;
		}
		if (OutputBuffer[0] != 0x00) {
			status = SCARD_F_INTERNAL_ERROR;
			goto quitCrypto;
		}
		
		switch (ccid_descriptor->uemCardType)
		{
		case CT_MIFARE_CLASSIC:
			InputBuffer[0] = 0x17;
			InputBuffer[1] = (BYTE)0x60; // Key A
			InputBuffer[2] = (BYTE)ucKeyNumber;
			memcpy(
				&InputBuffer[3],
				lpKey,
				6
			);
			InputBufferLength = 9;
			
			status = UEM_Transcieve(
				reader_index,
				InputBuffer,
				InputBufferLength,
				OutputBuffer,
				&ulBytesReadOrWritten
			);
			
			if (status != IFD_SUCCESS)
				break;

			if (ulBytesReadOrWritten == 0) {
				status = SCARD_F_COMM_ERROR;
				break;
			}
			if (OutputBuffer[0] != 0x00) {
				status = SCARD_F_INTERNAL_ERROR;
				break;
			}
			
			InputBuffer[1] = (BYTE)0x61; // Key B
			
			if (ucKeyLength == 12)
				memcpy(
					&InputBuffer[3],
					&lpKey[6],
					6
				);

			if (ucKeyLength == 16)
				memcpy(
					&InputBuffer[3],
					&lpKey[10],
					6
				);

			status = UEM_Transcieve(
				reader_index,
				InputBuffer,
				InputBufferLength,
				OutputBuffer,
				&ulBytesReadOrWritten
			);
			
			if (status != IFD_SUCCESS)
				break;

			if (ulBytesReadOrWritten == 0) {
				status = SCARD_F_COMM_ERROR;
				break;
			}
			if (OutputBuffer[0] != 0x00) {
				status = SCARD_F_INTERNAL_ERROR;
				break;
			}

			status = IFD_SUCCESS;
			
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;
		default:
			status = IFD_SUCCESS;
			
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x82;//card key not supported
		}
	}
	
quitCrypto:

	if (TRUE == fSecuredTransmission)
	{ // Reset auth reader
		InputBuffer[0] = 0x6F;
		InputBuffer[1] = 0x00;
		InputBuffer[2] = 0x00;
		InputBuffer[3] = 0x00;
		InputBufferLength = 4;
		
		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);
		
		if (status != IFD_SUCCESS)
			return status;

		if (ulBytesReadOrWritten == 0)
			return SCARD_F_COMM_ERROR;
		if (OutputBuffer[0] != 0x00)
			return SCARD_F_INTERNAL_ERROR;
	}

	return status;
}

RESPONSECODE
UEM_CRD_CmdGeneralAuthenticate(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	ULONG 					ulAddress;
	BYTE					ucKeyType;
	BYTE					ucKeyNumber;

	BYTE ucUemKeyType = 0x00;
	BYTE ucSectorNumber = 0x00;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	ulAddress = tx_buffer[6] * 256 + tx_buffer[7];
	ucKeyType = tx_buffer[8];
	ucKeyNumber = tx_buffer[9];
	
	switch (ccid_descriptor->uemCardType)
	{
	case CT_MIFARE_CLASSIC: // FF 86 00 00 05 01 00 28 60 0A
		InputBuffer[0] = 0x15;
		InputBuffer[1] = ucKeyType;
		memcpy(
			&InputBuffer[2],
			&ccid_descriptor->uemCardUID[ccid_descriptor->uemCardUIDLength - 4],
			4
		);
		InputBuffer[6] = ucKeyNumber;
		InputBuffer[7] = (BYTE)ulAddress;
		InputBufferLength = 8;

		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);

#ifdef UEM_ENABLE_DEBUG_OUTPUT
		for( unsigned int j = 0; j < InputBufferLength; j++ )	                                
			DEBUG_CRITICAL2("InputBuffer: %02X", InputBuffer[j]);
		for( unsigned int j = 0; j < ulBytesReadOrWritten; j++ )	                                
			DEBUG_CRITICAL2("OutputBuffer: %02X", OutputBuffer[j]);
#endif
						
		if (status != IFD_SUCCESS)
			return status;

		if ((ulBytesReadOrWritten == 0) || (OutputBuffer[0] != 0x00)) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x83;
			break;
		}
		
		ccid_descriptor->uemLastAuthCommandLength = (BYTE)InputBufferLength;
		memcpy(
			ccid_descriptor->uemLastAuthCommand,
			InputBuffer,
			ccid_descriptor->uemLastAuthCommandLength
		);
		ccid_descriptor->uemfAuthenticated = TRUE;
		
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;

		break;
	case CT_MIFARE_ULTRALIGHT: // FF 86 00 00 05 01 00 10 00 04
		InputBuffer[0] = 0x2D;
		InputBuffer[1] = ucKeyNumber & 0xFF;
		InputBuffer[2] = 0x00;//(ucKeyNumber >> 8) & 0xFF;
		InputBufferLength = 3;

		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);

		if (status != IFD_SUCCESS)
			return status;

		if ((ulBytesReadOrWritten == 0) || (OutputBuffer[0] != 0x00)) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x83;
			break;
		}
		
		ccid_descriptor->uemLastAuthCommandLength = 2;
		ccid_descriptor->uemLastAuthCommand[0] = 0x19;
		ccid_descriptor->uemLastAuthCommand[1] = 0x00;
		ccid_descriptor->uemfAuthenticated = TRUE;
		
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;

		break;
	case CT_MIFARE_PLUS: // FF 86 00 00 05 01 40 00 60 00
		//if (ccid_descriptor->uemfAuthenticated == FALSE)
		if (ccid_descriptor->uemfModeTCL == FALSE)
		{
		    InputBuffer[0] = 0x35; // RATS
			InputBuffer[1] = 0x00;
			InputBufferLength = 2;
			
            status = UEM_Transcieve(
				reader_index, 
				InputBuffer, 
				InputBufferLength, 
				OutputBuffer, 
				&ulBytesReadOrWritten);
				
			if (status != IFD_SUCCESS)
				return status;
			
			if ((ulBytesReadOrWritten == 0) || (OutputBuffer[0] != 0x00)) {
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x83;
				break;
			}
			
			ccid_descriptor->uemfModeTCL = TRUE;
		}
		
		UEM_AddressToSectorAndType(ulAddress, &ucUemKeyType, &ucSectorNumber, NULL);

		if (ucKeyType == 0x60) {
			ucUemKeyType = 0x4A;
		} else if (ucKeyType == 0x61) {
			ucUemKeyType = 0x4B;
		}
		
		InputBuffer[0] = 0xA0;
		InputBuffer[1] = 0x01;
		InputBuffer[2] = ucUemKeyType;
		InputBuffer[3] = ucSectorNumber;
		InputBuffer[4] = ucKeyNumber;
		InputBuffer[5] = 0x00;
		InputBuffer[6] = 0x00;
		InputBufferLength = 7;

		status = UEM_Transcieve(
			reader_index,
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten
		);

		if (status != IFD_SUCCESS)
			return status;

		if ((ulBytesReadOrWritten == 0) || (OutputBuffer[0] != 0x00)) {
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x83;
			break;
		}
		
		ccid_descriptor->uemLastAuthCommandLength = (BYTE)InputBufferLength;
		memcpy(
			ccid_descriptor->uemLastAuthCommand,
			InputBuffer,
			ccid_descriptor->uemLastAuthCommandLength
		);
		ccid_descriptor->uemfAuthenticated = TRUE;
		
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;

		break;
	default:
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x83;
	}	
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdReadBinary(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;
	
    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	ULONG 					ulAddress;
	ULONG 					ulNumBytesToRead;
	ULONG 					ulNumBytesRead;
	BOOLEAN					fErrorOccured;
	
	BYTE 					ucSectorNumber;
	BYTE 					ucBlockNumber;
	BYTE					ucUemValueType;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	ulAddress = tx_buffer[2] * 256 + tx_buffer[3];
	ulNumBytesToRead = tx_buffer[4];
	
	if (ccid_descriptor->uemCardStandard == CS_ISO14443A)
	{
#ifdef UEM_ENABLE_DEBUG_OUTPUT
		DEBUG_CRITICAL("CS_ISO14443A");
#endif
		switch (ccid_descriptor->uemCardType)
		{
		case CT_MIFARE_ULTRALIGHT: // FF B0 00 28 10
#ifdef UEM_ENABLE_DEBUG_OUTPUT
			DEBUG_CRITICAL("case CT_MIFARE_ULTRALIGHT");
#endif
			fErrorOccured = FALSE;
			ulNumBytesRead = 0;
			if (ulNumBytesToRead == 0)
				ulNumBytesToRead = 0xFFFFFFFF;
			while (ulNumBytesRead < ulNumBytesToRead)
			{
				InputBuffer[0] = 0x19;
				InputBuffer[1] = (BYTE)ulAddress;
				InputBufferLength = 2;
				
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				
				if (status != IFD_SUCCESS) {
					fErrorOccured = TRUE;
					break;
				}
				
				if (ulBytesReadOrWritten == 0) {
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten < 17) {
					status = SCARD_E_COMM_DATA_LOST;
					fErrorOccured = TRUE;
					break;
				}
				
				memcpy(
					&rx_buffer[ulNumBytesRead],
					&OutputBuffer[1],
					16
				);

				ulAddress += 4;
				ulNumBytesRead += 16;
			}

			if (ulNumBytesRead == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if ((ulNumBytesToRead != 0xFFFFFFFF) && (fErrorOccured == TRUE))
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			if (ulNumBytesToRead == 0xFFFFFFFF)
				ulNumBytesToRead = ulNumBytesRead;
			
			*rx_length = ulNumBytesToRead + 2;

			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;
		case CT_MIFARE_CLASSIC: // FF B0 00 28 10
#ifdef UEM_ENABLE_DEBUG_OUTPUT
			DEBUG_CRITICAL("case CT_MIFARE_CLASSIC");
#endif
			fErrorOccured = FALSE;
			ulNumBytesRead = 0;
			if (ulNumBytesToRead == 0)
				ulNumBytesToRead = 0xFFFFFFFF;
			while (ulNumBytesRead < ulNumBytesToRead)
			{
				InputBuffer[0] = 0x19;
				InputBuffer[1] = (BYTE)ulAddress;
				InputBufferLength = 2;
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				
				if (status != IFD_SUCCESS) {
					DEBUG_CRITICAL2("status=0x%08X", status);
					fErrorOccured = TRUE;
					break;
				}
				
				if (ulBytesReadOrWritten == 0) {
					DEBUG_CRITICAL("status=SCARD_F_COMM_ERROR");
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
#ifdef UEM_ENABLE_DEBUG_OUTPUT
					DEBUG_CRITICAL("status=SCARD_F_INTERNAL_ERROR");
					for( unsigned int j = 0; j < InputBufferLength; j++ )	                                
						DEBUG_CRITICAL2("InputBuffer: %02X", InputBuffer[j]);
					for( unsigned int j = 0; j < ulBytesReadOrWritten; j++ )	                                
						DEBUG_CRITICAL2("OutputBuffer: %02X", OutputBuffer[j]);
#endif
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten < 17) {
#ifdef UEM_ENABLE_DEBUG_OUTPUT
					DEBUG_CRITICAL("status=SCARD_E_COMM_DATA_LOST");
#endif
					status = SCARD_E_COMM_DATA_LOST;
					fErrorOccured = TRUE;
					break;
				}
			
				memcpy(
					&rx_buffer[ulNumBytesRead],
					&OutputBuffer[1],
					16
				);

				ulAddress += 1;
				ulNumBytesRead += 16;
			}

			if (ulNumBytesRead == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if ((ulNumBytesToRead != 0xFFFFFFFF) && (fErrorOccured == TRUE))
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			if (ulNumBytesToRead == 0xFFFFFFFF)
				ulNumBytesToRead = ulNumBytesRead;
			
			*rx_length = ulNumBytesToRead + 2;

			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;
		case CT_MIFARE_PLUS: // FF B0 00 01 10
#ifdef UEM_ENABLE_DEBUG_OUTPUT
			DEBUG_CRITICAL("case CT_MIFARE_PLUS");
#endif
			// 4806001B003600000300001000
			fErrorOccured = FALSE;
			ulNumBytesRead = 0;
			
			if (ulNumBytesToRead == 0)
				ulNumBytesToRead = 0x00000FF0;
			
			while (ulNumBytesRead < ulNumBytesToRead)
			{
				UEM_AddressToSectorAndType(ulAddress, &ucUemValueType, 
										&ucSectorNumber, &ucBlockNumber);
			
				if ((ulNumBytesToRead > 1) && ((ucBlockNumber == 3) || (ucBlockNumber == 15)))
				{
					ulAddress += 1;
					continue;
				}
				
				InputBuffer[0] = 0xA4;
				InputBuffer[1] = 0x01;//0x06;
				InputBuffer[2] = ucUemValueType;
				InputBuffer[3] = ucSectorNumber;
				InputBuffer[4] = ucBlockNumber;
				InputBuffer[5] = 0x01;//ulNumBytesToRead;
				InputBufferLength = 6;
				
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				
				if (status != IFD_SUCCESS) {
					fErrorOccured = TRUE;
					break;
				}
				
				if (ulBytesReadOrWritten == 0) {
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten < 19) {
					status = SCARD_E_COMM_DATA_LOST;
					fErrorOccured = TRUE;
					break;
				}
				
				memcpy(
					&rx_buffer[ulNumBytesRead],
					&OutputBuffer[3],
					16
				);

				ulAddress += 1;
				ulNumBytesRead += 16;
			}

			if (ulNumBytesRead == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if ((ulNumBytesToRead != 0x00000FF0) && (fErrorOccured == TRUE))
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			if (ulNumBytesToRead == 0x00000FF0)
				ulNumBytesToRead = ulNumBytesRead;
			
			*rx_length = ulNumBytesToRead + 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;
		default:
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x6A;
			rx_buffer[(*rx_length) - 1] = 0x81;
		}
	}
	else if (ccid_descriptor->uemCardStandard == CS_ISO15693)
	{
#ifdef UEM_ENABLE_DEBUG_OUTPUT
		DEBUG_CRITICAL("CS_ISO15693");
#endif
		fErrorOccured = FALSE;
		ulNumBytesRead = 0;
		if (ulNumBytesToRead == 0)
			ulNumBytesToRead = 0xFFFFFFFF;
		while (ulNumBytesRead < ulNumBytesToRead)
		{
			BYTE ucFlags = F_DATA_RATE | F_ADDRESS;
			BYTE BlockSecurityStatus;
			BYTE Error = 0;
			
			if( ccid_descriptor->uemCardUID[6] == 0x07 ) // Texas Instruments
				ucFlags |= F_OPTION;

			status = UEM_CRD_ISO15693_ReadSingleBlock(
				reader_index,
				ucFlags,
				(BYTE)ulAddress,
				&ucFlags,
				&BlockSecurityStatus,
				&rx_buffer[ulNumBytesRead],
				&Error );
			
			if (status != IFD_SUCCESS) 
			{
				fErrorOccured = TRUE;
				break;
			}
			
			ulAddress += 1;
			ulNumBytesRead += 4;
		}

		if (ulNumBytesRead == 0)
		{
			// 69 81
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x81;
			return IFD_SUCCESS;
		}
		
		if ((ulNumBytesToRead != 0xFFFFFFFF) && (fErrorOccured == TRUE))
		{
			// 62 82
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x62;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return IFD_SUCCESS;
		}
		
		status = IFD_SUCCESS;
		
		if (ulNumBytesToRead == 0xFFFFFFFF)
			ulNumBytesToRead = ulNumBytesRead;
		
		*rx_length = ulNumBytesToRead + 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;
	}
	else
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x6A;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdUpdateBinary(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
	// FF D6 00 0A 04 00 00 00 02
    RESPONSECODE   			status = IFD_SUCCESS;
	
    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;

	ULONG 					ulAddress;
	ULONG 					ulNumBytesToWrite;
	ULONG 					ulNumBytesWritten;
	BOOLEAN					fErrorOccured;
	
	BYTE 					ucSectorNumber;
	BYTE 					ucBlockNumber;
	BYTE					ucUemValueType;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	ulAddress = tx_buffer[2] * 256 + tx_buffer[3];
	ulNumBytesToWrite = tx_buffer[4];
	
	if (ulNumBytesToWrite & 0x03)
	{	// 6A 81 Data object not supported
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x6A;
		rx_buffer[(*rx_length) - 1] = 0x81;
		return status;
	}
	
	if (ulNumBytesToWrite == 0x00) 
	{ 	// 6A 81 Data object not supported
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x6A;
		rx_buffer[(*rx_length) - 1] = 0x81;
		return status;
	}
	
	if (ccid_descriptor->uemCardStandard == CS_ISO14443A)
	{
		switch (ccid_descriptor->uemCardType)
		{
		case CT_MIFARE_ULTRALIGHT:
			fErrorOccured = FALSE;
			ulNumBytesWritten = 0;
			while (ulNumBytesWritten < ulNumBytesToWrite)
			{
				InputBuffer[0] = 0x1E;
				InputBuffer[1] = (BYTE)ulAddress;
				memcpy(
					&InputBuffer[2],
					&tx_buffer[5 + ulNumBytesWritten],
					4
				);
				InputBufferLength = 6;
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				if (status != IFD_SUCCESS) {
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten == 0) {
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				
				ulAddress += 1;
				ulNumBytesWritten += 4;
			}

			if (ulNumBytesWritten == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if (fErrorOccured == TRUE)
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;
		case CT_MIFARE_CLASSIC: // FF D6 00 28 10 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 02
			fErrorOccured = FALSE;
			ulNumBytesWritten = 0;
			while (ulNumBytesWritten < ulNumBytesToWrite)
			{
				InputBuffer[0] = 0x1A;
				InputBuffer[1] = (BYTE)ulAddress;
				memcpy(
					&InputBuffer[2],
					&tx_buffer[5 + ulNumBytesWritten],
					16
				);
				InputBufferLength = 18;
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				if (status != IFD_SUCCESS) {
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten == 0) {
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}

				ulAddress += 4;
				ulNumBytesWritten += 16;
			}

			if (ulNumBytesWritten == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if (fErrorOccured == TRUE)
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;	
		case CT_MIFARE_PLUS: // FF D6 00 28 10 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 01
			fErrorOccured = FALSE;
			ulNumBytesWritten = 0;
			while (ulNumBytesWritten < ulNumBytesToWrite)
			{
				UEM_AddressToSectorAndType(ulAddress, &ucUemValueType, 
										&ucSectorNumber, &ucBlockNumber);
			
				if ((ulNumBytesToWrite > 1) && ((ucBlockNumber == 3) || (ucBlockNumber == 15)))
				{
					ulAddress += 1;
					continue;
				}
				
				InputBuffer[0] = 0xA5;
				InputBuffer[1] = 0x01;//0x06;
				InputBuffer[2] = ucUemValueType;
				InputBuffer[3] = ucSectorNumber;
				InputBuffer[4] = ucBlockNumber;
				InputBuffer[5] = 0x01;//ulNumBytesToWrite;
				memcpy(
					&InputBuffer[6],
					&tx_buffer[5 + ulNumBytesWritten],
					16
				);
				InputBufferLength = 22;
				status = UEM_Transcieve(
					reader_index,
					InputBuffer,
					InputBufferLength,
					OutputBuffer,
					&ulBytesReadOrWritten
				);
				if (status != IFD_SUCCESS) {
					fErrorOccured = TRUE;
					break;
				}
				if (ulBytesReadOrWritten == 0) {
					status = SCARD_F_COMM_ERROR;
					fErrorOccured = TRUE;
					break;
				}
				if (OutputBuffer[0] != 0x00) {
					status = SCARD_F_INTERNAL_ERROR;
					fErrorOccured = TRUE;
					break;
				}

				ulAddress += 1;
				ulNumBytesWritten += 16;
			}

			if (ulNumBytesWritten == 0)
			{
				// 69 81
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x69;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			
			if (fErrorOccured == TRUE)
			{
				// 62 82
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x62;
				rx_buffer[(*rx_length) - 1] = 0x82;
				return IFD_SUCCESS;
			}
			
			status = IFD_SUCCESS;
			
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;

			break;

		default:
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x6A;
			rx_buffer[(*rx_length) - 1] = 0x81;
		}
	}
	else if (ccid_descriptor->uemCardStandard == CS_ISO15693)
	{
		fErrorOccured = FALSE;
		ulNumBytesWritten = 0;
		while (ulNumBytesWritten < ulNumBytesToWrite)
		{
			BYTE ucFlags = F_DATA_RATE | F_ADDRESS;
			BYTE Error = 0;
			if( ccid_descriptor->uemCardUID[6] == 0x07 ) // Texas Instruments
			{
				ucFlags |= F_OPTION;
			}

			status = UEM_CRD_ISO15693_WriteSingleBlock(
				reader_index,
				ucFlags,
				(BYTE)ulAddress,
				(BYTE*)&tx_buffer[5 + ulNumBytesWritten],
				(BYTE*)&ucFlags,
				(BYTE*)&Error );
			
			if (status != IFD_SUCCESS) {
				fErrorOccured = TRUE;
				break;
			}
			
			ulAddress += 1;
			ulNumBytesWritten += 4;
		}

		if (ulNumBytesWritten == 0)
		{
			// 69 81
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x81;
			return IFD_SUCCESS;
		}
		
		if (fErrorOccured == TRUE)
		{
			// 62 82
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x62;
			rx_buffer[(*rx_length) - 1] = 0x82;
			return IFD_SUCCESS;
		}
		
		status = IFD_SUCCESS;
		
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x90;
		rx_buffer[(*rx_length) - 1] = 0x00;
	}
	else
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x6A;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdIncrementDecrementValue(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;

	BYTE ucTLVLength;
	BYTE* pTLV;
	BYTE ucCmdFlags = 0x00;
	BYTE ucCurCmd = 0x00;
	DWORD dwCurSrcAddress = 0;
	DWORD dwCurDestAddress = 0;
	DWORD dwCurValue = 0;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	ucTLVLength = tx_buffer[4];
	pTLV = &tx_buffer[5];
	
	while (pTLV - &tx_buffer[5] < ucTLVLength)
	{
		switch (*pTLV)
		{
		case 0xA0: // Increment
		case 0xA1: // Decrement
			ucCurCmd = *pTLV;
			if (ucCmdFlags > 0x00)
			{ // FF C2 00 03 0B A0 09 80 01 28 81 04 01 00 00 00 MfCl
			// FF C2 00 03 0C A0 0A 80 02 01 00 81 04 01 00 00 00 MfPl
				status = UEM_CRD_IncrementDecrementValue(
					reader_index,
					rx_length,
					rx_buffer,
					ucCurCmd, 
					dwCurSrcAddress, 
					dwCurDestAddress, 
					dwCurValue);
					
				if (status != IFD_SUCCESS)
					return status;

				ucCmdFlags = 0x00;
			}
			
			pTLV += 2;
		
			break;
		case 0x80: // Block address
			if ((pTLV[1] == 0) || (pTLV[1] > 4))
			{
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x6A;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			if ((ucCmdFlags & 0x01) == 0x01) // Setting destination block address
			{
				ucCmdFlags |= 0x02;
				dwCurDestAddress = 0;
				memcpy(
					&dwCurDestAddress,
					&pTLV[2],
					pTLV[1]
				);
				pTLV += 2 + pTLV[1];
			} 
			else // Setting source block address
			{
				ucCmdFlags |= 0x01;
				dwCurSrcAddress = 0;
				memcpy(
					&dwCurSrcAddress,
					&pTLV[2],
					pTLV[1]
				);
				dwCurDestAddress = dwCurSrcAddress;
				pTLV += 2 + pTLV[1];
			}
			break;
		case 0x81: // Setting value
			ucCmdFlags |= 0x04;
			if ((pTLV[1] == 0) || (pTLV[1] > 4))
			{
				*rx_length = 2;
				rx_buffer[(*rx_length) - 2] = 0x6A;
				rx_buffer[(*rx_length) - 1] = 0x81;
				return IFD_SUCCESS;
			}
			dwCurValue = 0;
			memcpy(
				&dwCurValue,
				&pTLV[2],
				pTLV[1]
			);
			pTLV += 2 + pTLV[1];
			break;
		default:
			*rx_length = 2;
			rx_buffer[(*rx_length) - 2] = 0x69;
			rx_buffer[(*rx_length) - 1] = 0x81;
			return IFD_SUCCESS;
		}
	}
	
	if (pTLV - &tx_buffer[5] == ucTLVLength)
	{
		if (ucCmdFlags > 0x00)
		{
			status = UEM_CRD_IncrementDecrementValue(reader_index, rx_length, rx_buffer, ucCurCmd, dwCurSrcAddress, dwCurDestAddress, dwCurValue);
		}
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdManageSession(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;
	BYTE					ucTLVLength;
	BYTE*					pTLV;
	BYTE					ucIndexOfTLV = 0;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL2("rx_length=%d", *rx_length);
#endif
	
	ucTLVLength = tx_buffer[4];
	pTLV = &tx_buffer[5];
	
	*rx_length = 0;
	rx_buffer[(*rx_length)++] = 0xC0;
	rx_buffer[(*rx_length)++] = 0x03;
	*rx_length += 3;
	while (pTLV - &tx_buffer[5] < ucTLVLength)
	{
		ucIndexOfTLV++;
		switch (*pTLV)
		{
		case 0x80: // Version // 
			{
				//UCHAR ucMajor, ucMinor, ucBuild; // 02 01 00
				//DEBUG_CRITICAL2("Version begin, %d", (*rx_length));
				//ucMajor = pTLV[2];
				//ucMinor = pTLV[3];
				//ucBuild = pTLV[4];
				rx_buffer[(*rx_length)++] = 0x80;
				//DEBUG_CRITICAL2("d1, %d", (*rx_length));
				rx_buffer[(*rx_length)++] = 0x03;
				//DEBUG_CRITICAL2("d2, %d", (*rx_length));
				rx_buffer[(*rx_length)++] = 0x02;
				//DEBUG_CRITICAL2("d3, %d", (*rx_length));
				rx_buffer[(*rx_length)++] = 0x01;
				//DEBUG_CRITICAL2("d4, %d", (*rx_length));
				rx_buffer[(*rx_length)++] = 0x00;
				//DEBUG_CRITICAL2("Version end, %d", (*rx_length));
			}
			pTLV += 2 + pTLV[1];
			break;
		case 0x81: // Start Session FFC20000028100
			if (ccid_descriptor->uemfSessionActive != TRUE)
			{
				ccid_descriptor->uemfSessionActive = TRUE;
				UEM_SuspendCards(reader_index, FLAG_CRD);
			}
			pTLV += 2 + pTLV[1];
			break;
		case 0x82: // End Session FFC20000028200
			if (ccid_descriptor->uemfSessionActive != FALSE)
			{
				UEM_CRD_CmdRfOn(reader_index);
				UEM_SuspendCards(reader_index, 0x00);
				ccid_descriptor->uemfSessionActive = FALSE;
			}
			pTLV += 2 + pTLV[1];
			break;
		case 0x83: // RF Off
			if (ccid_descriptor->uemfSessionActive == TRUE)
				status = UEM_CRD_CmdRfOff(reader_index);
			pTLV += 2 + pTLV[1];
			break;
		case 0x84: // Rf On
			if (ccid_descriptor->uemfSessionActive == TRUE)
				status = UEM_CRD_CmdRfOn(reader_index);
			pTLV += 2 + pTLV[1];
			break;
		case 0x5F: // Timer
			if (pTLV[1] == 0x46)
			{
				DWORD dwTimeUS;
				//KFLOATING_SAVE  saveData;
				//NTSTATUS        status2;
				memcpy(
					 &dwTimeUS,
					 &pTLV[3],
					 pTLV[2]
				);
				/*status2 = KeSaveFloatingPointState(&saveData);
				if (status2 != IFD_SUCCESS)
				{
					return status2;
				}*/
				//SysDelay((DWORD)(((double)dwTimeUS) / 1000));
				usleep(dwTimeUS);
				//KeRestoreFloatingPointState(&saveData);
				pTLV += 3 + pTLV[2];
			}
			else
				pTLV += 2 + pTLV[1];
			break;
		case 0xFF: // Parameters
			if (pTLV[1] == 0x6D)
			{
				pTLV += 3 + pTLV[2];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else if (pTLV[1] == 0x6E)
			{
				pTLV += 3 + pTLV[2];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else
			{
				pTLV += 2 + pTLV[1];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			break;
		default:
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
		}
	}
	
	if (pTLV - &tx_buffer[5] == ucTLVLength)
	{
		rx_buffer[2] = 0x00;
		rx_buffer[3] = 0x90;
		rx_buffer[4] = 0x00;
		rx_buffer[(*rx_length)++] = 0x90;
		rx_buffer[(*rx_length)++] = 0x00;
	}
	else
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdTranscieveCustom(
	unsigned int reader_index, 
	BOOLEAN fTxAppendCRC,
	BOOLEAN fRxExpectCRC,
	BOOLEAN fUseParityBit,
	BOOLEAN fPrologueTxAppendOrRxDiscard,
	BYTE ucTxLastBitsCount,
	BYTE* pDataTLV, 
	DWORD dwLengthTLV,
	BYTE* pOutData,
	DWORD* pdwOutLength,
	BYTE* pucRxLastBitsCount)
{	
    RESPONSECODE   			status = IFD_SUCCESS;

    BYTE					InputBuffer[SCR_BUFFER_SIZE];
    ULONG					InputBufferLength = sizeof(InputBuffer);
    BYTE					OutputBuffer[SCR_BUFFER_SIZE];
    ULONG					OutputBufferLength = sizeof(OutputBuffer);
	ULONG					ulBytesReadOrWritten = 0;
	int iPos = 0;
	
	DWORD dwTimeout = 6800;
	
	DWORD dwOptions = 0x00000000;
	
	// Transmit FFC20001LL 95LL 48040030000014901A0000
	// FFC2000106950430000014
	
	BYTE ucPCB = 0x00, ucCID = 0x00;
	
	if (fTxAppendCRC == TRUE)
		dwOptions |= 0x00000001 << 4;
	if (fRxExpectCRC == TRUE)
		dwOptions |= 0x00000001 << 5;
	if (fUseParityBit == TRUE)
		dwOptions |= 0x00000001 << 6;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	InputBuffer[iPos++] = 0x46;
	memcpy(
		&InputBuffer[iPos],
		&dwTimeout,
		2
	);
	iPos += 2;
	memcpy(
		&InputBuffer[iPos],
		&dwOptions,
		2
	);
	iPos += 2;
	InputBuffer[iPos++] = ucTxLastBitsCount;
	if (fPrologueTxAppendOrRxDiscard == TRUE)
	{
		InputBuffer[iPos++] = (dwLengthTLV + 2) & 0x000000FF;
		InputBuffer[iPos++] = ((dwLengthTLV + 2) >> 8) & 0x000000FF;
		InputBuffer[iPos++] = ucPCB = 0x02 + ccid_descriptor->uembnBit; // PCB
		ccid_descriptor->uembnBit ^= 1;
		//InputBuffer[iPos++] = ucCID = 0x00; // CID
	}
	else
	{
		InputBuffer[iPos++] = dwLengthTLV & 0x000000FF;
		InputBuffer[iPos++] = (dwLengthTLV >> 8) & 0x000000FF;
	}
	memcpy(
		&InputBuffer[iPos],
		pDataTLV,
		dwLengthTLV
	);
	iPos += dwLengthTLV;
	InputBufferLength = iPos;

	status = UEM_Transcieve(
		reader_index,
		InputBuffer,
		InputBufferLength,
		OutputBuffer,
		&ulBytesReadOrWritten
	);

	if (status != IFD_SUCCESS) {
		return status;
	}
	
	if (ulBytesReadOrWritten == 0) {
		return SCARD_F_COMM_ERROR;
	}

	if (ulBytesReadOrWritten < 3) { // RxLastbitsCount and Length are absent
		if (status == IFD_SUCCESS)
			status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		return status;
	} else {
		*pucRxLastBitsCount = OutputBuffer[1];
		*pdwOutLength = OutputBuffer[2] + OutputBuffer[3] * 256;

		if (ulBytesReadOrWritten != ((*pdwOutLength) + 4)) {
			if (status == SCARD_S_SUCCESS)
				return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		
		if (fPrologueTxAppendOrRxDiscard == TRUE)
		{
			if (*pdwOutLength < 2)
				return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			
			if (OutputBuffer[4] != ucPCB)
				return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			
			//if (OutputBuffer[5] != ucCID)
			//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			
			*pdwOutLength -= 2;
			
			if (*pdwOutLength > 0)
				memcpy(
					pOutData,
					&OutputBuffer[6],
					*pdwOutLength
				);
		}
		else
		{
			if (*pdwOutLength > 0)
				memcpy(
					pOutData,
					&OutputBuffer[4],
					*pdwOutLength
				);
		}
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdTransparentExchange(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;

	DWORD					dwTLVsLength;
	BYTE*					pTLV;
	BYTE					ucIndexOfTLV = 0;
	DWORD					dwLengthTLV = 0;
	BYTE					ucTypeTLV = 0;

	BOOLEAN					fCustomTransceive = FALSE;
	BOOLEAN 				fTxAppendCRC = FALSE;
	BOOLEAN 				fRxExpectCRC = FALSE;
	BOOLEAN 				fUseParityBit = FALSE;
	BOOLEAN					fPrologueTxAppendOrRxDiscard = FALSE;
	BYTE 					ucTxLastBitsCount = 0x00;
	BYTE 					ucRxLastBitsCount = 0x00;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
	dwTLVsLength = tx_buffer[4];
	pTLV = &tx_buffer[5];
	
	*rx_length = 0;
	rx_buffer[(*rx_length)++] = 0xC0;
	rx_buffer[(*rx_length)++] = 0x03;
	*rx_length += 3;
	while ((DWORD)(pTLV - &tx_buffer[5]) < dwTLVsLength)
	{
		ucTypeTLV = *pTLV;
		ucIndexOfTLV++;
		dwLengthTLV = 0;
		if (pTLV[1] <= 0x7F)
		{
			dwLengthTLV = pTLV[1];
			pTLV += 1;
		}
		else if (pTLV[1] == 0x81)
		{
			memcpy(
				 &dwLengthTLV,
				 &pTLV[2],
				 1
			);
			pTLV += 2;
		}
		else if (pTLV[1] == 0x82)
		{
			memcpy(
				 &dwLengthTLV,
				 &pTLV[2],
				 2
			);
			pTLV += 3;
		}
		else if (pTLV[1] == 0x83)
		{
			memcpy(
				 &dwLengthTLV,
				 &pTLV[2],
				 3
			);
			pTLV += 4;
		}
		else if (pTLV[1] == 0x84)
		{
			memcpy(
				 &dwLengthTLV,
				 &pTLV[2],
				 4
			);
			pTLV += 5;
		}
		else
		{
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
		}
		
		switch (ucTypeTLV)
		{
		case 0x90: // Transmission and reception flag
			fCustomTransceive = TRUE;
			if (dwLengthTLV == 2)
			{
				if ((pTLV[1] & 0x01) == 0x00)
					fTxAppendCRC = TRUE;
				if ((pTLV[1] & 0x02) == 0x00)
					fRxExpectCRC = TRUE;
				if ((pTLV[1] & 0x04) == 0x00)
					fUseParityBit = TRUE;//Tx Parity Bit
				if ((pTLV[1] & 0x10) == 0x00)
					fPrologueTxAppendOrRxDiscard = TRUE;//Append prologue in transmit or discard from response
			}
			pTLV += 1 + dwLengthTLV;
			break;
		case 0x91: // Transmission bit framing
			fCustomTransceive = TRUE;
			if (dwLengthTLV == 1)
				ucTxLastBitsCount = pTLV[1];
			pTLV += 1 + dwLengthTLV;
			break;
		case 0x92: // Reception bit framing
			fCustomTransceive = TRUE;
			if (dwLengthTLV == 1)
				ucRxLastBitsCount = pTLV[1];
			pTLV += 1 + dwLengthTLV;
			break;
		case 0x93: // Transmit
			pTLV += 1 + dwLengthTLV;
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
			break;
		case 0x94: // Receive
			pTLV += 1 + dwLengthTLV;
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
			break;
		case 0x95: // Transceive
			{
				BYTE OutData[1024];
				unsigned int uiOutLength = 0;
				DWORD dwOutLength = 0;
				
				if ((fCustomTransceive == FALSE) || (ccid_descriptor->uemfModeTCL == TRUE))
				{
					//status = CRD_CmdTranscieve(SmartcardExtension, &pTLV[1], dwLengthTLV, OutData, &dwOutLength);
					status = UEM_CRD_toCard(reader_index, &pTLV[1], dwLengthTLV, OutData, sizeof(OutData), &uiOutLength);
					dwOutLength = uiOutLength;
				}
				else
				{
					status = UEM_CRD_CmdTranscieveCustom(
						reader_index, 
						fTxAppendCRC,
						fRxExpectCRC,
						fUseParityBit,
						fPrologueTxAppendOrRxDiscard,
						ucTxLastBitsCount,
						&pTLV[1], 
						dwLengthTLV, 
						OutData, 
						&dwOutLength,
						&ucRxLastBitsCount);
				}
				if (status != IFD_SUCCESS)
				{
					rx_buffer[0x02] = ucIndexOfTLV;
					rx_buffer[0x03] = 0x6A;
					rx_buffer[0x04] = 0x80;
					rx_buffer[(*rx_length)++] = 0x90;
					rx_buffer[(*rx_length)++] = 0x00;
					return IFD_SUCCESS;
				}
				rx_buffer[(*rx_length)++] = 0x97;
				if (dwOutLength <= 0x7F)
				{
					rx_buffer[(*rx_length)++] = (BYTE)dwOutLength;
				}
				else if (dwOutLength <= 0xFF)
				{
					rx_buffer[(*rx_length)++] = 0x81;
					rx_buffer[(*rx_length)++] = (BYTE)dwOutLength;
				}
				else if (dwOutLength <= 0xFFFF)
				{
					rx_buffer[(*rx_length)++] = 0x82;
					memcpy(
						 &rx_buffer[*rx_length],
						 &dwOutLength,
						 2
					);
					*rx_length += 2;
				}
				else if (dwOutLength <= 0xFFFFFF)
				{
					rx_buffer[(*rx_length)++] = 0x83;
					memcpy(
						 &rx_buffer[*rx_length],
						 &dwOutLength,
						 3
					);
					*rx_length += 3;
				}
				else if (dwOutLength <= 0xFFFFFFFF)
				{
					rx_buffer[(*rx_length)++] = 0x84;
					memcpy(
						 &rx_buffer[*rx_length],
						 &dwOutLength,
						 4
					);
					*rx_length += 4;
				}
				
				memcpy(
					 &rx_buffer[*rx_length],
					 OutData,
					 dwOutLength
				);
				*rx_length += dwOutLength;
				
				if ((fCustomTransceive == TRUE) && (ccid_descriptor->uemfModeTCL == FALSE) && (ucRxLastBitsCount > 0))
				{
					rx_buffer[(*rx_length)++] = 0x92;
					rx_buffer[(*rx_length)++] = 0x01;
					rx_buffer[(*rx_length)++] = ucRxLastBitsCount;
				}
			}
			pTLV += 1 + dwLengthTLV;
			break;
		case 0xFF: // Parameters
			if (pTLV[1] == 0x6D)
			{
				pTLV += 2 + dwLengthTLV;
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else if (pTLV[1] == 0x6E)
			{
				pTLV += 2 + dwLengthTLV;
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else
			{
				pTLV += 1 + dwLengthTLV;
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			break;
		default:
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
		}
	}
	
	if ((DWORD)(pTLV - &tx_buffer[5]) == dwTLVsLength)
	{
		rx_buffer[2] = 0x00;
		rx_buffer[3] = 0x90;
		rx_buffer[4] = 0x00;
		rx_buffer[(*rx_length)++] = 0x90;
		rx_buffer[(*rx_length)++] = 0x00;
	}
	else
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}
	
	return status;
}

RESPONSECODE
UEM_CRD_CmdSwitchProtocol(unsigned int reader_index, 
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
    RESPONSECODE   			status = IFD_SUCCESS;
	
	BYTE					ucTLVLength;
	BYTE*					pTLV;
	BYTE					ucIndexOfTLV = 0;
	
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CRD_CmdSwitchProtocol");
#endif
	
	ucTLVLength = tx_buffer[4];
	pTLV = &tx_buffer[5];
	
	*rx_length = 0;
	rx_buffer[(*rx_length)++] = 0xC0;
	rx_buffer[(*rx_length)++] = 0x03;
	*rx_length += 3;
	while (pTLV - &tx_buffer[5] < ucTLVLength)
	{
		ucIndexOfTLV++;
		switch (*pTLV)
		{
		case 0x8F: // Switch Protocol // 
			{
				switch (pTLV[2])
				{
					case 0x00: // ISO/IEC 14443 type A
						break;
					case 0x01: // ISO/IEC 14443 type B
						break;
					case 0x02: // ISO/IEC 15693
						break;
					case 0x03: // Felica
						break;
					case 0x04: // ICODE-EPC/UID
						break;
					case 0x05: // ICODE - 1
						break;
					case 0x06: // HF EPC G2/ISO18000-3 Mode 3
						break;
					case 0x07: // Innovatron
						break;
					default:
						break;
				}
				
				switch (pTLV[3])
				{
					case 0x00: // if no layer separation
					case 0x02: // Switch to layer 2
					case 0x03: // Switch or activate up to layer 3
						pTLV += 2 + pTLV[1];
						rx_buffer[0x02] = ucIndexOfTLV;
						rx_buffer[0x03] = 0x6A;
						rx_buffer[0x04] = 0x81;
						rx_buffer[(*rx_length)++] = 0x90;
						rx_buffer[(*rx_length)++] = 0x00;
						return IFD_SUCCESS;	
						break;
					case 0x04: // Activate up to layer 4
						if (pTLV[2] == 0x00) // ISO/IEC 14443 type A
						{
							if (ccid_descriptor->uemfModeTCL == TRUE)
							{
								pTLV += 2 + pTLV[1];
								rx_buffer[0x02] = ucIndexOfTLV;
								rx_buffer[0x03] = 0x6F;
								rx_buffer[0x04] = 0x00;
								rx_buffer[(*rx_length)++] = 0x90;
								rx_buffer[(*rx_length)++] = 0x00;
								return IFD_SUCCESS;		
							}
							status = UEM_CRD_CmdARATS(reader_index);
							if (status != IFD_SUCCESS)
							{
								pTLV += 2 + pTLV[1];
								rx_buffer[0x02] = ucIndexOfTLV;
								rx_buffer[0x03] = 0x6F;
								rx_buffer[0x04] = 0x00;
								rx_buffer[(*rx_length)++] = 0x90;
								rx_buffer[(*rx_length)++] = 0x00;
								return IFD_SUCCESS;								
							}
							status = UEM_CRD_CmdAPPS(reader_index);
							if (status != IFD_SUCCESS)
							{
								pTLV += 2 + pTLV[1];
								rx_buffer[0x02] = ucIndexOfTLV;
								rx_buffer[0x03] = 0x6F;
								rx_buffer[0x04] = 0x00;
								rx_buffer[(*rx_length)++] = 0x90;
								rx_buffer[(*rx_length)++] = 0x00;
								return IFD_SUCCESS;								
							}
							ccid_descriptor->uemfModeTCL = TRUE;
							ccid_descriptor->uembnBit = 0;
						}
						else
						{
							pTLV += 2 + pTLV[1];
							rx_buffer[0x02] = ucIndexOfTLV;
							rx_buffer[0x03] = 0x6A;
							rx_buffer[0x04] = 0x81;
							rx_buffer[(*rx_length)++] = 0x90;
							rx_buffer[(*rx_length)++] = 0x00;
							return IFD_SUCCESS;	
						}
						break;
					default:
						if ((pTLV[3] & 0xF0) == 0x20) // Switch to layer 2
						{
							pTLV += 2 + pTLV[1];
							rx_buffer[0x02] = ucIndexOfTLV;
							rx_buffer[0x03] = 0x6A;
							rx_buffer[0x04] = 0x81;
							rx_buffer[(*rx_length)++] = 0x90;
							rx_buffer[(*rx_length)++] = 0x00;
							return IFD_SUCCESS;	
						} 
						else if ((pTLV[3] & 0xF0) == 0x30) // Switch (activate) to layer 3
						{
							pTLV += 2 + pTLV[1];
							rx_buffer[0x02] = ucIndexOfTLV;
							rx_buffer[0x03] = 0x6A;
							rx_buffer[0x04] = 0x81;
							rx_buffer[(*rx_length)++] = 0x90;
							rx_buffer[(*rx_length)++] = 0x00;
							return IFD_SUCCESS;	
						}
						else if ((pTLV[3] & 0xF0) == 0x40) // Switch to high level protocol
						{
							BYTE ucProprietaryProtocolType = pTLV[3] & 0x0F;
							
							if (ucProprietaryProtocolType == 0x00) // ISO/IEC 7816-4
							{
							}
							pTLV += 2 + pTLV[1];
							rx_buffer[0x02] = ucIndexOfTLV;
							rx_buffer[0x03] = 0x6A;
							rx_buffer[0x04] = 0x81;
							rx_buffer[(*rx_length)++] = 0x90;
							rx_buffer[(*rx_length)++] = 0x00;
							return IFD_SUCCESS;	
						}
						else
						{
							pTLV += 2 + pTLV[1];
							rx_buffer[0x02] = ucIndexOfTLV;
							rx_buffer[0x03] = 0x6A;
							rx_buffer[0x04] = 0x81;
							rx_buffer[(*rx_length)++] = 0x90;
							rx_buffer[(*rx_length)++] = 0x00;
							return IFD_SUCCESS;	
						}
					break;
				}
			}
			pTLV += 2 + pTLV[1];
			break;
		case 0xFF: // Parameters
			if (pTLV[1] == 0x6D)
			{
				pTLV += 3 + pTLV[2];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else if (pTLV[1] == 0x6E)
			{
				pTLV += 3 + pTLV[2];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			else
			{
				pTLV += 2 + pTLV[1];
				rx_buffer[0x02] = ucIndexOfTLV;
				rx_buffer[0x03] = 0x6A;
				rx_buffer[0x04] = 0x81;
				rx_buffer[(*rx_length)++] = 0x90;
				rx_buffer[(*rx_length)++] = 0x00;
				return IFD_SUCCESS;
			}
			break;
		default:
			rx_buffer[0x02] = ucIndexOfTLV;
			rx_buffer[0x03] = 0x6A;
			rx_buffer[0x04] = 0x81;
			rx_buffer[(*rx_length)++] = 0x90;
			rx_buffer[(*rx_length)++] = 0x00;
			return IFD_SUCCESS;
		}
	}
	
	if (pTLV - &tx_buffer[5] == ucTLVLength)
	{
		rx_buffer[2] = 0x00;
		rx_buffer[3] = 0x90;
		rx_buffer[4] = 0x00;
		rx_buffer[(*rx_length)++] = 0x90;
		rx_buffer[(*rx_length)++] = 0x00;
	}
	else
	{
		*rx_length = 2;
		rx_buffer[(*rx_length) - 2] = 0x69;
		rx_buffer[(*rx_length) - 1] = 0x81;
	}
	
	return status;
}

static RESPONSECODE UEM_CmdXfrBlockTPDU_T1(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
	RESPONSECODE return_value = IFD_SUCCESS;
	int ret;
	BYTE APDU_GetUID[] = {0xFF, 0xCA, 0x00, 0x00, 0x00};
	BYTE APDU_LoadKeys[] = {0xFF, 0x82};
	BYTE APDU_GeneralAuthenticate[] = {0xFF, 0x86, 0x00, 0x00, 0x05};
	BYTE APDU_ReadBinary[] = {0xFF, 0xB0};//, 0x00, 0x00, 0x10};
	BYTE APDU_UpdateBinary[] = {0xFF, 0xD6};
	BYTE APDU_IncrementDecrementValue[] = {0xFF, 0xC2, 0x00, 0x03};
	BYTE APDU_ManageSession[] = {0xFF, 0xC2, 0x00, 0x00};
	BYTE APDU_TransparentExchange[] = {0xFF, 0xC2, 0x00, 0x01};
	BYTE APDU_SwitchProtocol[] = {0xFF, 0xC2, 0x00, 0x02};
	//unsigned char APDU_VendorSpecificUEM[] = {0xFF, 0x70, 0xC2/*VID MSB*/, 0x51/*VIDLSB, LC, DATA, LE*/};
	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
	
#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	DEBUG_CRITICAL("UEM_CmdXfrBlockTPDU_T1");
#endif
	//for( unsigned int i = 0; i < tx_length; i++ ) 
	//	DEBUG_CRITICAL2("Transmit: %02X", tx_buffer[i] );
		
	DEBUG_COMM3("T=1: %d and %d bytes", tx_length, *rx_length);

	*rx_length = 2;
	rx_buffer[(*rx_length) - 2] = 0x6A;
	rx_buffer[(*rx_length) - 1] = 0x81;

	if (ccid_descriptor->bCurrentSlotIndex == 0)
	{
		if (memcmp(APDU_GetUID, tx_buffer, sizeof(APDU_GetUID)) == 0)
		{
			memcpy(rx_buffer, ccid_descriptor->uemCardUID, ccid_descriptor->uemCardUIDLength);
			*rx_length = ccid_descriptor->uemCardUIDLength + 2;
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;
			return IFD_SUCCESS;
		}

		if (memcmp(APDU_LoadKeys, tx_buffer, sizeof(APDU_LoadKeys)) == 0)
		{
			return UEM_CRD_CmdLoadKeys(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_GeneralAuthenticate, tx_buffer, sizeof(APDU_GeneralAuthenticate)) == 0)
		{
			return UEM_CRD_CmdGeneralAuthenticate(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_ReadBinary, tx_buffer, sizeof(APDU_ReadBinary)) == 0)
		{
			return UEM_CRD_CmdReadBinary(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_UpdateBinary, tx_buffer, sizeof(APDU_UpdateBinary)) == 0)
		{
			return UEM_CRD_CmdUpdateBinary(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_IncrementDecrementValue, tx_buffer, sizeof(APDU_IncrementDecrementValue)) == 0)
		{
			return UEM_CRD_CmdIncrementDecrementValue(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_ManageSession, tx_buffer, sizeof(APDU_ManageSession)) == 0)
		{
			return UEM_CRD_CmdManageSession(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_TransparentExchange, tx_buffer, sizeof(APDU_TransparentExchange)) == 0)
		{
			return UEM_CRD_CmdTransparentExchange(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (memcmp(APDU_SwitchProtocol, tx_buffer, sizeof(APDU_SwitchProtocol)) == 0)
		{
			return UEM_CRD_CmdSwitchProtocol(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		if (ccid_descriptor->uemfModeTCL == TRUE)// && (ccid_descriptor->uemCardType != CT_MIFARE_PLUS))// && ccid_descriptor->uemCardType != CT_MIFARE_DESFIRE))
		{
			return UEM_CmdXfrBlockTPDU_RAW(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}
		
		/*if (memcmp(APDU_VendorSpecificUEM, tx_buffer, sizeof(APDU_VendorSpecificUEM)) == 0)
		{
			return UEM_CRD_CmdLoadKeys(reader_index, tx_length, tx_buffer, rx_length, rx_buffer);
		}*/
		
		/*ret = t1_transceive(&((get_ccid_slot(reader_index)) -> t1), 0,
			tx_buffer, tx_length, rx_buffer, *rx_length);

		if (ret < 0)
			return_value = IFD_COMMUNICATION_ERROR;
		else
			*rx_length = ret;*/
	}
	else
	{
		BYTE cmd[1024 * 64];//[SCR_BUFFER_SIZE];
		unsigned int cmd_length;
		ULONG length;
		BYTE buffer_tmp[1024 * 64];//[SCR_BUFFER_SIZE];

		// SAM
		// Set Holder
		cmd[0] = 0xB2;
		cmd[1] = 0x00 + (BYTE)ccid_descriptor->bCurrentSlotIndex; 
		cmd_length = 2;
		
		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			buffer_tmp, &length);
		if (return_value != IFD_SUCCESS)
		{
			//ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			return return_value;
		}
		if (buffer_tmp[0] != 0x00)
		{
			//ccid_descriptor->uemCardStatus = CCID_ICC_ABSENT;
			return IFD_COMMUNICATION_ERROR;
		}
		//ccid_descriptor->uemCardStatus = CCID_ICC_PRESENT_ACTIVE;
		if (memcmp(APDU_GetUID, tx_buffer, sizeof(APDU_GetUID)) == 0)
		{
			cmd[0] = 0xB6;
			cmd[1] = 0x05;
			cmd[2] = 0x00;
			cmd[3] = 0x80;
			cmd[4] = 0x60;
			cmd[5] = 0x00;
			cmd[6] = 0x00;
			cmd[7] = 0x00;
			cmd_length = 8;
			
			return_value = UEM_Transcieve(reader_index, 
				cmd, cmd_length,
				buffer_tmp, &length);
			if (return_value != IFD_SUCCESS)
				return return_value;
			if (buffer_tmp[0] != 0x00)
				return IFD_COMMUNICATION_ERROR;
			if (length < 30)
				return IFD_COMMUNICATION_ERROR;
				
			*rx_length = 7 + 2;
			memcpy(rx_buffer, &buffer_tmp[3+14], (*rx_length) - 2);
			rx_buffer[(*rx_length) - 2] = 0x90;
			rx_buffer[(*rx_length) - 1] = 0x00;
			
			return return_value;
		}
		
		cmd[0] = 0xB6;
		cmd[1] = 0xFF & tx_length;
		cmd[2] = 0xFF & (tx_length >> 8);
		memcpy(
			&cmd[3],
			tx_buffer,
			tx_length
		);
		cmd_length = 3 + tx_length;
		return_value = UEM_Transcieve(reader_index, 
			cmd, cmd_length,
			buffer_tmp, &length);
		if (return_value != IFD_SUCCESS)
			return return_value;
		if (length == 0)
			return IFD_COMMUNICATION_ERROR;
		if (buffer_tmp[0] != 0x00)
			return IFD_COMMUNICATION_ERROR;
		
		if (length < (*rx_length) + 3) 
			return IFD_COMMUNICATION_ERROR;
		
		*rx_length = buffer_tmp[1] + buffer_tmp[2] * 256;
			
		memcpy(
			rx_buffer,
			&buffer_tmp[3],
			*rx_length
		);

	}
	
	return return_value;
}

static RESPONSECODE UEM_CmdXfrBlockTPDU_RAW(unsigned int reader_index,
	unsigned int tx_length, BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[])
{
	RESPONSECODE status = IFD_SUCCESS;
	int ret;

#if UEM_ENABLE_DEBUG_OUTPUT == UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
	DEBUG_CRITICAL("UEM_CmdXfrBlockTPDU_RAW");
#endif

	DEBUG_COMM3("RAW: %d and %d bytes", tx_length, *rx_length);

	//for( unsigned int i = 0; i < tx_length; i++ ) 
	//	DEBUG_CRITICAL2("TPDU_RAW: %02X", tx_buffer[i] );
		
	/*ret = t1_transceive(&((get_ccid_slot(reader_index)) -> t1), 0,
		tx_buffer, tx_length, rx_buffer, *rx_length);

	if (ret < 0)
		return_value = IFD_COMMUNICATION_ERROR;
	else
		*rx_length = ret;*/

#ifdef UEM_USE_EXTENDED_ACTIVATION
		BYTE					InputBuffer[1024 * 64];//[SCR_BUFFER_SIZE];
		ULONG					InputBufferLength = sizeof(InputBuffer);
		BYTE					OutputBuffer[1024 * 64];//[SCR_BUFFER_SIZE];
		ULONG					OutputBufferLength = sizeof(OutputBuffer);
		DWORD					ulBytesReadOrWritten = 0;
		
		InputBuffer[0] = 0x47;
		InputBuffer[1] = 0x00;
		InputBuffer[2] = 0x00;
		memcpy(&InputBuffer[3], &tx_length, 2);
		memcpy(&InputBuffer[5], tx_buffer, tx_length);
		InputBufferLength = 5 + tx_length;

		ulBytesReadOrWritten = OutputBufferLength;
		status = UEM_Transcieve(reader_index, 
			InputBuffer,
			InputBufferLength,
			OutputBuffer,
			&ulBytesReadOrWritten);

		/*if (status == SCARD_F_INTERNAL_ERROR) {
			//m_fModeTCL = FALSE;
			//m_state = SCARD_POWERED;

			usleep(200 * 1000);
			status = m_device->m_pUemReader->iso14443AHalt();
			usleep(200 * 1000);
			status = m_device->m_pUemReader->iso14443AWakeUp(m_cardUID, m_ucCardUIDLength);

			//if (status != 0) {
			//	Sleep(200);
			//	status = m_device->m_pUemReader->iso14443AHalt();
			//	Sleep(200);
			//	status = m_device->m_pUemReader->iso14443AWakeUp(m_cardUID, m_ucCardUIDLength);
			//}

			if (status != 0) {
				m_state = SCARD_ABSENT;
				return status;
			}

			//Sleep(200);

			if (QueryATR(m_ATR, &m_ATRLength, false)) {
				rx_buffer[0] = 0x63;
				rx_buffer[1] = 0x00;
				*rx_length = 2;
				status = SCARD_S_SUCCESS;
			}
			else {
				return status;
			}

			//Sleep(200);
		}*/

		if (status == SCARD_S_SUCCESS) {
			if (ulBytesReadOrWritten < 3) {
				return SCARD_F_COMM_ERROR;
			}
			else {
				memcpy(rx_length, &OutputBuffer[1], 2);
				if (ulBytesReadOrWritten - 2 < *rx_length) {
					return SCARD_E_COMM_DATA_LOST;
				}
				memcpy(rx_buffer, &OutputBuffer[3], *rx_length);
			}
		}
#else
	#ifdef UEM_USE_CLESCAR_CHAINING
		status = transmit_ISO_14443_4(reader_index, tx_buffer, tx_length, rx_buffer, rx_length);
	#else
		status = UEM_CRD_toCard(reader_index, tx_buffer, tx_length, rx_buffer, *rx_length, rx_length);
	#endif
	//usleep(10 * 1000);
#endif

	return status;
}

static RESPONSECODE UEM_CmdSelectCardType(unsigned int reader_index, BYTE card_type)
{
	BYTE cmd[5];
	status_t res;
	unsigned int length;
	RESPONSECODE return_value = IFD_SUCCESS;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CmdSelectCardType NOT IMPLEMENTED");
#endif

	/*
	// SELECT_CARD_TYPE
	cmd[0] = 0x01;
	cmd[1] = 0x02;
	cmd[2] = 0x00;
	cmd[3] = 0x01;
	cmd[4] = card_type;

	res = WritePort(reader_index, sizeof(cmd), cmd);
	CHECK_STATUS(res)

	length = sizeof(cmd);
	res = ReadPort(reader_index, &length, cmd);
	CHECK_STATUS(res)

	if (cmd[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(cmd[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return_value = IFD_COMMUNICATION_ERROR;
	}
	*/
	
	return return_value;
}

static RESPONSECODE UEM_CmdSetOption(unsigned int reader_index, BYTE option)
{
	BYTE cmd[5];
	status_t res;
	unsigned int length;
	RESPONSECODE return_value = IFD_SUCCESS;

#ifdef UEM_ENABLE_DEBUG_OUTPUT
	DEBUG_CRITICAL("UEM_CmdSetOption NOT IMPLEMENTED");
#endif

	/*
	// SET_OPTION
	cmd[0] = 0x01;
	cmd[1] = 0x07;
	cmd[2] = 0x00;
	cmd[3] = 0x01;
	cmd[4] = option;

	res = WritePort(reader_index, sizeof(cmd), cmd);
	CHECK_STATUS(res)

	length = sizeof(cmd);
	res = ReadPort(reader_index, &length, cmd);
	CHECK_STATUS(res)

	if (cmd[UEM_STATUS_OFFSET] != 0)
	{
		uem_error(cmd[UEM_STATUS_OFFSET], __FILE__, __LINE__, __FUNCTION__);
		return_value = IFD_COMMUNICATION_ERROR;
	}
	*/
	
	return return_value;
}
